;; nomadic-pict-mode.el
;; ====================
;;
;; A major mode that supports Nomadic-Pict
;;
;; Copyright (c) 2002, Laurent Buffler 
;; 
;; Author : Laurent Buffler (Laurent@buffler.ch)
;; Created : April 2002
;; 
;; This file is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.
;;
;; This file is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this file; see the file COPYING.  If not, write to the
;; Free Software Foundation, Inc., 59 Temple Place - Suite 330,
;; Boston, MA 02111-1307, USA.

(defvar nomadic-pict-mode-syntax-table nil
  "Syntax table for nomadic pict mode")

(defvar nomadic-pict-mode-map (make-sparse-keymap)
  "Keymap used in nomadic-pict-mode buffers")

(defvar nomadic-pict-mode-hook nil
  "*List of functions to call when Nomadic Pict mode is invoked.")

(defvar nomadic-pict-indent-width 2
  "*Indent width")

(defvar nomadic-pict-auto-indent nil
  "*If this variable is nil, don't indent automatically after a newline")

(defvar nomadic-pict-hungry-backspace nil
  "*When this is non-nil, backspace will erase all spaces on left of the point")

(define-key nomadic-pict-mode-map "\t" 'nomadic-pict-indent-line)
(define-key nomadic-pict-mode-map "\C-m" 'nomadic-pict-newline)
(define-key nomadic-pict-mode-map "\C-c\C-s" 'nomadic-pict-skip-comment)
(define-key nomadic-pict-mode-map "|" 'nomadic-pict-electric-bar)
(define-key nomadic-pict-mode-map [backspace] 'nomadic-pict-backspace)
(define-key nomadic-pict-mode-map ")" 'nomadic-pict-electric-rpar)

(defun nomadic-pict-electric-bar ()
  (interactive)
  (insert "|")
  (insert-char ?\  (- nomadic-pict-indent-width 1))
  (if nomadic-pict-auto-indent
      (nomadic-pict-indent-line)))

(defun nomadic-pict-electric-rpar ()
  (interactive)
  (insert ")")
  (if nomadic-pict-auto-indent
      (nomadic-pict-indent-line))
  (blink-matching-open))

(defun nomadic-pict-backspace ()
  "Remove char before the point, and, if nomadic-pict-hungry-backspace is
not nil, all spaces"
  (interactive)
  (if (not nomadic-pict-hungry-backspace)
      (delete-backward-char 1)
    (let ((pos (point))
	  beginning-of-line)
      (beginning-of-line)
      (setq beginning (point))
      (goto-char pos)
      (if (= pos beginning)
	    (delete-backward-char 1)
	(skip-syntax-backward " " beginning)
	(if (= (point) pos)
	    (delete-backward-char 1)
	  (delete-region (point) pos))))))

(defun nomadic-pict-newline ()
  "Insert a new line and possibly indent the next one"
  (interactive)
  (newline)
  (if nomadic-pict-auto-indent
      (nomadic-pict-indent-line)))
  

(defun nomadic-pict-in-comment-p (pos)
  "Return t if pos is inside a comment, nil otherwise"
  (if (search-forward "{-" pos t) 
      (if (search-forward "-}" pos t)
	  (let ((save (point)))
	    (end-of-line)
	    (if (>= (point) pos) 
		(progn
		  (search-backward "{-")
		  (forward-char 2)
		  t)
	      (goto-char save)
	      (nomadic-pict-in-comment-p pos)))
	(end-of-line)
	(if (< (point) pos)
	    (progn
	      (search-backward "{-")
	      (forward-char 2)
	      t)
	  (goto-char pos)
	  (let (end)
	    (end-of-line)
	    (setq end (point))
	    (beginning-of-line)
	    (if (search-forward-regexp "^[ \\t]*{-" end t) 
		t
	      nil))))))

(defun nomadic-pict-skip-comment ()
  "Moves the point backwards, skipping comments and blank lines"
  (interactive)
  (let (end)
    (end-of-line)
    (setq end (point))
    (beginning-of-line)
    (if (search-forward "-}" end t)
	(progn
	  (search-backward "{-")
	  (skip-chars-backward " \t")
	  (if (= (point) (progn (beginning-of-line) (point)))
	      (progn 
		(if (= (point) (point-min))
		    nil
		  (backward-char)
		  (nomadic-pict-skip-comment)))))
      (if (not (search-forward-regexp "[^ \\t]" end t))
	  (progn
	    (previous-line 1)
	    (nomadic-pict-skip-comment))
	(beginning-of-line)))))

(defun nomadic-pict-matching-paren ()
  "Finds the open parenthesis that match the last one on the current line"
  (save-excursion
    (next-line 1)
    (end-of-line)
    (let ((stack '(0)) pos)
      (search-backward ")")
      (while (not (null stack))
	(search-backward-regexp "[()]")
	(if (= (char-after) ?\))
	    (setq stack (cons 0 stack))
	  (setq stack (cdr stack))))
      (beginning-of-line)
      (skip-syntax-forward " ")
      (setq pos (point))
      (beginning-of-line)
      (- pos (point)))))

(defun nomadic-pict-calc-indent-outside-comment (pos)
  (goto-char pos)
  (beginning-of-line)
  (if (= (point) (point-min)) ;; beginning of buffer => indent=0
      0
    (backward-char 1)
    (beginning-of-line)
    (let (end)
      (end-of-line)
      (setq end (point))
      (beginning-of-line)
      (nomadic-pict-skip-comment)
      (let ((bol (point)) 
	    last-indent
	    depth
	    lev
	    (stack nil))
	(skip-syntax-forward " ")
	(setq last-indent (- (point) bol))
	(end-of-line)
	(setq end (point))
	(beginning-of-line)
	(setq depth (nomadic-pict-parse-paren))
	;(debug)
	(cond
	 ((> (nth 2 depth) 0) ; un [ ouvert  la ligne prcdente
	  (end-of-line)
	  (setq end (point))
	  (beginning-of-line)
	  (setq bol (point))
	  (while (search-forward-regexp "\\(\\]\\|\\[\\)" end t)
	    (if (= (char-before) ?\[)
		(setq stack (cons (point) stack))
	      (setq stack (cdr stack))))
	  (- (car stack) bol))
	 ((< (nth 2 depth) 0) 
	  (setq stack (list (point)))
	  (backward-char)
	  (while (not (null stack))
	    (search-backward-regexp "\\(\\]\\|\\[\\)")
	    (if (= (char-after) ?\[)
		(setq stack (cdr stack))
	      (setq stack (cons (point) stack))))
	  (setq pos (point))
	  (beginning-of-line)
	  (setq bol (point))
	  (if (search-forward "[" pos t)
	      (- (point) bol)
	    (skip-syntax-forward " ")
	    (- (point) bol)))
	 ((save-excursion
	    (goto-char pos)
	    (end-of-line)
	    (setq end (point))
	    (beginning-of-line)
	    (search-forward-regexp "^[ \t]*)" end t))
	  (nomadic-pict-matching-paren))
	 ((save-excursion
	    (goto-char pos)
	    (beginning-of-line)
	    (skip-syntax-forward " ")
	    (and (< (point) (point-max))
		 (= (char-after) ?|)))
	  (if (save-excursion
		(nomadic-pict-skip-comment)
		(beginning-of-line)
		(skip-syntax-forward " ")
		(and (< (point) (point-max))
		     (= (char-after) ?|)))
	      (if (search-backward-regexp "[^ \t)]" bol t)
		  (+ last-indent (* nomadic-pict-indent-width (car depth)))
		last-indent)
	    (if (search-backward-regexp "[^ \t)]" bol t)
		(+ last-indent (* nomadic-pict-indent-width (car depth)) -2)
	      (- last-indent nomadic-pict-indent-width))))
	 ((and
	   (save-excursion
	     (goto-char pos)
	     (end-of-line)
	     (setq end (point))
	     (beginning-of-line)
	     (not (search-forward-regexp "^[ \t]*(" end t)))
	   (setq lev (nomadic-pict-if-then-else)))
	  (+ last-indent (* nomadic-pict-indent-width lev)))
	 ((search-backward-regexp "[^ \t)]" bol t)
	  (if (save-excursion
		(nomadic-pict-skip-comment)
		(beginning-of-line)
		(skip-syntax-forward " ")
		(and (< (point) (point-max))
		     (= (char-after) ?|)))
	      (+ last-indent (* nomadic-pict-indent-width (car depth)) nomadic-pict-indent-width)
	    (+ last-indent (* nomadic-pict-indent-width (car depth)))))
	 
	 (t
	  (if (save-excursion
		(nomadic-pict-skip-comment)
		(beginning-of-line)
		(skip-syntax-forward " ")
		(and (< (point) (point-max))	    
		     (= (char-after) ?|)))
	      (+ last-indent nomadic-pict-indent-width)
	    last-indent)))))))

(defun nomadic-pict-if-then-else (&optional rec)
  "This function computes the indentation for if-then-else blocks. 
Returns nil if we are not in an if-then-else block, otherwise the 
indent level to use. 
The optional parameter rec is used to see if this function is called 
recursively"
  (let (beg end)
    (save-excursion
;      (debug)
      (end-of-line)
      (setq end (point))
      (beginning-of-line)
      (cond
       ((and
	 (not rec)
	 (search-forward-regexp "[ \t]*else\\([ \t]\\|{-.*-}\\)*$" end t))
	1)
       ((and
	 (not rec)
	 (search-forward-regexp "[ \t]*if.*then\\([ \t]\\|{-.*-}\\)*$" end t))
	1)
       ((and
	 (not rec)
	 (search-forward-regexp "^[ \t]*(" end t))
	nil)
       (t
	(previous-line 1)
	(nomadic-pict-skip-comment)
	(end-of-line)
	(setq end (point))
	(beginning-of-line)
	(or (if (search-forward-regexp "[ \t]*if.*then\\([ \t]\\|{-.*-}\\)*$" end t)
		-1)
	    (if (search-forward-regexp "[ \t]*else\\([ \t]\\|{-.*-}\\)*$" end t)
		(progn
		  (search-backward-regexp "if.*then")
		  (nomadic-pict-skip-comment)
		  (- (or (nomadic-pict-if-then-else t) 0) 1)))))))))
	    
	    
	
(defun nomadic-pict-parse-paren ()
  (let ((par1 0)
	(par2 0)
	(par3 0)
	end)
    (end-of-line)
    (setq end (point))
    (beginning-of-line)
    (while (search-forward-regexp "(\\|)\\|\\[\\|\\]\\|{\\|}" end t)
      (cond
       ((= (char-before) ?\() (setq par1 (1+ par1)))
       ((= (char-before) ?\)) (setq par1 (1- par1)))
       ((= (char-before) ?{) (setq par2 (1+ par2)))
       ((= (char-before) ?}) (setq par2 (1- par2)))
       ((= (char-before) ?\[) (setq par3 (1+ par3)))
       ((= (char-before) ?\]) (setq par3 (1- par3)))
       (t (error "Shouldn't happen!"))))
    (list par1 par2 par3)))

(defun nomadic-pict-calc-indent-level ()
  "This function computes the number of spaces to put at the beginning 
of the current line"
  (let (pos beg depth)
    (setq pos (point))
    (save-restriction
      (widen)
      (save-excursion
	(save-excursion
	  (setq depth (nomadic-pict-parse-paren))
	  (beginning-of-line)
	  (setq beg (point)))
	(backward-char (min (- pos 1) 3000))
	(if (nomadic-pict-in-comment-p pos)
	    ;; le point est plac juste aprs le dbut du commentaire
	    (let (cur end)
	      (setq cur (point))
	      (end-of-line)
	      (setq end (point))
	      (beginning-of-line)
	      (if (or (= (point) beg) 
		      (save-excursion
			(goto-char pos)
			(end-of-line)
			(setq end (point))
			(beginning-of-line)
			(if (search-forward-regexp "^[ \\t]*-}[ \\t]*$" end t)
			    (progn 
			      (search-backward "{-")
			      (setq pos (point)))
			  nil)))
		  (nomadic-pict-calc-indent-outside-comment pos)
		
		(+ 1 (- cur (point)))))
	  ;; Pas dans un commentaire
	    (nomadic-pict-calc-indent-outside-comment pos))))))


(defun nomadic-pict-indent-line ()
  "Indent the current line"
  (interactive)
  (let (indent-width)
    (setq indent-width (nomadic-pict-calc-indent-level))
    (save-excursion 
      (let (beginning
	    end)
	(end-of-line)
	(setq end (point))
	(beginning-of-line)
	(setq beginning (point))
	(skip-syntax-forward " " end)
	(delete-region beginning (point))
	(indent-to indent-width)))
    (let ((pos (point)))
      (beginning-of-line)
      (if (< (- pos (point)) indent-width)
	  (forward-char indent-width)
	(goto-char pos)))))

;(cancel-debug-on-entry 'nomadic-pict-indent-line)
 (defconst nomadic-pict-keywords 
   (list 
    '("\\(\\^\\|!\\|\\?\\|/\\)?\\(Agent\\|Bool\\|Char\\|Int\\|Site\\|String\\|T\\(op\\|ype\\)\\)\\>" 0 'font-lock-type-face)
    (list (concat "\\<\\(" (regexp-opt '("agent" "and" "ccode" "def"
 					"dynamic" "else" "false" "if" 
 					"iflocal" "import" "inline" "in"
 					"migrate" "new" "now" "of" "program"
 					"rec" "run" "terminate" "then"
 					"timeout" "to" "toplevel" "true"
 					"type" "typecase" "val" "switch"
 					"wait" "where" "with")) "\\)\\>")
 	 1 'font-lock-keyword-face)
    '("{[a-zA-Z]*>" . 'font-lock-constant-face))
   "Keywords for syntax highlighting")



(defun nomadic-pict-mode ()
  "Major mode for editing Nomadic Pict programs. This mode supports 
indentation and syntax highlighting

\\{nomadic-pict-mode-map}"
  (interactive)
  (kill-all-local-variables)
  (setq major-mode 'nomadic-pict-mode)
  (setq mode-name "Nomadic Pict")
  (setq nomadic-pict-mode-syntax-table nil)
  (if nomadic-pict-mode-syntax-table 
      nil
    (setq nomadic-pict-mode-syntax-table (make-syntax-table))
    (modify-syntax-entry ?{ ".1" nomadic-pict-mode-syntax-table)
    (modify-syntax-entry ?} ".4" nomadic-pict-mode-syntax-table)
    (modify-syntax-entry ?- ". 23" nomadic-pict-mode-syntax-table)
    (modify-syntax-entry ?\\ "." nomadic-pict-mode-syntax-table)
    (modify-syntax-entry ?\' "'" nomadic-pict-mode-syntax-table)
    (modify-syntax-entry ?\" "\"" nomadic-pict-mode-syntax-table)
    (modify-syntax-entry ?\( "()" nomadic-pict-mode-syntax-table)
    (modify-syntax-entry ?\) ")(" nomadic-pict-mode-syntax-table)
    (modify-syntax-entry ?< "(>" nomadic-pict-mode-syntax-table)
    (modify-syntax-entry ?> ")<" nomadic-pict-mode-syntax-table)
    )
  (set-syntax-table nomadic-pict-mode-syntax-table)
  (use-local-map nomadic-pict-mode-map)
  (make-variable-buffer-local 'font-lock-defaults)
  (make-variable-buffer-local 'indent-line-function)
  (setq font-lock-defaults (list 'nomadic-pict-keywords nil nil nil nil))
  (setq indent-tabs-mode nil)
  (setq indent-line-function 'nomadic-pict-indent-line)
  (make-local-variable 'comment-start)
  (setq comment-start "{- ")
  (make-local-variable 'comment-end)
  (setq comment-end " -}")
  (make-local-variable 'comment-start-skip)
  (setq comment-start-skip "{-+ *")
  (make-local-variable 'comment-multi-line)
  (setq comment-multi-line t)
  (make-variable-buffer-local 'nomadic-pict-hungry-backspace)
  (make-variable-buffer-local 'nomadic-pict-auto-indent)
  (make-variable-buffer-local 'nomadic-pict-indent-width)
  (run-hooks 'nomadic-pict-mode-hook)
)

(provide 'nomadic-pict-mode)

