(*
 * Intermediate code.
 *)

let cc =
  Flags.createFlag "cc"
    "*Compile to C and native code" true  (* do not change this setting, 
                        otherwise Nomadic Pict libraries will not compile! *)

let exact =
  Flags.createFlag "exactCCode"
  "*Show exact representation of abstract C code" false

type var =
  Var.var

type kind = EXP | STATEMENT | NEXP | NSTATEMENT

type info = {
  alloc: int;       (* Number of words allocated by the C code *)
  const: bool;      (* True if the C code is a constant *)
  reader: bool;     (* True if the C code reads from updatable storage *)
  writer: bool;     (* True if the C code writes to updatable storage *)
  kind: kind;       (* Tells us what kind of expression this is *)
  code: string list (* The list of pieces of C code *)
}

type ty =
  INTEGER
| POINTER

type decl =
  BLOCK of var * (var * ty) list * var * var * int * code
| CONST of var * exp list
| BYTES of var * string
| TOPLEVEL of string
| FORWARD of var
| EXTERN of var

and code =
  SEQ of code * code
| ASSIGN of exp * exp
| IF of exp * code * code
| CPROC of info * exp list
| PROFILING of string
| NULL

and exp =
  INT of int
| VAR of var
| ADDR of var
| OFFSET of bool * exp * int
| DEREF of var * int
| TAG of var * int
| TAGS of var
| INDEX of var * exp
| CCODE of info * exp list

let printDecls os l =
  let printInt = function
      -2 -> output_string os "-2"
    | -1 -> output_string os "-1"
    |  0 -> output_char os '0'
    |  1 -> output_char os '1'
    |  2 -> output_char os '2'
    |  3 -> output_char os '3'
    |  4 -> output_char os '4'
    |  5 -> output_char os '5'
    |  6 -> output_char os '6'
    |  7 -> output_char os '7'
    |  8 -> output_char os '8'
    |  9 -> output_char os '9'
    |  i -> output_string os (string_of_int i)
  in

  let rec printString os s x =
    if x = String.length s then output_string os "\\000" else
    let c = String.get s x in
    if Char.code c < 32 || Char.code c > 126 then
      (output_char os '\\';
       output_byte os ((Char.code '0') + ((Char.code c) lsr 6));
       output_byte os ((Char.code '0') + (((Char.code c) lsr 3) land 7));
       output_byte os ((Char.code '0') + ((Char.code c) land 7));
       printString os s (x+1))
    else if c = '\\' then
      (output_string os "\\\\"; printString os s (x+1))
    else if c = '"' then
      (output_string os "\\\""; printString os s (x+1))
    else
      (output_char os c; printString os s (x+1))
  in

  let printLocals os = function
    (x,INTEGER) -> Var.print os x
  | (x,POINTER) -> output_string os "*"; Var.print os x
  in

  let rec printCode = function
    SEQ(c,d) ->
      printCode c; printCode d
  | ASSIGN(e,f) ->
      printExp os e; output_string os "=";
      printExp os f; output_string os ";\n"
  | IF(e,c,NULL) ->
      output_string os "if("; printExp os e; output_string os "){\n";
      printCode c; output_string os "}\n"
  | IF(e,c,d) ->
      output_string os "if("; printExp os e; output_string os "){\n";
      printCode c; output_string os "}else{\n";
      printCode d; output_string os "}\n"
  | CPROC(ci,l) ->
      printCCode ci l; output_char os '\n'
  | PROFILING(s) ->
      output_string os s
  | NULL ->
      if !exact then output_string os "/**/;\n" else ()

  and printExp os = function
    INT(i) ->
      printInt i
  | VAR(var) ->
      Var.print os var
  | ADDR(v) ->
      output_string os "(Val)&"; Var.print os v
  | INDEX(v,INT i) ->
      if i > 0 then
	(Var.print os v; output_char os '+'; printInt i)
      else if i < 0 then
	(Var.print os v; output_char os '-'; printInt (-i))
      else
	Var.print os v
  | INDEX(v,e) ->
      Var.print os v; output_char os '+'; printExp os e
  | TAG(v,0) ->
      output_string os "TAG("; Var.print os v; output_char os ')'
  | TAG(v,i) ->
      output_string os "TAG("; Var.print os v; output_char os '+';
      printInt i; output_char os ')'
  | TAGS(v) ->
      output_string os "TAG(&"; Var.print os v; output_char os ')'
  | DEREF(v,i) ->
      Var.print os v; output_char os '['; printInt i; output_char os ']'
  | OFFSET(m,e,i) ->
      if !exact && m then output_string os "/*mutable*/";
      output_string os "OFFSET("; printExp os e; output_char os ',';
      printInt i; output_char os ')'
  | CCODE(ci,l) ->
      printCCode ci l

  and printCCode ci l =
    let rec print sl al =
      match sl with [] -> () | (s :: sl) ->
      (match al with
         [] -> output_string os s
       | (e :: el) -> output_string os s; printExp os e; print sl el)
  in
    print ci.code l

  and printDecl = function
    BLOCK(var,l,free,endq,i,c) ->
      output_string os "void "; Var.print os var;
      (match l with [] -> output_string os "(void){\n" | _ ->
       Misc.printList "(void){\nVal " ";\n" "," printLocals os l);
      if i > 0 then
        (Var.print os free; output_string os "=Free;";
         Var.print os endq; output_string os "=EndQ;\n";
	 output_string os "if("; Var.print os free;
	 output_string os "+"; printInt i;
	 output_string os ">"; Var.print os endq;
	 output_string os "){Gc("; printInt i;
	 output_string os ");";
	 Var.print os free; output_string os "=Free;";
	 Var.print os endq; output_string os "=EndQ;}\n");
      printCode c; output_string os "}\n"
  | TOPLEVEL(s) ->
      output_string os s; output_char os '\n'
  | CONST(x,l) ->
      output_string os "const Val "; Var.print os x;
      output_string os "["; printInt(List.length l);
      Misc.printList "]={" "};\n" "," printExp os l
  | BYTES(x,s) ->
      let sz = String.length s + 1 in
      output_string os "const struct{Val h;char c["; printInt sz;
      output_string os "];}"; Var.print os x;
      output_string os "={STRING("; printInt sz; output_string os "),\"";
      printString os s 0; output_string os "\"};\n"
  | FORWARD(x) ->
      output_string os "void "; Var.print os x; output_string os "(void);\n"
  | EXTERN(x) ->
      output_string os "extern const Val "; Var.print os x;
      output_string os "[];\n"
in
  List.iter printDecl l

let formatDecls l = Format.print_flush(); printDecls stdout l

let complexOutput = {
  alloc = 9; const = false; reader = false; writer = true;
  kind = STATEMENT; code = ["ComplexOutput("; ","; ");"]
}

let complexInput = {
  alloc = 9; const = false; reader = false; writer = true;
  kind = STATEMENT; code = ["ComplexInput("; ","; ");"]
}

let sizeOf = {
  alloc = 0; const = false; reader = false; writer = false;
  kind = EXP; code = ["SIZE("; ")"]
}

let free = {
  alloc = 0; const = false; reader = true; writer = false;
  kind = EXP; code = ["Free"]
}

let startq = {
  alloc = 0; const = false; reader = true; writer = false;
  kind = EXP; code = ["StartQ"]
}

let endq = {
  alloc = 0; const = false; reader = true; writer = false;
  kind = EXP; code = ["EndQ"]
}

(*
 * Printing CAML representation of ccode info
 *)

let caml_format_string s =
  Format.print_string "\"";
  Format.print_string s;
  Format.print_string "\""

let caml_formatInfo ci =
  Format.open_hbox(); 
  Format.print_char '{';
  Format.print_string "Ccode.alloc="; Format.print_int ci.alloc; 
  Format.print_char ';'; Format.print_newline();
  Format.print_string "Ccode.const="; Format.print_bool ci.const; 
  Format.print_char ';'; Format.print_newline();
  Format.print_string "Ccode.reader="; Format.print_bool ci.reader; 
  Format.print_char ';'; Format.print_newline();
  Format.print_string "Ccode.writer=";  Format.print_bool ci.writer; 
  Format.print_char ';'; Format.print_newline();
  Format.print_string "Ccode.kind=";
  (match ci.kind with
    EXP        -> Format.print_string "Ccode.EXP"
  | STATEMENT  -> Format.print_string "Ccode.STATEMENT"
  | NEXP       -> Format.print_string "Ccode.NEXP"
  | NSTATEMENT -> Format.print_string "Ccode.NSTATEMENT"
); 
  Format.print_char ';'; Format.print_newline();
  Format.print_string "Ccode.code="; 
  Misc.formatList "[" "]" "; " caml_format_string ci.code;
  Format.print_char '}';
  Format.close_box()

