(*
 * Toplevel compilation loop.
 *)

open Genlex
open Unix

let defaultConfigFile = "config"

let nomad_configFile = ref (Some defaultConfigFile)
let nomad_metaFile   = ref (None : string option)
let nomad_tempFile   = ref (None : string option)

let sep =
  Flags.createFlag "sep" "Separately compile this module" false

let executeAfterCompiling =
  Flags.createFlag "execute"
    "Execute after compiling (autoset when no output file given)" false 

let lib = Flags.createFlag "lib" 
  "Automatically import the standard library lib.src" true

let debuggingInfo =
  Flags.createFlag "debuggingInfo"
  "Include debugging information in the compiled C code" false

let optC = Flags.createFlag "optC" "Optimise the generated C code" true

let keep = Flags.createFlag "keep" "Keep temporary files" false

let globalOpt = Flags.createFlag "globalOpt" "Do global optimisation" false

let showCamlTyHigh =			                (* Nomadic Pict *)
  Flags.createFlag "showCamlTyHigh"
    "@Show CAML representation of high-level code with type annotations" false

let showCamlHigh =					(* Nomadic Pict *)
  Flags.createFlag "showCamlHigh"
    "@Show CAML representation of high-level code" false
    
let showCamlSimplified =				(* Nomadic Pict *)
  Flags.createFlag "showCamlSimplified"
    "@Show CAML representation of intermediate code before joining imports" false

(*
 * Search path stuff.
 *)

let dirs = try ["."; Sys.getenv "PICTLIB"] with Not_found -> ["."];;
let searchPath = ref dirs;;

let _ = Flags.createCmd "setPath"
  "Set the import search path" (fun _ l -> searchPath := l)

let _ = Flags.createCmd "appendToPath"
  "Append some directories to the import search path"
  (fun _ l -> searchPath := l @ !searchPath)

exception BadFile of string

let find name suffix =
  let rec trypath = function
      [] -> raise (BadFile
	 ("Could not find " ^ name ^ suffix ^ " in the directories " ^
	  (List.fold_left (fun s t -> s ^ " " ^ t) "" !searchPath)))
    | (p::t) ->
      try let full = Filename.concat p name in (open_in (full ^ suffix),full)
      with Sys_error _ -> trypath t
in
  if String.length name = 0 then
    raise (BadFile "Null file name")
(* ocaml-1.05: else if Filename.is_absolute name then *)
  else if not (Filename.is_implicit name) then
    try (open_in (name ^ suffix),name) with Sys_error m -> 
    raise (BadFile ("Could not open " ^ name ^ suffix ^ " (" ^ m ^ ")"))
  else 
    trypath !searchPath

let findFile fi name suffix =
  try find name suffix with BadFile msg -> Error.errAt fi msg

let findDep fi name suffix =
  try let (pi,full) = find name suffix in close_in pi; Some full
  with BadFile msg -> Error.warningAt fi msg; None

(*
 * Parsing.
 *)

let parseFile inFile =
  let pi =
    try open_in inFile with Sys_error m -> 
    Error.err ("Could not open " ^ inFile ^ " (" ^ m ^ ")")
  in let lexbuf = Lexer.create inFile pi
  in let result =
    try Parser.toplevel Lexer.main lexbuf with Parsing.Parse_error -> 
    Error.errAt (Lexer.info lexbuf) "Parse error"
in
  Parsing.clear_parser(); close_in pi; result

(*
 * Parsing config file.
 *)

let check  message r s =
  match !r with
    None -> r := Some s
  | Some _ -> Error.err message

let checkwarn message r s =
  match !r with
    None -> r := Some s
  | Some _ -> Error.warning message; r := Some s 

let checkignore message r s =
  match !r with
    None -> r := Some s
  | Some _ -> Error.warning message

let shortname n =
let rec finddot s i = 
  if i < String.length s then
    if s.[i] = '.' then i
    else finddot s (i+1)
  else i
in let pos = finddot n 0 in String.sub n 0 pos 

let rec parse_line = parser
  [< 'String addr; 'Int port_no >] ->
      let dotted_addr = try 
        if (addr = "localhost") 
          then string_of_inet_addr((gethostbyname(gethostname())).h_addr_list.(0))
          else string_of_inet_addr((gethostbyname(addr)).h_addr_list.(0))
      with Not_found -> Error.errAt Error.progName ("Unknown IP address " ^ addr)
      in
(*
      if (dotted_addr = !Nrun.my_host_addr) then
        checkwarn "Another port binding found" Nrun.serverPortOpt port_no;
*)
      if (dotted_addr = !Nrun.my_host_addr) then
        checkignore ("Port binding \""^addr^"\" "^(string_of_int port_no)^" in config ignored") 
          Nrun.serverPortOpt port_no;

      Nrun.sites_l := (dotted_addr,port_no)::!Nrun.sites_l;
      true
  | [< 'Kwd "METAFILE"; 'Kwd "="; 'String filename >] ->
      check "Encodings file name already specified" nomad_metaFile filename;
      true
  | [< 'Kwd "#"; r=parse_comment >] -> r
  | [< eof >] -> 
      Nrun.sites_l  := List.rev !Nrun.sites_l;
      Nrun.no_sites := List.length !Nrun.sites_l;
      false
and parse_comment = parser
    [< 'String addr; 'Int port_no >] -> true
  | [< 'Kwd "METAFILE"; 'Kwd "="; 'String filename >] -> true

let parseConfig cFile =
  match (try Some (open_in cFile) with Sys_error m -> 
    if cFile <> defaultConfigFile then
      Error.warning ("Could not open " ^ cFile ^ " (" ^ m ^ ")");
    None) 
  with
    (Some pi) ->
      let lexer = make_lexer ["METAFILE";"=";"END";"#"] in 
      let lexerfun = lexer (Stream.of_channel pi) in
      let rec read_file() = 
    	if parse_line lexerfun then read_file()
      in
      let () = try read_file() with 
    	_ -> Error.err ("Parse error in a file "^cFile)
      in
      close_in pi
  | None -> ()

(*
 * Add encodings to input file.
 *)

let addSys inFile =

  Stats.mark "Reading config file";
  (match !nomad_configFile with
    None -> Error.warning "A config file not specified"
  | Some filename -> parseConfig filename);

  Stats.mark "Enclosing encodings";
  match !nomad_metaFile with
    None ->
 (*     Error.warning ("Encodings file name not found."); *)
      inFile
  | Some metafile ->
      if not (Sys.file_exists metafile) then
    	Error.err ("Could not find " ^ metafile ^ " (a file with meta-definitions)")
      else
    	let tempfile = inFile ^ "_NOMAD" ^ string_of_int(Unix.getpid()) ^ ".pi" in
    	if (Sys.file_exists tempfile) then 
	  Error.err ("A randomly generated " ^ tempfile ^ " already used as file name. Compile again.")
    	else 
	  begin
	    let cmd = 
	      "cp "^inFile^" "^tempfile^"; "^      (* copy input file to temporal file *)
              "cat "^metafile^" >> "^tempfile  (* enclose encodings *)
	    in
(*	    let exitcode = *)
	    let _ =
	      try Sys.command cmd with Sys_error m ->
	      	Error.err ("System error while executing: "^cmd^" ("^m^")")
	    in
	    check "A temporary file already specified" nomad_tempFile tempfile;
	    tempfile
    	  end

(*
 * Dependency generator stuff.
 *)

let dep = Flags.createFlag "dep" 
  "Generate dependencies (disables all compilation)" false

let printDep os (fi,name) =
  match findDep fi name ".pi" with
    Some full -> output_string os full; output_string os ".px "
  | None -> ()

let fileDeps os inFile =
  Stats.progress ("checking dependencies for " ^ inFile);
  let imports =
    try
      let pi = open_in inFile in
      let lexbuf = Lexer.create inFile pi in
      let result =
        try
	  (Parser.toplevel Lexer.main lexbuf).Syntax.imports
        with Parsing.Parse_error -> 
          Error.warningAt (Lexer.info lexbuf) "Parse error"; []
      in Parsing.clear_parser(); close_in pi; result 
    with Sys_error m -> 
      Error.warning ("Could not open " ^ inFile ^ " (" ^ m ^ ")"); []
  in
  let libfile = if !Ccode.cc then "lib" else "nlib" in
  let imports =
    if !lib then (Error.unknown,libfile) :: imports else imports
  in
    if Filename.check_suffix inFile ".pi" then
      begin
        let base = Filename.chop_suffix inFile ".pi" in
	output_string os base; output_string os " ";
	output_string os base; output_string os ".px : ";
	output_string os base; output_string os ".pi ";
	List.iter (printDep os) imports; output_string os "\n"
      end
    else
      Error.errAt Error.progName "Unknown file suffix"

(*
 * Importing separate compiled (or parsed) modules.
 *)

module Set = Misc.StringSet

let rec importAll inProgress ok files imports = function
  [] -> (ok,files,imports)
| ((fi,inFile) :: rest) ->
  if Set.mem inFile ok then
    importAll inProgress ok files imports rest
  else if Set.mem inFile inProgress then
    Error.errAt fi "Cyclic import dependency"
  else
    begin
      Stats.progress ("importing " ^ inFile);
      let (pi,full) = findFile fi inFile ".px" in
      Stats.progress (" loading " ^ full);
      let (deps,prog) = input_value pi in close_in pi;
      let (ok,files,imports) =
	importAll (Set.add inFile inProgress) ok files imports deps
      in
	importAll inProgress (Set.add inFile ok)
	(full :: files) (prog :: imports) rest
    end

let parseArgs () =
  let inFiles = ref ([] : string list) in
  let outfile = ref (None : string option) in
  let ccFlags = ref "-Wall" in
  Arg.parse [
    "-set", Arg.String (Flags.setFlag Error.unknown true), "Set a flag";
    "-reset", Arg.String (Flags.setFlag Error.unknown false), "Reset a flag";
    "-do", Arg.String (fun c -> Flags.doCmd Error.unknown c []),
    "Execute an internal command";
    "-I", Arg.String (fun f -> Flags.doCmd Error.unknown "appendToPath" [f]),
    "Append a directory to the search path";
    "-o", Arg.String (check "Output file already specified" outfile),
    "Specify the output file";
    "-cc", Arg.String (fun s -> ccFlags := !ccFlags ^ " " ^ s),
    "Specify an argument for the C compiler";
    "-f", Arg.String (fun s -> nomad_configFile := (Some s)),
    "Specify the config file";
    "-p", Arg.Int (fun p -> check "Port number already specified in the config file" Nrun.serverPortOpt p),
    "Specify port number (option)";
    "-e", Arg.String (check "File with encodings already specified" nomad_metaFile),
    "Specify the name of file with encodings (option)";
    "-trader", Arg.String (fun s -> 
      let s = try 
        if (s = "localhost") 
          then string_of_inet_addr((gethostbyname(gethostname())).h_addr_list.(0))
          else string_of_inet_addr((gethostbyname(s)).h_addr_list.(0))
      with Not_found -> Error.errAt Error.progName ("Unknown IP address " ^ s)
      in
        Nrun.trader_hostname := (Some s)),
    "Specify trader's host address";
    "-tport",  Arg.Int (fun n -> Nrun.trader_port := n), 
    "Specify trader's port number";
    "-recover", Arg.String (check "Recovery file already specified" Nrun.recoveryFileOpt),
    "Specify the recovery file (omit suffix)"
  ] (fun s -> inFiles := s :: !inFiles) "";
  match !outfile with
    None ->
      executeAfterCompiling := true;
      (List.rev !inFiles, "./a.out", !ccFlags)
  | Some oFile -> (List.rev !inFiles, oFile, !ccFlags)

let gcc link ccFlags outFile files decls =
  let tempFile = ".PICT" ^ string_of_int(Unix.getpid()) ^ ".c" in
  if (Sys.file_exists tempFile) then 
    Error.err ("A randomly generated " ^ tempFile ^ " already used as file name. Compile again.")
  else begin
    let objStream = try open_out tempFile with Sys_error s ->
      Error.err ("Couldn't open " ^ tempFile ^ " for output" ^ "(" ^ s ^ ")")
    in
    (try
      Stats.mark "Printing C code";
      Ccode.printDecls objStream decls;
      close_out objStream
    with Error.Exit x -> close_out objStream; raise (Error.Exit x));
    Stats.mark "Compiling C code";
    let pictrts = Sys.getenv "PICTRTS" in
    let cmd =
      Sys.getenv "GCC" ^
      " " ^ ccFlags ^
 (* " -pass-exit-codes " ^ for debugging *)
      (if link then "" else " -c") ^
      (if !debuggingInfo then " -g" else "") ^
      (if link && not !debuggingInfo then " -Xlinker -s" else "") ^
      (if !optC && not !debuggingInfo then " -O -fomit-frame-pointer" else "") ^
      " -o " ^ outFile ^
      " -I " ^ pictrts ^
      " -idirafter " ^ Sys.getenv "X11INCLUDE" ^
      " " ^ tempFile ^
      (if link then
       	(List.fold_left (fun s f -> s ^ " " ^ f ^ ".px.o") "" files) ^
       	(if !debuggingInfo then " " ^ pictrts ^ "/pictLibG.a"
        else if !Codegen.profiling then " " ^ pictrts ^ "/pictLibP.a"
	else " " ^ pictrts ^ "/pictLib.a") ^
       	" -lm " ^
       	" -L" ^ Sys.getenv "X11LIB" ^ " -lX11 " ^
       	" " ^ Sys.getenv "X11EXTRA"
      else
       	"")
    in
    Stats.progress cmd;
    if Sys.command cmd <> 0 then Error.err "C compilation failed";
    if !keep then
      Error.warning ("Temporary file " ^ tempFile ^ " not removed")
    else
      Sys.remove tempFile
  end

let runit executable =
  Stats.mark "Executing";
  let cmd = executable in
  if Sys.command cmd <> 0
    then Error.err "Nomadic Pict program exited with nonzero status"

let rec pluginCompilations level imports prog = 
  Stats.mark "Pre-Typechecking";
  let prog = Precheck.precheck imports prog in
  Stats.mark "Plugging-in encodings";
  (* print_endline ("Translation pass " ^ (string_of_int level) ^ ".."); *)
  let prog = Plugin.encodings imports prog in
  if !showCamlTyHigh then
    begin
      print_endline ("The encodings level "^(string_of_int level));
      print_endline  "___________________________";
      Tyhigh.caml_format prog
    end;
  if !Enco.anotherPass then 
    begin
      Enco.anotherPass := false;
      pluginCompilations (level + 1) imports prog
    end
  else
    prog

let main () =
  Stats.mark "Checking args";
  let (inFiles,outFile,ccFlags) = parseArgs() in
  if !dep then
    let os = try open_out outFile with Sys_error s ->
      Error.errAt Error.progName
      	("Couldn't open " ^ outFile ^ " for output (" ^ s ^ ")")
    in
    (List.iter (fileDeps os) inFiles; close_out os)
  else
    begin
      let inFile =
	match inFiles with
	  [inFile] -> 
      	    if !Ccode.cc then
	      inFile
	    else
	      addSys inFile
        | _ ->
	    if !Ccode.cc then Error.err "You must specify exactly one input file"
	    else
	      let inFile = 
		match findDep (Error.unknown) "bootstrap" ".pi" with
		  Some full -> full ^ ".pi"
	      	| None -> Error.err "Bootstrapping file not found. Specify the input file."
	      in addSys inFile
      in
      Stats.mark "Parsing";
      let prog = parseFile inFile in
      let libfile = if !Ccode.cc then "lib" else "nlib" in
      let deps =
	if !lib then (Error.unknown,libfile) :: prog.Syntax.imports 
	else prog.Syntax.imports
      in
      Stats.mark "Importing";
      let (files,imports) =
	let (_,f,i) = importAll Set.empty Set.empty [] [] deps in
	(List.rev f,List.rev i)
      in
      if Filename.check_suffix inFile ".pi" then
	Var.setModuleName(Filename.chop_suffix inFile ".pi")
      else
	Var.setModuleName inFile;
      if !showCamlTyHigh then
	(print_endline "The program high-level code (O'Caml format):";
	 print_endline "********************************************";
	 Tyhigh.caml_format prog);
      if !sep then
	begin
          Stats.mark "Typechecking";
          let prog = Check.check imports prog in	  
	  Stats.mark "Simplification";
	  let prog = Simplify.simplify imports prog in
	  Stats.mark "Optimisation";
	  let prog = Optimise.optimise imports prog in
	  let px = try open_out outFile with Sys_error s ->
	    Error.errAt Error.progName
	      ("Couldn't open " ^ outFile ^ " for output (" ^ s ^ ")")
	  in
	  Stats.mark "Exporting";
	  output_value px (deps,prog); close_out px;
	  if !Ccode.cc then
            begin
	      Stats.mark "C code generation";
	      let decls = Codegen.constants imports prog in
	      gcc false ccFlags (outFile ^ ".o") files decls
	    end
	end
      else
	begin
	  Stats.mark "Compilations Loop";
	  let prog = pluginCompilations 0 imports prog in
          Stats.mark "Typechecking";
          (* print_endline "Type checking"; *)
          let prog = Check.check imports prog in
          if !showCamlHigh then
            (print_endline "Typechecked code with a typing context (O'Caml format)";
	     print_endline "######################################################";
	     High.caml_format prog;
	     print_endline "The typing context";
	     Types.printC prog.Simplify.types;
 	     print_endline "The typing context (O'Caml format)";
	     Types.caml_formatC prog.Simplify.types Inter.caml_formatAtom);
	  Stats.mark "Simplification";
	  let prog = Simplify.simplify imports prog in
	  if !showCamlSimplified then
	    (print_endline "Intermediate code (O'Caml format)";
 	     print_endline "#################################";
	     Inter.caml_format (fun _ -> ()) prog);
	  Stats.mark "Joining imports";
	  let prog = Inter.joinImports imports prog in
	  Stats.mark "Optimisation";
	  let prog = Optimise.optimise imports prog in
	  if !Ccode.cc then
	    (Stats.mark "C code generation";
	     let decls = Codegen.trans imports prog in
	     gcc true ccFlags outFile files decls;
             if !executeAfterCompiling then runit outFile)
	  else
  	    begin
	      (match !nomad_tempFile with
	      	None -> ()
	      | Some filename -> 
		  if !keep then
		    Error.warning ("Temporary file " ^ filename ^ " not removed")
	    	  else
		    Sys.remove filename);
	      if !executeAfterCompiling then
	       	(Stats.mark "Execute nomadic bytecode";
		 (* print_endline "Starting the Nomadic Pict virtual machine"; *)
		 Nrun.execute imports prog)
	      else
	      	(if (Sys.file_exists outFile) then
		  Error.err ("A file " ^ outFile ^ " already exists (hmm...do we need be informed?)")
		else
		  let objStream = try open_out outFile with Sys_error s ->
		    Error.err ("Couldn't open " ^ outFile ^ " for output" ^ "(" ^ s ^ ")")
	      	  in
	      	  (try
		    Stats.mark "Printing nomadic bytecode";
		    (* Bcode.printBytecode objStream imports prog; *)
		    output_string objStream "ALA MA KOTA.";
		    close_out objStream;
		    print_endline ("Bytecode saved to file " ^ outFile)
	      	  with Error.Exit x -> close_out objStream; raise (Error.Exit x)))
	    end
    	end
    end

let () = Format.set_max_boxes 1000
let res = Printexc.catch (fun () -> try main();0 with Error.Exit x -> x) ()
let () = Stats.summary()
let () = Format.print_flush()
let () = exit res
