(*
 * Utilities of the encodings mechanism. 
 *)

(*

  TO DO:

  getId: write a recursive function which simplifies values
  to get x (then {{ }} is not needed)

*)

let anotherPass = ref (false : bool)

type id = Var.var

module Map = Var.Map

module IdMap =
  Misc.Make(struct
    type t = id
    let compare = compare
  end)

type proc =
  PRL of Error.info * proc * proc
| IF of Error.info * value * proc * proc
| INPUT of Error.info * value * pat * proc
| INPA of Error.info * value * pat * proc
| OUTPUT of Error.info * value * value
| OUTA of Error.info * value * value
| LDOUTA of Error.info * value * value * value
| TRYHERE of Error.info * value * value * proc * proc
| REP_INPUT of Error.info * value * pat *  proc
| TIMED_INPUT of Error.info * value * value * pat *  proc * proc
| REP_TIMED_INPUT of Error.info * value * value * pat *  proc * proc
| LI_SEND of Error.info * value * value * value
| REMOTE_SEND of Error.info * value * value * value * value
| LOCAL_SEND of Error.info * value * value * value
| IFLOCAL of Error.info * value * value * value * proc * proc
| LET of Error.info * dec * proc
| SKIP of Error.info
| TERMINATE of Error.info
| TYPECASE of Error.info * value * (pat * proc) list * proc
| CASE of Error.info * ty option * value * (string * pat * proc) list
| ENCODE_PROC of Error.info * id * id

and pat =
  RECORDp of Error.info * fieldPat list
| ASp of Error.info * id * ty option * pat
| VARp of Error.info * id * ty option
| WILDp of Error.info * ty option
| RECp of Error.info * ty option * pat
| ENCODE_PAT of Error.info * id

and fieldPat =
  ANONp of Error.info * pat
| LABELEDp of Error.info * string * pat
| TYp of Error.info * id * constr

and dec =
  DEF of Error.info * (Error.info * id * pat * proc) list
| INLINE of Error.info * id * pat * proc
| NEW of Error.info * id * ty
| VAL of Error.info * pat * value
| SEQ of Error.info * dec * dec
| RUN of Error.info * proc
| NEWLOC of Error.info * id
| CREATE of Error.info * (Error.info * id * ty * proc) list
| CREATE_AGENT of Error.info * (Error.info * id * proc) list
| AT of Error.info * id
| MIGRATE of Error.info * value
| MACRO of Error.info * string * value
| MACROLET of Error.info * string * id * value
| INLINE_DEC of Error.info * dec
| INLINE_VAL of Error.info * value
| DEFER_DEC of Error.info * Syntax.dec
| ENCODE_DEC of Error.info * id * id

and ty =
  TOP of Error.info
| TVAR of Error.info * id
| TABS of Error.info * (id * Kind.kind) list * ty
| REC of Error.info * id * Kind.kind * ty
| TAPP of Error.info * ty * ty list
| RECORD of Error.info * fieldTy list
| WITH of Error.info * ty * fieldTy list
| WHERE of Error.info * ty * fieldTy list
| CHAN of Error.info * ty
| OCHAN of Error.info * ty
| ICHAN of Error.info * ty
| RCHAN of Error.info * ty
| STRING of Error.info
| BOOL of Error.info
| CHAR of Error.info
| INT of Error.info
| SITE of Error.info
| LOC of Error.info
| AGENTT of Error.info * ty
| AGENT of Error.info
| DYN of Error.info
| VART of Error.info * (string * ty) list
| MAP of Error.info * ty * ty
| ENCODE_TYPE of Error.info * id

and constr =
  LT of ty
| EQ of ty

and fieldTy =
  ANON of Error.info * ty
| LABELED of Error.info * string * ty
| TYP of Error.info * id * constr

and value =
  CCODEv of Error.info * Ccode.info * fieldVal list
| APPv of Error.info * ty option * value * fieldVal list
| RECORDv of Error.info * fieldVal list
| WITHv of Error.info * ty option * value * fieldVal list
| WHEREv of Error.info * ty option * value * fieldVal list
| STRINGv of Error.info * string
| SITEv of Error.info * string * int
| SELECTv of Error.info * value * string
| VARv of Error.info * id
| BOOLv of Error.info * bool
| CHARv of Error.info * char
| INTv of Error.info * int
| LETv of Error.info * dec * value
| RECv of Error.info * value * ty option
| IFv of Error.info * ty option * value * value * value
| ABSv of Error.info * pat * proc
| DYNAMICv of Error.info * value * ty option
| TYPECASEv of Error.info * ty option * value * (pat * value) list * value
| VARIANTv of Error.info * string * value
| CASEv of Error.info * ty option * value * (string * pat * value) list
| ENCODE_VALUE of Error.info * id

and fieldVal =
  ANONv of Error.info * value
| LABELEDv of Error.info * string * value
| TYv of Error.info * ty

and enco =
  DEF_PROC of id list * proc
| DEF_DEC of id list * dec
| DEF_TYPE of id list * ty
| DEF_DEF of Error.info * id *id * pat * proc

(* 
 * The types of used in the language translator. 
 *)

type f =
    NAME of id
  | PROC of proc
  | PAT of pat
  | DEC of dec
  | TY of ty
  | TYOPT of ty option
  | VALUE of value
  | LIST of f list
  | TUPLE of f * f
  | TRIPLE of f * f * f

	
type e = f Map.t 
let enull = Map.empty

let fresh n = 
  let x = Var.fresh() in 
  Var.createVar ((Var.str_of x)^n) (Var.int_of x)

let str_of a = Var.str_of a           (* Return string-part of id *)
let int_of a = Var.int_of a           (* Return integer-part of id *)
let createId s i = Var.createVar s i  (* Create value of abstract type id *)

let name a =
  let n = str_of a in
  let base = Var.getModuleName() in
  try 
    if (String.sub n 0 (String.length base) <> base) 
    then n           (* return name defined in imported files *)
    else Var.name a  (* return composite name *)
  with Invalid_argument s -> n 

let format a = Var.format a
let caml_format a = Var.format a


(* 
 * The translation context: types and definitions of the compositional translation. 
 *)

type 'a ec =
    {
     types: Types.tag Map.t;           (* Maps type variables to unique tags *)
     constr: Types.constr Types.Map.t; (* Maps tags to constraints *)
     vars: (Types.t * 'a) IdMap.t;     (* Maps identifiers to types, plus an 'a *)
     
     tvid: id Types.Map.t;             (* Maps tags to identifiers *)
  
     forms: enco Misc.StringMap.t;     (* Definitions of syntactic forms *)
     funs: enco Misc.StringMap.t       (* Definitions of library and user-defined functions *)
   }

type toplevel =
    {imports : (Error.info * string) list;
     static : string list;
     includes : string list;
     dec : dec;
     ec : Inter.atom ec}
      
let empty =
  {types=Map.empty; constr=Types.Map.empty; vars=IdMap.empty; 
   tvid=Types.Map.empty; 
   forms=Misc.StringMap.empty; funs=Misc.StringMap.empty}
    
let abind ec n tag =
  {types=Map.add n tag ec.types; constr=ec.constr; vars=ec.vars; 
   tvid=ec.tvid;
   forms=ec.forms; funs=ec.funs}

let constr ec tag constr =
  {types=ec.types; constr=Types.Map.add tag constr ec.constr; vars=ec.vars; 
   tvid=ec.tvid;
   forms=ec.forms; funs=ec.funs}

let xbind n t x ec =
  {types=ec.types; constr=ec.constr; vars=IdMap.add n (t,x) ec.vars; 
   tvid=ec.tvid;
   forms=ec.forms; funs=ec.funs}

let ibind ec tag id =
  {types=ec.types; constr=ec.constr; vars=ec.vars; 
   tvid=Types.Map.add tag id ec.tvid;
   forms=ec.forms; funs=ec.funs}

let ebind s def ec =
  {types=ec.types; constr=ec.constr; vars=ec.vars;
   tvid=ec.tvid;
   forms=Misc.StringMap.add s def ec.forms; funs=ec.funs}
    
let fbind n f ec =
  {types=ec.types; constr=ec.constr; vars=ec.vars;
   tvid=ec.tvid;
   forms=ec.forms; funs=Misc.StringMap.add n f ec.funs}

(* 
 * Iterate over value bindings in the translation context. 
 *)

let lookupTVId fi tag ec =
  try
    Types.Map.find tag ec.tvid
  with Not_found ->
    (* Error.warningAt fi ("Enco.lookupTVId: Tag " ^ (Types.name tag) ^ " not found"); *) 
    createId (Types.str_of tag) (Types.int_of tag)

let lookupTVTag fi nid ec =
  try
    Map.find nid ec.types
  with Not_found ->
    Error.bugAt fi ("Enco.lookupTVTag: Id " ^ (name nid) ^ " not found")

(* 
 * Bind formal parameters to actual values. 
 *)

let rec bind fi l el e =
  (match (l,el) with
    ([],[]) -> e
  | (l, []) -> Error.errAt fi "Encoding definition has extra fields" (* ; Enco.mkUnit? *)
  | ([],el) -> Error.errAt fi "Encoding definition does not have enough fields" (* ; Enco.mkUnit? *)
  | (x::l,v::el)   -> bind fi l el (Map.add x v e))

(* 
 * Modify the translation definition to compute the actual parameters first 
 *)

let modifyDec fi v def =
  let x = Var.fresh() in
  (x, SEQ(fi, VAL(fi, VARp(fi,x,None),v), def))

let modifyProc fi v def =
  let x = Var.fresh() in
  (x, LET(fi, VAL(fi, VARp(fi,x,None),v), def))

(* For debugging:

  let f fi = function
    CCODEv(_) -> Error.bugAt fi "ccd"
  | APPv(_) -> Error.bugAt fi "app"
  | RECORDv(_) -> Error.bugAt fi "record"
  | WITHv(_) -> Error.bugAt fi "with"
  | WHEREv(_) -> Error.bugAt fi "where"
  | STRINGv(_) -> Error.bugAt fi "string"
  | SITEv(_) -> Error.bugAt fi "site"
  | SELECTv(_) -> Error.bugAt fi "sel"
  | VARv(_) -> Error.bugAt fi "var"
  | BOOLv(_) -> Error.bugAt fi "bool"
  | CHARv(_) -> Error.bugAt fi "char"
  | INTv(_) -> Error.bugAt fi "intv"
  | LETv(_) -> Error.bugAt fi "letv"
  | RECv(_) -> Error.bugAt fi "recv"
  | IFv(_) -> Error.bugAt fi "ifv"
  | ABSv(_) -> Error.bugAt fi "abs"
  | DYNAMICv(_) -> Error.bugAt fi "dynamic"
  | TYPECASEv(_) -> Error.bugAt fi "typecase"
  | VARIANTv(_) -> Error.bugAt fi "variant"
  | CASEv(_) -> Error.bugAt fi "casev"
  | ENCODE_VALUE(_) -> Error.bugAt fi "encod"     
*)

let getId fi x e =
  try match Map.find x e with
    NAME(x) -> x
  | VALUE(VARv(fi,x)) -> x
  | _ -> Error.errAt fi ("Expected {{"^name x^"}} but found "^name x)
  with Not_found -> (* Error.warningAt fi ("Enco.getId: Name " ^ (name x) ^ " not encoded"); *) 
    x

let getName fi x e = name (getId fi x e)
      
let getDec fi x e =
  try
    match Map.find x e with
      DEC(d) -> d
    | _ -> Error.errAt fi (name x^" was supposed to be a declaration here.")
  with Not_found -> Error.bugAt fi ("Enco.getDec: "^name x^" not defined.")

let getProc fi x e = 
  try
    match Map.find x e with
      PROC(p) -> p
    | _ -> Error.errAt fi (name x^" was supposed to be a process here.")
  with Not_found -> Error.bugAt fi ("Enco.getProc: "^name x^" not defined.")
      
let getPat fi x e =
  try match Map.find x e with
    PAT(a) -> a
  | NAME(id) -> VARp(fi,id,None)
  | VALUE(VARv(fi,id)) -> VARp(fi,id,None)  (* Note: not sure if good but it allowed for some 
					       tricky simulation of replicated input encoding *)
  | _ -> Error.bugAt fi "Enco.getPat: Expected pattern"
  with Not_found -> Error.bugAt fi ("Enco.getPat: Pattern " ^ (name x) ^ " not defined")

let getTy fi x e =
  try match Map.find x e with
    TY(t) -> t
  | _ -> Error.bugAt fi "Enco.getTy: Expected type"
  with Not_found -> Error.bugAt fi ("Enco.getTy: Type " ^ (name x) ^ " not defined")
    
let rec convertToVal = function
    RECORDp(fi,l) -> RECORDv(fi,List.map convertToFieldVal l)
  | ASp(fi,x,topt,pat) -> convertToVal pat
  | VARp(fi,x,topt) -> VARv(fi,x)
  | WILDp(fi,topt) -> Error.bugAt fi "Enco.convertToVal: compiler bug"
  | RECp(fi,topt,pat) -> RECv(fi,convertToVal pat,topt)
  | ENCODE_PAT(fi,x) -> ENCODE_VALUE(fi,x)

and convertToFieldVal = function
    ANONp(fi,pat) -> ANONv(fi,convertToVal pat)
  | LABELEDp(fi,s,pat) -> LABELEDv(fi,s,convertToVal pat)
  | TYp(fi,x,constr) -> Error.bugAt fi "Enco.convertToFieldVal: compiler bug"

let getVal fi x e =
  try match Map.find x e with
    VALUE(v) -> v
  | PAT(a) -> (convertToVal a)
  | NAME(id) -> VARv(fi,id)
  | _ -> Error.bugAt fi "Enco.getVal: Expected value"
  with Not_found -> Error.errAt fi ("Value " ^ (name x) ^ " not defined")

(*
 * Extract file position information.
 *)

let procInfo = function
  PRL(i,_,_) -> i | IF(i,_,_,_) -> i | INPUT(i,_,_,_) -> i | OUTPUT(i,_,_) -> i 
| INPA(i,_,_,_) -> i | OUTA(i,_,_) -> i | LDOUTA(i,_,_,_) -> i | TRYHERE(i,_,_,_,_) -> i
| REP_INPUT(i,_,_,_) -> i | TIMED_INPUT(i,_,_,_,_,_) -> i | REP_TIMED_INPUT(i,_,_,_,_,_) -> i 
| LI_SEND(i,_,_,_) -> i | REMOTE_SEND(i,_,_,_,_) -> i 
| LOCAL_SEND(i,_,_,_) -> i | IFLOCAL(i,_,_,_,_,_) -> i
| LET(i,_,_) -> i | SKIP(i) -> i
| TERMINATE(i) -> i | TYPECASE(i,_,_,_) -> i
| CASE(i,_,_,_) -> i | ENCODE_PROC(i,_,_) -> i

let patInfo = function
  RECORDp(i,_) -> i | ASp(i,_,_,_) -> i | VARp(i,_,_) -> i
| WILDp(i,_) -> i | RECp(i,_,_) -> i | ENCODE_PAT(i,_) -> i

let valInfo = function
  CCODEv(i,_,_) -> i | APPv(i,_,_,_) -> i | RECORDv(i,_) -> i
| STRINGv(i,_) -> i | SITEv(i,_,_) -> i | SELECTv(i,_,_) -> i | VARv(i,_) -> i
| BOOLv(i,_) -> i | CHARv(i,_) -> i | INTv(i,_) -> i | LETv(i,_,_) -> i
| RECv(i,_,_) -> i | IFv(i,_,_,_,_) -> i | WITHv(i,_,_,_) -> i
| WHEREv(i,_,_,_) -> i | ABSv(i,_,_) -> i
| DYNAMICv(i,_,_) -> i | TYPECASEv(i,_,_,_,_) -> i
| VARIANTv(i,_,_) -> i | CASEv(i,_,_,_) -> i | ENCODE_VALUE(i,_) -> i

let fvInfo = function
  ANONv(i,_) -> i | LABELEDv(i,_,_) -> i | TYv(i,_) -> i

let decInfo = function
  DEF(i,_) -> i | INLINE(i,_,_,_) -> i | NEW(i,_,_) -> i
| VAL(i,_,_) -> i | SEQ(i,_,_) -> i | RUN(i,_) -> i
| NEWLOC(i,_) -> i | CREATE(i,_) -> i | CREATE_AGENT(i,_) -> i | AT(i,_) -> i | MIGRATE(i,_) -> i | MACRO(i,_,_) -> i | MACROLET(i,_,_,_) -> i
| INLINE_DEC(i,_) -> i | INLINE_VAL(i,_) -> i 
| DEFER_DEC(i,_) -> i | ENCODE_DEC(i,_,_) -> i

let typeInfo = function
  TVAR(i,_) -> i | TABS(i,_,_) -> i | TAPP(i,_,_) -> i | TOP(i) -> i
| REC(i,_,_,_) -> i | RECORD(i,_) -> i | WITH(i,_,_) -> i
| RCHAN(i,_) -> i | OCHAN(i,_) -> i | STRING(i) -> i
| BOOL(i) -> i | CHAN(i,_) -> i | CHAR(i) -> i | INT(i) -> i
| ICHAN(i,_) -> i | WHERE(i,_,_) -> i
| SITE(i) -> i | LOC(i) -> i | AGENTT(i,_) -> i | AGENT(i) -> i 
| DYN(i) -> i | VART(i,_) -> i | MAP(i,_,_) -> i | ENCODE_TYPE(i,_) -> i





