(*
 * Syntax and simple utilities for encodings.
 *)


val anotherPass: bool ref 
    
type id
module Map : Map.S with type key = id
module IdMap : Misc.S with type key = id
      
type proc =
    PRL of Error.info * proc * proc
  | IF of Error.info * value * proc * proc
  | INPUT of Error.info * value * pat * proc
  | INPA of Error.info * value * pat * proc
  | OUTPUT of Error.info * value * value
  | OUTA of Error.info * value * value
  | LDOUTA of Error.info * value * value * value
  | TRYHERE of Error.info * value * value * proc * proc
  | REP_INPUT of Error.info * value * pat * proc
  | TIMED_INPUT of Error.info * value * value * pat * proc * proc
  | REP_TIMED_INPUT of Error.info * value * value * pat * proc * proc
  | LI_SEND of Error.info * value * value * value
  | REMOTE_SEND of Error.info * value * value * value * value
  | LOCAL_SEND of Error.info * value * value * value
  | IFLOCAL of Error.info * value * value * value * proc * proc
  | LET of Error.info * dec * proc
  | SKIP of Error.info
  | TERMINATE of Error.info
  | TYPECASE of Error.info * value * (pat * proc) list * proc
  | CASE of Error.info * ty option * value * (string * pat * proc) list
  | ENCODE_PROC of Error.info * id * id
	
and pat =
    RECORDp of Error.info * fieldPat list
  | ASp of Error.info * id * ty option * pat
  | VARp of Error.info * id * ty option
  | WILDp of Error.info * ty option
  | RECp of Error.info * ty option * pat
  | ENCODE_PAT of Error.info * id
	
and fieldPat =
    ANONp of Error.info * pat
  | LABELEDp of Error.info * string * pat
  | TYp of Error.info * id * constr
	
and dec =
    DEF of Error.info * (Error.info * id * pat * proc) list
  | INLINE of Error.info * id * pat * proc
  | NEW of Error.info * id * ty
  | VAL of Error.info * pat * value
  | SEQ of Error.info * dec * dec
  | RUN of Error.info * proc
  | NEWLOC of Error.info * id
  | CREATE of Error.info * (Error.info * id * ty * proc) list
  | CREATE_AGENT of Error.info * (Error.info * id * proc) list
  | AT of Error.info * id
  | MIGRATE of Error.info * value
  | MACRO of Error.info * string * value
  | MACROLET of Error.info * string * id * value
  | INLINE_DEC of Error.info * dec
  | INLINE_VAL of Error.info * value 
  | DEFER_DEC of Error.info * Syntax.dec
  | ENCODE_DEC of Error.info * id * id

and ty =
    TOP of Error.info
  | TVAR of Error.info * id
  | TABS of Error.info * (id * Kind.kind) list * ty
  | REC of Error.info * id * Kind.kind * ty
  | TAPP of Error.info * ty * ty list
  | RECORD of Error.info * fieldTy list
  | WITH of Error.info * ty * fieldTy list
  | WHERE of Error.info * ty * fieldTy list
  | CHAN of Error.info * ty
  | OCHAN of Error.info * ty
  | ICHAN of Error.info * ty
  | RCHAN of Error.info * ty
  | STRING of Error.info
  | BOOL of Error.info
  | CHAR of Error.info
  | INT of Error.info
  | SITE of Error.info
  | LOC of Error.info
  | AGENTT of Error.info * ty
  | AGENT of Error.info
  | DYN of Error.info
  | VART of Error.info * (string * ty) list
  | MAP of Error.info * ty * ty
  | ENCODE_TYPE of Error.info * id

and constr =
    LT of ty
  | EQ of ty

and fieldTy =
    ANON of Error.info * ty
  | LABELED of Error.info * string * ty
  | TYP of Error.info * id * constr
	
and value =
    CCODEv of Error.info * Ccode.info * fieldVal list
  | APPv of Error.info * ty option * value * fieldVal list
  | RECORDv of Error.info * fieldVal list
  | WITHv of Error.info * ty option * value * fieldVal list
  | WHEREv of Error.info * ty option * value * fieldVal list
  | STRINGv of Error.info * string
  | SITEv of Error.info * string * int
  | SELECTv of Error.info * value * string
  | VARv of Error.info * id
  | BOOLv of Error.info * bool
  | CHARv of Error.info * char
  | INTv of Error.info * int
  | LETv of Error.info * dec * value
  | RECv of Error.info * value * ty option
  | IFv of Error.info * ty option * value * value * value
  | ABSv of Error.info * pat * proc
  | DYNAMICv of Error.info * value * ty option
  | TYPECASEv of Error.info * ty option * value * (pat * value) list * value
  | VARIANTv of Error.info * string * value
  | CASEv of Error.info * ty option * value * (string * pat * value) list
  | ENCODE_VALUE of Error.info * id
	
and fieldVal =
    ANONv of Error.info * value
  | LABELEDv of Error.info * string * value
  | TYv of Error.info * ty

and enco =
    DEF_PROC of id list * proc
  | DEF_DEC of id list * dec
  | DEF_TYPE of id list * ty
  | DEF_DEF of Error.info * id * id * pat * proc

type f =
    NAME of id
  | PROC of proc
  | PAT of pat
  | DEC of dec
  | TY of ty
  | TYOPT of ty option
  | VALUE of value
  | LIST of f list
  | TUPLE of f * f
  | TRIPLE of f * f * f

type e = f Map.t  (* Maps ids to actual syntactic forms *)
val enull : e

(*
 * Extract file position Error.information.
 *)

val procInfo: proc -> Error.info
val patInfo: pat -> Error.info
val valInfo: value -> Error.info
val fvInfo: fieldVal -> Error.info
val decInfo: dec -> Error.info
val typeInfo: ty -> Error.info

(*
 * Translation utilities.
 *)

val fresh : string -> id
val str_of : id -> string
val int_of : id -> int
val createId : string -> int -> id
val name : id -> string
val format : id -> unit
val caml_format : id -> unit

type 'a ec =
    {
     types: Types.tag Map.t;           (* Maps type variables to unique tags *)
     constr: Types.constr Types.Map.t; (* Maps tags to constraints *)
     vars: (Types.t * 'a) IdMap.t;     (* Maps identifiers to types, plus an 'a  *)
     
     tvid: id Types.Map.t;             (* Maps tags to identifiers *)
     
     forms: enco Misc.StringMap.t;     (* Definitions of syntactic forms *)
     funs: enco Misc.StringMap.t       (* Definitions of library 
                                             and user-defined functions *)
   }

type toplevel =
    {imports : (Error.info * string) list;
     static : string list;
     includes : string list;
     dec : dec;
     ec : Inter.atom ec}

val empty : 'a ec
val abind : 'a ec -> id -> Types.tag -> 'a ec
val constr : 'a ec -> Types.tag -> Types.constr -> 'a ec
val xbind : IdMap.key -> Types.t -> 'a -> 'a ec -> 'a ec

val ibind : 'a ec -> Types.tag -> id -> 'a ec

val ebind : string -> enco -> 'a ec -> 'a ec
val fbind : string -> enco -> 'a ec -> 'a ec

val lookupTVId : Error.info -> Types.tag -> 'a ec -> id
val lookupTVTag : Error.info -> id -> 'a ec -> Types.tag

val bind : Error.info -> id list -> f list -> e -> e

val modifyDec : Error.info -> value -> dec -> (id * dec)
val modifyProc : Error.info -> value -> proc -> (id * proc)

val getId : Error.info -> id -> e -> id
val getName : Error.info -> id -> e -> string
val getDec : Error.info -> id -> e -> dec
val getProc : Error.info -> id -> e -> proc
val getPat : Error.info -> id -> e -> pat
val getTy : Error.info -> id -> e -> ty
val getVal : Error.info -> id -> e -> value








