(*
 * Intermediate code
 *)

let exactInter =
  Flags.createFlag "exactInter"
  "*Show exact representation of intermediate code" false

type status =
  EMPTY
| REPLICATED
| ONEREADER
| ONEWRITER
| KNOWN
| UNKNOWN

let unionStatus a b =
  match (a,b) with
    (EMPTY,EMPTY) -> EMPTY
  | (REPLICATED,REPLICATED) -> REPLICATED
  | (ONEREADER,ONEREADER) -> ONEREADER
  | (ONEWRITER,ONEWRITER) -> ONEWRITER
  | (KNOWN,KNOWN) -> KNOWN
  | (_,_) -> UNKNOWN

let printStatus = function
  EMPTY -> Format.print_string "{-e-}"
| REPLICATED -> Format.print_string "{-*-}"
| ONEREADER -> Format.print_string "{-r-}"
| ONEWRITER -> Format.print_string "{-w-}"
| KNOWN -> Format.print_string "{-k-}"
| UNKNOWN -> ()

let interStatus a b =
  match (a,b) with
    (EMPTY,EMPTY) -> EMPTY
  | (REPLICATED,REPLICATED) -> REPLICATED
  | (ONEREADER,ONEREADER) -> ONEREADER
  | (ONEWRITER,ONEWRITER) -> ONEWRITER
  | (KNOWN,KNOWN) -> KNOWN
  | (UNKNOWN,s) -> s
  | (s,UNKNOWN) -> s
  | (KNOWN,s) -> s
  | (s,KNOWN) -> s
  | (_,_) -> Error.bugf (fun () ->
      Format.print_string "interStatus: inconsistent status values: ";
      printStatus a; Format.print_string " and "; printStatus b
    )

type inlinable = INLINABLE | NOTINLINABLE
type init = REQUIRED | NOTREQUIRED

type call =
    TEST_EQUALITY of Var.var * Var.var
  | DUMMY
  | GETSITE

type proc =
  PRL of proc * proc
| VAL of value list * proc
| CCODE of Var.var option * Ccode.info * atom list * proc
| CWAIT of int 
| ATOM of Var.var * atom * proc
| IF of atom * proc * proc
| NEW of Var.var * init * proc
| INPUT of atom * atom * status
| OUTPUT of atom * atom * status
| SEND of atom * atom list
| STRUCT of atom Types.c
| SKIP 

| ICODE of Var.var option * call * proc
| REP_INPUT of atom * atom
| TIMED_INPUT of atom * atom * atom * proc * int
| REP_TIMED_INPUT of atom * atom * atom * proc * int
| IFLOCAL of atom * atom * atom * proc * proc
| MIGRATE of atom * int * proc
| TERMINATE

and atom =
  INT of int
| CHAR of char
| BOOL of bool
| ADDR of Var.var
| STATIC of Var.var
| DYNAMIC of Var.var
| PROJECT of atom * int
| CCONST of Ccode.info

and value =
  TUPLE of Var.var * atom list
| DEF of Var.var * inlinable * Var.var option list * proc
| STRING of Var.var * string
| CREATE_AGENT of Var.var * proc

type toplevel =
  {includes : string list;
   static : string list;
   constants : value list;
   proc : proc}

let small = function
  INPUT(_) | OUTPUT(_) | SEND(_) | SKIP -> true
| PRL(_) | VAL(_) | CCODE(_) | CWAIT(_) | ATOM(_)
| IF(_) | NEW(_) | STRUCT(_) 
| ICODE(_) | IFLOCAL(_) | MIGRATE(_) -> false
| REP_INPUT(_) | TIMED_INPUT(_) | REP_TIMED_INPUT(_) | TERMINATE -> true

let formatBinder = function
  None -> Format.print_string "_"
| Some x -> Var.format x

let caml_formatCall = function
    TEST_EQUALITY(x,y) -> 
      Format.print_string "TEST_EQUALITY(";
      Var.format x; Format.print_char ','; 
      Var.format y; 
      Format.print_char ')'
  | DUMMY -> Format.print_string "DUMMY"
  | GETSITE -> Format.print_string "GETSITE"

let rec printProc bi = function
  PRL(p,q) ->
    if small p || !exactInter then
      (Format.print_string "run"; Format.print_break 1 2;
       printBoxProc bi p; Format.print_space(); Format.print_string "end";
       Format.print_space(); printBoxProc bi q)
    else
      (Format.print_string "run"; Format.print_break 1 2;
       printBoxProc bi q; Format.print_space(); Format.print_string "end";
       Format.print_space(); printBoxProc bi p)
| VAL(l,p) ->
    printVals bi l; Format.print_space(); printProc bi p
| CCODE(None,ci,l,p) ->
    Format.open_hovbox 0; printCi ci l; Format.close_box();
    Format.print_space(); printProc bi p
| CCODE(Some x,ci,l,p) ->
    Format.open_hovbox 0; Format.print_string "val "; Var.format x; bi x;
    Format.print_string " ="; Format.print_space(); printCi ci l;
    Format.close_box(); Format.print_space(); printProc bi p
| CWAIT(i) -> Format.print_string "cwait"
| ATOM(x,a,p) ->
    Format.open_hovbox 0; Format.print_string "val "; Var.format x; bi x;
    Format.print_string " ="; Format.print_space(); printAtom a;
    Format.close_box(); Format.print_space(); printProc bi p
| IF(a,p,q) ->
    Format.print_string "if";
      Format.print_break 1 2; printAtom a; Format.print_space();
    Format.print_string "then";
      Format.print_break 1 2; printBoxProc bi p; Format.print_space();
    Format.print_string "else";
      Format.print_break 1 2; printBoxProc bi q; Format.print_space();
    Format.print_string "end"
| NEW(x,REQUIRED,p) ->
    Format.print_string "new "; Var.format x; bi x;
    Format.print_space(); printProc bi p
| NEW(x,NOTREQUIRED,p) ->
    Format.print_string "NEW "; Var.format x; bi x;
    Format.print_space(); printProc bi p
| INPUT(x,y,s) ->
    printAtom x; Format.print_string "?"; printStatus s; printAtom y
| OUTPUT(x,y,s) ->
    printAtom x; Format.print_string "!"; printStatus s; printAtom y
| SEND(x,l) ->
    printAtom x; Format.print_string "!!";
    Misc.formatList "[" "]" " " printAtom l
| STRUCT(c) ->
    Format.open_hovbox 0;
      Format.print_string "(struct ";
      Format.print_break 1 1; Types.printC c; Format.close_box();
      Format.print_cut(); Format.print_string ")";
    Format.close_box()
| SKIP ->
    Format.print_string "skip"
| ICODE(None,e,p) -> 
    Format.open_hovbox 0; caml_formatCall e; Format.close_box();
    Format.print_space(); printProc bi p
| ICODE(Some x,e,p) -> 
    Format.open_hovbox 0; Format.print_string "val "; Var.format x; bi x;
    Format.print_string " ="; Format.print_space(); caml_formatCall e;
    Format.close_box(); Format.print_space(); printProc bi p
| REP_INPUT(x,y) ->
    printAtom x; Format.print_string "?*"; printAtom y
| TIMED_INPUT(x,y,z,p,i) ->
    Format.print_string "timeout";
    Format.print_break 1 2; printAtom x; Format.print_string "on";
      printAtom y; Format.print_string "?"; printAtom z; Format.print_space();
    Format.print_string "then";
      Format.print_break 1 2; printBoxProc bi p; Format.print_space();
    Format.print_string "end"; print_int i;
| REP_TIMED_INPUT(x,y,z,p,i) ->
    Format.print_string "timeout";
    Format.print_break 1 2; printAtom x; Format.print_string "on";
      printAtom y; Format.print_string "?*"; printAtom z; Format.print_space();
    Format.print_string "then";
      Format.print_break 1 2; printBoxProc bi p; Format.print_space();
    Format.print_string "end"; print_int i;
| IFLOCAL(x,y,z,p,q) ->
    Format.print_string "iflocal";
      Format.print_break 1 2; 
      Format.print_string "<"; printAtom x; Format.print_string ">"; 
      printAtom y; Format.print_string "!"; printAtom z; Format.print_space();
    Format.print_string "then";
      Format.print_break 1 2; printBoxProc bi p; Format.print_space();
    Format.print_string "else";
      Format.print_break 1 2; printBoxProc bi q; Format.print_space();
    Format.print_string "end"
| MIGRATE(x,i,p) -> 
    Format.print_string "migrate to "; printAtom x; Format.print_string " then";
    Format.print_break 1 2; printBoxProc bi p; Format.print_space();
| TERMINATE -> 
    Format.print_string "terminate"

and printBoxProc bi p =
  Format.open_hvbox 0; printProc bi p; Format.close_box()

and printAtom = function
  INT(i) -> Format.print_int i
| CHAR(c) ->
    Format.print_string "'"; Format.print_char c; Format.print_string "'"
| BOOL(b) -> Format.print_bool b
| ADDR(x) -> Format.print_char '&'; Var.format x
| STATIC(x) -> Format.print_char '%'; Var.format x
| DYNAMIC(x) -> Var.format x
| PROJECT(a,i) -> printAtom a; Format.print_char '.'; Format.print_int i
| CCONST(ci) -> printCi ci []

and printVals bi l =
  let rec pr needSep = function
    [] -> ()
  | h :: t ->
    if needSep then Format.print_cut();
    Format.open_hovbox 0; printVal bi h; Format.close_box();
    pr true t
in
  Format.open_hvbox 0; pr false l; Format.close_box()

and printVal bi = function
  TUPLE(x,l) ->
    Format.print_string "val "; Var.format x; bi x;
    Format.print_string " ="; Format.print_space();
    Misc.formatList "[" "]" " " printAtom l
| DEF(x,i,args,p) ->
    (match i with
       INLINABLE -> Format.print_string "inline "
     | NOTINLINABLE -> ());
    Format.print_string "def "; Var.format x; bi x;
    Misc.formatList " [" "]" " " formatBinder args;
    Format.print_string " >"; Format.print_break 1 2; printBoxProc bi p
| STRING(x,s) ->
    Format.print_string "val "; Var.format x; bi x;
    Format.print_string " ="; Format.print_space(); Format.print_string "\"";
    Format.print_string (String.escaped s); Format.print_string "\""
| CREATE_AGENT(x,p) ->
    Format.print_string "agent "; Var.format x;
    Format.print_string " ="; Format.print_break 1 2; printBoxProc bi p

and printCi ci l =
  let rec print sl al = match (sl,al) with
    ([s],[]) -> Format.print_string s
  | (s :: sl,a :: al) -> Format.print_string s; printAtom a; print sl al
  | _ -> Error.bug "printCCode: Wrong number of arguments in CCODE"
in
  Format.open_hbox();
  Format.print_string "(ccode ";
  Format.print_int ci.Ccode.alloc; Format.print_string " ";
  (match ci.Ccode.kind with
    Ccode.EXP ->
    (match (ci.Ccode.const,ci.Ccode.reader,ci.Ccode.writer) with
       ( true,false,false) -> Format.print_string "C "
     | (false,false,false) -> Format.print_string "P "
     | (false, true,false) -> Format.print_string "R "
     | (false, true, true) -> Format.print_string "W "
     | (    _,    _,    _) -> Error.bug "Inter.printCCode: Bad purity")
  | Ccode.STATEMENT -> Format.print_string "E "
  | Ccode.NEXP -> Format.print_string "NEXP "
  | Ccode.NSTATEMENT -> Format.print_string "NSTATEMENT ");
  if !Ccode.cc then 
    print ci.Ccode.code l 
  else
    begin
      Format.print_space(); Ccode.caml_formatInfo ci;
      Format.print_char ','; Misc.formatList "[" "]" "; " printAtom l
    end; 
  Format.print_string ")";
  Format.close_box()

let printToplevel s = Format.print_string s; Format.print_newline()

let print bi prog =
  Format.print_string "{- Includes -}";
  Format.print_newline(); List.iter printToplevel prog.includes;
  Format.print_string "{- Static C Code -}";
  Format.print_newline(); List.iter printToplevel prog.static;
  Format.print_string "{- Constants -}";
  Format.print_newline(); printVals bi prog.constants; Format.print_newline();
  Format.print_string "{- Processes -}";
  Format.print_newline(); printBoxProc bi prog.proc; Format.print_newline()

(*
 * printing an intermediate language (O'Caml format).
 *)

let caml_formatStatus = function
    EMPTY      -> Format.print_string "EMPTY"
  | REPLICATED -> Format.print_string "REPLICATED"
  | ONEREADER  -> Format.print_string "ONEREADER"
  | ONEWRITER  -> Format.print_string "ONEWRITER"
  | KNOWN      -> Format.print_string "KNOWN"
  | UNKNOWN    -> Format.print_string "UNKNOWN"

let rec caml_formatProc bi = function
    PRL(p,q) ->
      Format.print_string "PRL("; Format.print_break 1 2; caml_formatBoxProc bi p; 
      Format.print_char ','; Format.print_space(); caml_formatBoxProc bi q; 
      Format.print_char ')'
  | VAL(l,p) ->
      Format.print_string "VAL("; caml_formatVals bi l; Format.print_char ',';
      Format.print_space(); caml_formatProc bi p; Format.print_char ')'
  | CCODE(None,ci,l,p) ->
      Format.open_hovbox 0; 
      Format.print_string "CCODE(None,"; Ccode.caml_formatInfo ci; 
      Format.print_char ','; Misc.formatList "[" "]" "; " caml_formatAtom l; 
      Format.close_box(); Format.print_space(); 
      caml_formatProc bi p; Format.print_char ')'
  | CCODE(Some x,ci,l,p) ->
      Format.open_hovbox 0;  
      Format.print_string "CCODE(Some "; Var.format x; bi x;
      Format.print_char ','; Format.print_space(); Ccode.caml_formatInfo ci; 
      Format.print_char ','; Misc.formatList "[" "]" "; " caml_formatAtom l;
      Format.print_char ',';
      Format.close_box(); Format.print_space(); 
      caml_formatProc bi p; Format.print_char ')'
  | CWAIT(i) ->   
      Format.print_string "CWAIT("; Format.print_int i; Format.print_char ')'
  | ATOM(x,a,p) ->
      Format.open_hovbox 0; 
      Format.print_string "ATOM("; Var.format x; bi x; 
      Format.print_char ','; Format.print_space(); 
      caml_formatAtom a; Format.print_char ','; 
      Format.close_box(); Format.print_space(); 
      caml_formatProc bi p; Format.print_char ')'
  | IF(a,p,q) ->
      Format.print_string "IF("; Format.print_break 1 2; caml_formatAtom a; 
      Format.print_char ','; Format.print_space(); Format.print_break 1 2; 
      caml_formatBoxProc bi p; Format.print_char ','; Format.print_space(); 
      Format.print_break 1 2; caml_formatBoxProc bi q;  Format.print_char ')'
  | NEW(x,REQUIRED,p) ->
      Format.print_string "NEW("; Var.format x; bi x; 
      Format.print_string ",REQUIRED,"; Format.print_space(); 
      caml_formatProc bi p; Format.print_char ')'
  | NEW(x,NOTREQUIRED,p) ->
      Format.print_string "NEW("; Var.format x; bi x;
      Format.print_string ",NOTREQUIRED,"; Format.print_space(); 
      caml_formatProc bi p; Format.print_char ')'
  | INPUT(x,y,s) ->
      Format.print_string "INPUT("; caml_formatAtom x; Format.print_char ','; 
      caml_formatAtom y; Format.print_char ','; caml_formatStatus s;
      Format.print_char ')'
  | OUTPUT(x,y,s) ->
      Format.print_string "OUTPUT("; caml_formatAtom x; Format.print_char ','; 
      caml_formatAtom y; Format.print_char ','; caml_formatStatus s; 
      Format.print_char ')'
  | SEND(x,l) ->
      Format.print_string "SEND("; caml_formatAtom x; Format.print_char ',';
      Misc.formatList "[" "]" "; " caml_formatAtom l; Format.print_char ')'
  | STRUCT(c) ->
      Format.open_hovbox 0;
      Format.print_string "STRUCT(";
      Format.print_break 1 1; Format.print_newline();
      Types.caml_formatC c caml_formatAtom; Format.close_box();
      Format.print_cut(); Format.print_char ')';
      Format.close_box()
  | SKIP -> 
      Format.print_string "SKIP"
  | ICODE(None,e,p) -> 
      Format.print_string "ICODE(None,"; caml_formatCall e;
      Format.print_char ','; caml_formatProc bi p; Format.print_char ')'
  | ICODE(Some x,e,p) -> 
      Format.print_string "ICODE(Some "; Var.format x; bi x;
      Format.print_char ','; caml_formatCall e; 
      Format.print_char ','; caml_formatProc bi p; Format.print_char ')'
  | REP_INPUT(x,y) ->
      Format.print_string "REP_INPUT("; caml_formatAtom x; Format.print_char ',';
      caml_formatAtom y; Format.print_char ')';
  | TIMED_INPUT(x,y,z,p,i) ->
      Format.print_string "TIMED_INPUT("; Format.print_break 1 2; caml_formatAtom x; 
      Format.print_char ','; caml_formatAtom y; Format.print_char ','; 
      caml_formatAtom z; Format.print_char ','; Format.print_space(); 
      Format.print_break 1 2; caml_formatBoxProc bi p; 
      Format.print_char ',';Format.print_int i; Format.print_char ')'
  | REP_TIMED_INPUT(x,y,z,p,i) ->
      Format.print_string "REP_TIMED_INPUT("; Format.print_break 1 2; caml_formatAtom x; 
      Format.print_char ','; caml_formatAtom y; Format.print_char ','; 
      caml_formatAtom z; Format.print_char ','; Format.print_space(); 
      Format.print_break 1 2; caml_formatBoxProc bi p; 
      Format.print_char ',';Format.print_int i; Format.print_char ')'
  | IFLOCAL(x,y,z,p,q) ->
      Format.print_string "IFLOCAL("; Format.print_break 1 2; caml_formatAtom x; 
      Format.print_char ','; caml_formatAtom y; Format.print_char ','; 
      caml_formatAtom z; Format.print_char ','; Format.print_space(); 
      Format.print_break 1 2; caml_formatBoxProc bi p; Format.print_char ','; 
      Format.print_space(); Format.print_break 1 2; caml_formatBoxProc bi q;  
      Format.print_char ')'
  | MIGRATE(x,i,p) -> 
      Format.print_string "MIGRATE("; caml_formatAtom x; Format.print_char ','; 
      Format.print_int i; Format.print_char ',';
      caml_formatProc bi p; Format.print_char ')'
  | TERMINATE -> 
      Format.print_string "TERMINATE"

and caml_formatBoxProc bi p = 
  Format.open_hvbox 0; caml_formatProc bi p; Format.close_box()
    
and caml_formatAtom = function
    INT(i) -> 
      Format.print_string "INT("; Format.print_int i; Format.print_char ')'
  | CHAR(c) ->
      Format.print_string "CHAR("; Format.print_char c; Format.print_char ')'
  | BOOL(b) -> 
      Format.print_string "BOOL("; Format.print_bool b; Format.print_char ')'
  | ADDR(x) -> 
      Format.print_string "ADDR("; Var.format x; Format.print_char ')'
  | STATIC(x) ->
      Format.print_string"STATIC("; Var.format x; Format.print_char ')'
  | DYNAMIC(x) -> 
      Format.print_string"DYNAMIC("; Var.format x; Format.print_char ')'
  | PROJECT(a,i) -> 
      Format.print_string"PROJECT("; caml_formatAtom a; Format.print_char ','; 
      Format.print_int i; Format.print_char ')'
  | CCONST(ci) -> 
      Format.print_string"CCONST("; Ccode.caml_formatInfo ci; Format.print_char ')'

and caml_formatVals bi l =
  let rec pr needSep = function
      [] -> ()
    | h :: t ->
      	if needSep then begin
	  Format.print_char ',';
	  Format.print_cut();
          Format.open_hovbox 0; 
	  caml_formatVal bi h; 
	  Format.close_box();
    	  pr true t
      	end 
	else begin
	  Format.open_hovbox 0; 
	  caml_formatVal bi h; 
	  Format.close_box();
    	  pr true t
    	end
  in
  Format.print_string "[";
  Format.open_hvbox 0; 
  pr false l; 
  Format.print_string "]";
  Format.close_box()
    
and caml_formatVal bi = function
    TUPLE(x,l) ->
      Format.print_string "TUPLE("; Var.format x; bi x; Format.print_char ',';
      Format.print_space(); Misc.formatList "[" "]" "; " caml_formatAtom l;
      Format.print_char ')'
  | DEF(x,i,args,p) ->
      let caml_formatBinder = function
	  None -> Format.print_string "None"
  	| Some x -> Format.print_string "Some "; Var.format x
      in
      Format.print_string "DEF(";
      Var.format x; bi x; Format.print_char ',';
      (match i with
       	INLINABLE -> Format.print_string "INLINABLE,"
      | NOTINLINABLE -> Format.print_string "NOTINLINABLE,");
      Misc.formatList " [" "]" "; " caml_formatBinder args; Format.print_char ','; 
      Format.print_break 1 2; caml_formatBoxProc bi p; Format.print_char ')'
  | STRING(x,s) ->
      Format.print_string "STRING("; Var.format x; bi x; 
      Format.print_char ','; Format.print_space(); Format.print_string "\""; 
      Format.print_string (String.escaped s); Format.print_string "\"";
      Format.print_char ')'
  | CREATE_AGENT(x,p) ->
      Format.print_string "CREATE_AGENT("; Var.format x; Format.print_char ','; 
      Format.print_break 1 2; caml_formatBoxProc bi p; Format.print_char ')'

let caml_format bi prog =
  Format.print_string "{- Includes -}";
  Format.print_newline(); List.iter printToplevel prog.includes;
  Format.print_string "{- Static C Code -}";
  Format.print_newline(); List.iter printToplevel prog.static;
  Format.print_string "{- Constants -}";
  Format.print_newline(); caml_formatVals bi prog.constants; 
  Format.print_newline();
  Format.print_string "{- Processes -}";
  Format.print_newline(); caml_formatBoxProc bi prog.proc; Format.print_newline()

(* end of printing *)


let rec join c p = match p with
  STRUCT(_) -> c
| PRL(p,q) -> PRL(join c p,join c q)
| VAL(l,p) -> VAL(List.map (joinVal c) l,join c p)
| CCODE(x,ci,l,p) -> CCODE(x,ci,l,join c p) 
| CWAIT(_) -> p
| ATOM(x,a,p) -> ATOM(x,a,join c p)
| NEW(x,i,p) -> NEW(x,i,join c p)
| IF(_) | INPUT(_) | OUTPUT(_) | SEND(_) | SKIP -> p  
| ICODE(o,e,p) -> ICODE(o,e,join c p)
| REP_INPUT(_) | TIMED_INPUT(_) | REP_TIMED_INPUT(_) | IFLOCAL(_) | TERMINATE -> p
| MIGRATE(x,i,p) -> MIGRATE(x,i,join c p)

and joinVal c v = match v with
  DEF(x,i,args,p) -> DEF(x,i,args,join c p)
| TUPLE(_) | STRING(_) -> v
| CREATE_AGENT(x,p) -> CREATE_AGENT(x,join c p)

let joinImports imports prog =
  let p =
    List.fold_right (fun prog q -> join q prog.proc) imports
    (join SKIP prog.proc)
  in
    {includes = prog.includes; static = prog.static;
     constants = prog.constants; proc = p}

let cloneTable = (Hashtbl.create 200 : (Var.var,Var.var) Hashtbl.t)
let clone x = let y = Var.fresh() in Hashtbl.add cloneTable x y; y
let cloneOpt = function None -> None | Some x -> Some(clone x)
let cloneVar x = try Hashtbl.find cloneTable x with Not_found -> x

let rec cloneProc = function
  PRL(p,q) -> PRL(cloneProc p,cloneProc q)
(* ocaml-1.05: | VAL(l,p) -> List.iter scanVal l; VAL(List.map cloneVal l,cloneProc p) *)
| VAL(l,p) -> ignore (List.map scanVal l); VAL(List.map cloneVal l,cloneProc p)
| IF(a,p,q) -> IF(cloneAtom a,cloneProc p,cloneProc q)
| STRUCT(inc) -> STRUCT(Types.mapBindings (fun (ty,a) -> (ty,cloneAtom a)) inc)
| CCODE(o,ci,l,p) ->
    let o = cloneOpt o in CCODE(o,ci,List.map cloneAtom l,cloneProc p)
| CWAIT(i) -> CWAIT(i)
| ATOM(x,a,p) -> let x = clone x in ATOM(x,cloneAtom a,cloneProc p)
| NEW(x,i,p) -> let x = clone x in NEW(x,i,cloneProc p)
| INPUT(x,y,s) -> INPUT(cloneAtom x,cloneAtom y,s)
| OUTPUT(x,y,s) -> OUTPUT(cloneAtom x,cloneAtom y,s)
| SEND(x,l) -> SEND(cloneAtom x,List.map cloneAtom l)
| SKIP -> SKIP
| ICODE(o,e,p) -> let o = cloneOpt o in ICODE(o,e,cloneProc p)
| REP_INPUT(x,y) -> REP_INPUT(cloneAtom x,cloneAtom y)
| TIMED_INPUT(x,y,z,p,i) -> TIMED_INPUT(cloneAtom x,cloneAtom y,cloneAtom z,cloneProc p,i)
| REP_TIMED_INPUT(x,y,z,p,i) -> REP_TIMED_INPUT(cloneAtom x,cloneAtom y,cloneAtom z,cloneProc p,i)
| IFLOCAL(x,y,z,p,q) -> IFLOCAL(cloneAtom x,cloneAtom y,cloneAtom z,
				cloneProc p,cloneProc q)
| MIGRATE(x,i,p) -> MIGRATE(cloneAtom x,i,cloneProc p)
| TERMINATE -> TERMINATE

and scanVal = function
  DEF(x,_,_,_) -> clone x
| TUPLE(x,_) -> clone x
| STRING(x,_) -> clone x
| CREATE_AGENT(x,_) -> clone x

and cloneVal = function
  DEF(x,i,args,p) ->
    let args = List.map cloneOpt args in DEF(cloneVar x,i,args,cloneProc p)
| TUPLE(x,l) -> TUPLE(cloneVar x,List.map cloneAtom l)
| STRING(x,s) -> STRING(cloneVar x,s)
| CREATE_AGENT(x,p) -> CREATE_AGENT(cloneVar x,cloneProc p)

and cloneAtom a = match a with
  DYNAMIC(x) -> DYNAMIC(cloneVar x)
| PROJECT(a,i) -> PROJECT(cloneAtom a,i)
| INT(_) | CHAR(_) | BOOL(_) | ADDR(_) | STATIC(_) | CCONST(_) -> a

let clone args p =
  let args = List.map cloneOpt args in let p = cloneProc p in
  Hashtbl.clear cloneTable; (args,p)



