(*
 * Types for PICT
 * ==============
 *)

open Types

let debug = Flags.createFlag "leq" "*Trace subtype checking" false

(*
 * Exception for internal use only, must be caught by all top-level functions.
 *)

exception Error of (unit -> unit)
let fail f = raise (Error f)

(* 
 * Instantiation of existential variables
 * ======================================
 *)

(* 
 * The occur check
 * ===============
 *
   Check that all of the free variables in s have binders above i,
   and that the reference r does not occur in any UVARs in s.
 *)

let occurCheck tag s =
  let rec check = function
    TVAR(_,_,_) -> ()
  | UVAR({contents = Some t},_) -> check t
  | UVAR({contents = None},tg) as v ->
    if tag <> tg then () else fail (fun () ->
      Format.print_string "since the occur check failed when instantiating";
      displayT v; Format.print_space(); Format.print_string "with"; displayT s
    )
  | TABS(_,_,s) -> check s
  | REC(_,_,_,t) -> check t
  | TAPP(_,s,l) -> check s; List.iter check l
  | TOP -> ()
  | RECORD(l) -> List.iter checkF l
  | RCHAN(t) -> check t
  | OCHAN(t) -> check t
  | ICHAN(t) -> check t
  | CHAN(t) -> check t
  | STRING -> ()
  | CHAR -> ()
  | BOOL -> ()
  | INT -> ()
  | SITE -> ()
  | LOC -> ()
  | AGENTT(t) -> check t
  | AGENT -> ()
  | DYN -> ()
  | VART(l) -> List.iter checkVF l
  | MAP(t1,t2) -> check t1; check t2
  | ENCODE_TYPE(l) -> ()  (* Bogus? *)
  | IGNORE -> ()

  and checkF = function
    ANON(t) -> check t
  | LABELED(_,t) -> check t
  | TY(_,_,LT t) -> check t
  | TY(_,_,EQ t) -> check t

  and checkVF (_,t) = check t
in
  check s

let rec bindUvar c r tag t =
  let bind r tag t = 
    match r with
      {contents = Some _} -> Error.bug "Attempt to bind known Uvar"
    | {contents = None} -> occurCheck tag t; r := Some t
in
  if !debug then 
    Debug.wrap "bindUvar"
    (fun () -> bind r tag t)
    (fun () -> Types.printTag tag; Format.print_string " <- "; printT t)
    (fun () -> ())
  else 
    bind r tag t

module Map = Types.Map

(*
 * Helper function which creates a fresh type variable W, and adds
 * assumptions W<B,X=W,Y=W, for a given constraint C and variables X and Y.
 *)

let makeEqual c ks ns tags kt nt tagt con =
  let t = freshTag() in
  constr (constr (constr c t con) tags (EQ(TVAR(ks,ns,t))))
  tagt (EQ (TVAR(kt,nt,t)))

(* 
 * Subtyping checker
 *)

let rec leq c s t = 
  try
    if !debug then
      Debug.wrap "leq"
      (fun () -> leqLight c s t)
      (fun () ->
         printT s; Format.print_space(); Format.print_string "<";
	 Format.print_space(); printT t)
      (fun () -> ())
    else
      leqLight c s t
  with
    Error(why1) -> fail(fun()->
      Format.print_string "since"; displayT s; Format.print_space();
      Format.print_string "is not a subtype of"; displayT t;
      Format.print_space(); why1())

and leqLight c s t = match (s,t) with
  (_, TOP) -> ()
| (_, IGNORE) -> ()
| (IGNORE, _) -> ()
| (TVAR(_,_,tags), TVAR(_,_,tagt)) ->
    if tags = tagt then () else leqExposed c (expose c s) (expose c t)
| (ENCODE_TYPE(k), ENCODE_TYPE(l)) -> 
    if k = l then () else leqExposed c (expose c s) (expose c t)
| (AGENTT(k), AGENTT(l)) ->
    if k = l then () else leqExposed c (expose c s) (expose c t)
| (_,_) ->
    leqExposed c (expose c s) (expose c t)

and leqExposed c s t = match (s,t) with
(*
 * Any type is less than Top.
 *)
| (_, TOP) -> ()
| (_, IGNORE) -> ()
| (IGNORE, _) -> ()
| (UVAR(rs,tags),UVAR(rt,tagt)) ->
  if tags = tagt then () else bindUvar c rs tags t
| (UVAR(rs,tags), t) -> bindUvar c rs tags t
| (s, UVAR(rt,tagt)) -> bindUvar c rt tagt s
(*
 * Subtyping of primitive types. ^ is invariant, ? is covariant,
 * ! and !* are contravariant, ^ < !, ^ < ?, and Char < Int.
 *)
| (CHAN s, CHAN t) -> leqLight c s t; leqLight c t s
| (OCHAN s, OCHAN t) -> leqLight c t s
| (ICHAN s, ICHAN t) -> leqLight c s t
| (CHAN s, ICHAN t) -> leqLight c s t
| (CHAN s, OCHAN t) -> leqLight c t s
| (RCHAN s, RCHAN t) -> leqLight c t s
| (RCHAN s, OCHAN t) -> leqLight c t s
| (STRING, STRING) -> ()
| (CHAR, CHAR) -> ()
| (BOOL, BOOL) -> ()
| (INT, INT) -> ()
| (CHAR, INT) -> ()
| (SITE, SITE) -> ()
| (LOC, LOC) -> ()
| (AGENTT s, AGENTT t) -> leqLight c s t; leqLight c t s
| (AGENT, AGENT) -> ()
| (DYN, DYN) -> ()
| (VART ls, VART lt) -> () (* Bogus *)
| (MAP(s1,s2), MAP(t1,t2)) -> leqLight c t1 s1; leqLight c t2 s2;
                              leqLight c s1 t1; leqLight c s2 t2

(*
 * Type operators are subtyped pointwise.
 *)
| (TABS(_,ls,s), TABS(_,lt,t)) ->
  let c =
    List.fold_left2 (fun c (ns,tags,ks) (nt,tagt,kt) ->
      makeEqual c ks ns tags kt nt tagt (LT(mkTop ks))
    ) c ls lt
  in
    leq c s t
(*
 * For recursive types, we do something slightly tricky.  First, we
 * try comparing the types for equivalence.  If this succeeds, we're
 * happy.  If not, then we try the Amber rule ("assuming the bound vars
 * are LEQ, show that the bodies are LEQ...").  We believe this strategy 
 * is complete, but it introduces a slight incompleteness in type
 * inference, since the first (unsuccessful) check may instantiate some
 * unification variables, and these will *not* be uninstantiated before
 * we try the Amber rule.  So if you want two REC types to stand in the
 * subtype relation because of the Amber rule, better make sure that there 
 * are no unification variables around!
 *)
| (REC(ks,ns,tags,s), REC(kt,nt,tagt,t)) -> 
  (try
     let c = makeEqual c ks ns tags kt nt tagt (LT(mkTop kt))
     in leq c s t; leq c t s
   with Error(_) ->
     let c = constr (constr c tagt (LT(mkTop kt))) tags (LT(TVAR(kt,nt,tagt)))
     in leq c s t)
(*
 * If s and t are identical type variables then we succeed. Otherwise,
 * we try and promote s so that we can continue.
 *)
| (TVAR(_,_,tags), TVAR(_,_,tagt)) ->
  if tags = tagt then ()
  else begin
    match promote c s with
      Some s -> leq c s t
    | None -> fail (fun () ->
      Format.print_string "since"; displayT s; Format.print_space();
      Format.print_string "cannot be (promoted to) a subtype of";
      displayT t)
  end
| (ENCODE_TYPE(k), ENCODE_TYPE(l)) ->
  if k = l then ()
  else begin
    match promote c s with
      Some s -> leq c s t
    | None -> fail (fun () ->
      Format.print_string "since"; displayT s; Format.print_space();
      Format.print_string "cannot be (promoted to) a subtype of";
      displayT t)
  end
(*
 * It must be the case that t is something other than a type variable.
 * Therefore, we must try and promote s so that we can continue.
 *)
| (TVAR(_,_,_), t) -> begin
  match promote c s with
    Some s -> leq c s t
  | None -> fail (fun()->
      Format.print_string "since"; displayT s; Format.print_space();
      Format.print_string "cannot be (promoted to) a subtype of";
      displayT t)
  end
| (TAPP(_,s,ls), TAPP(_,t,lt)) -> 
    if (kindOf s) <> (kindOf t) then
      fail (fun () ->
        Format.print_string "since the kind of";
        displayT s; Format.print_space();
        Format.print_string "is not the same as the kind of";
        displayT t
      )
    else
      (leq c s t; List.iter2 (leq c) ls lt; List.iter2 (leq c) lt ls)
| (TAPP(_,_,_), t) -> begin
    match promote c s with
      Some s -> leq c s t
    | None -> fail (fun () ->
      Format.print_string "since"; displayT s; Format.print_space();
      Format.print_string "and"; displayT t; Format.print_space();
      Format.print_string "do not match")
  end
(*
 * By this time, we have already tried reducing t, so we just fail.
 * There is no way s can match t since t is an application and s is not.
 *)
| (s, TAPP(_,_,_)) ->
  fail (fun () ->
    Format.print_string "since"; displayT s; Format.print_space();
    Format.print_string "and"; displayT t; Format.print_space();
    Format.print_string "do not match"
  )
(*
 * Subtyping of records (new fields can only be added on the right)
 * (Note that types of the form (... with ...) are always converted
 * to record types by expose, so we do not need an explicit clause
 * for them.)
 *)
| (RECORD ls, RECORD lt) ->
  (let rec loop c i ls lt = match (ls,lt) with
     (_, []) -> ()
   | ([],_) ->
       fail (fun()->
         Format.print_string "since"; displayT t; Format.print_space();
         Format.print_string "has more fields than"; displayT s
       )
   | (ANON(ss) :: ls,ANON(tt) :: lt) ->
       (leq c ss tt; loop c (i+1) ls lt)
   | ((LABELED(ns,ss) :: ls), (LABELED(nt,tt) :: lt)) ->
        if ns <> nt then
          fail (fun()->
            Format.print_string "since the field "; 
            Format.print_string ns; Format.print_string " in";
            displayT s; Format.print_space();
            Format.print_string "is matched by ";
            Format.print_string nt; Format.print_string " in";
            displayT t)
        else
          (leq c ss tt; loop c (i+1) ls lt)
   | (TY(ns,tags,EQ ss) :: ls,TY(nt,tagt,EQ tt) :: lt) ->
       let ks = kindOf ss and kt = kindOf tt in
       if ks = kt then
	 (leq c ss tt; leq c tt ss;
	  let c = constr(constr c tags (EQ ss)) tagt (EQ(TVAR(ks,ns,tags))) in
	  loop c (i+1) ls lt)
       else
	 fail (fun () ->
	   Format.print_string "since the kinds "; Kind.print (kindOf ss);
	   Format.print_string " and "; Kind.print (kindOf tt);
	   Format.print_string " of the types";
	   displayT ss; Format.print_space();
	   Format.print_string "and";
	   displayT tt; Format.print_space();
	   Format.print_string "are not identical")
   | (TY(ns,tags,EQ ss) :: ls,TY(nt,tagt,LT tt) :: lt) ->
       let ks = kindOf ss and kt = kindOf tt in
       if ks = kt then
	 (leq c ss tt;
	  (*
	   * the following line, which implements the contravariant
	   * half of the "Kernel Fun" rule for comparing quantified types,
	   * is omitted in the current version because the widget library
	   * seems to need the less restrictive FSub rule
	   * leq c b2 b1;
	   *)
	  let c = constr(constr c tags (EQ ss)) tagt (EQ(TVAR(ks,ns,tags))) in
	  loop c (i+1) ls lt)
       else
	 fail (fun()->
	   Format.print_string "since the kinds "; Kind.print (kindOf ss);
	   Format.print_string " and "; Kind.print (kindOf tt);
	   Format.print_string " of the bounds";
	   displayT ss; Format.print_space();
	   Format.print_string "and";
	   displayT tt; Format.print_space();
	   Format.print_string "are not identical")
   | (TY(ns,tags,LT ss) :: ls,TY(nt,tagt,LT tt) :: lt) ->
       let ks = kindOf ss and kt = kindOf tt in
       if ks = kt then
	 (leq c ss tt;
	  (*
	   * the following line, which implements the contravariant
	   * half of the "Kernel Fun" rule for comparing quantified types,
	   * is omitted in the current version because the widget library
	   * seems to need the less restrictive FSub rule
	   * leq c b2 b1;
	   *)
	  loop (makeEqual c ks ns tags kt nt tagt (LT ss)) (i+1) ls lt)
       else
	 fail (fun()->
	   Format.print_string "since the kinds "; Kind.print (kindOf ss);
	   Format.print_string " and "; Kind.print (kindOf tt);
	   Format.print_string " of the bounds";
	   displayT ss; Format.print_space();
	   Format.print_string "and";
	   displayT tt; Format.print_space();
	   Format.print_string "are not identical")
   | (_,_) ->
       fail (fun()->
         Format.print_string "since the field at position ";
	 Format.print_int i; Format.print_string " in";
         displayT s; Format.print_string " does not match that in";
	 displayT t
       )
  in loop c 0 ls lt)
(*
 * If none of the above cases apply, then the types s and t
 * have different shapes and we fail.
 *)
| (s,t) -> fail (fun () ->
    Format.print_string "since"; displayT s; Format.print_space();
    Format.print_string "and"; displayT t; Format.print_space();
    Format.print_string "do not match"
  )

(*
 * Top level leq functions, which catch the Error exception and
 * format a nice message.
 *)

let isLeq fi c s t f =
  try leq c s t with Error(why) ->
  Error.errfAt fi (fun () ->
    Format.open_hovbox 0; f(); Format.close_box();
    Format.print_space(); Format.open_hovbox 0;
    Format.open_hovbox 0; why(); Format.close_box()
  )

let leqConstr fi c c1 c2 err =
  if Types.kindOfConstr c1 = Types.kindOfConstr c2 then
    match (c1,c2) with
      (Types.EQ t1, Types.EQ t2) -> isLeq fi c t1 t2 err; isLeq fi c t2 t1 err
    | (Types.LT t1, Types.LT t2) -> isLeq fi c t1 t2 err
    | (Types.EQ t1, Types.LT t2) -> isLeq fi c t1 t2 err
    | (Types.LT _,Types.EQ _) ->
        Error.errAt fi "An equality constraint was required"
  else
    Error.errfAt fi (fun () ->
      Format.print_string "The constraint ";
      Types.printConstr c1; Format.print_string " has kind ";
      Kind.print (Types.kindOfConstr c1); Format.print_space();
      Format.print_string "but the constraint ";
      Types.printConstr c2; Format.print_string " has kind ";
      Kind.print (Types.kindOfConstr c2)
    )
