(*
 * Typechecking for the pi-calculus before applying the compositional 
 * translation and generating the translation context.
 *)

(*

TO DO: 

- checking "Attempt to rebind top-level identifier" not working inside 
  agents (the last definition of channel/function is always taken).

*)


open Enco
open Unix

(* 
 * Agent names for coding the agent input (old calculus - can be removed?) 
 *)
let thisAgentInfo = ref (Error.unknown : Error.info)
let thisAgentName  = ref ("$TOP_AGENT" : string)
let thisAgentMsgType  = ref (Types.TVAR(Kind.TYPE,"$TOP_AGENT_MSG_TYPE",Types.freshTag()))

(*
 * Store temporarily when typechecking the translation definitions
 *               (some of that is not supported now and could be removed).  
 *)
let encodedAgentTypeX    = ref ("$TOP_TYPE_AGENT__X" : string)
let encodedAgentTypeXtag = ref ((Types.freshTag()) : Types.tag)
let encodedAgentTypeXid  = ref ((fresh "$TOP_TYPE_AGENT__X") : id) 

let encodedSiteType   = ref Types.SITE
let encodedDaemonType = ref (Types.AGENTT(Types.TOP))
let encodedAgentType  = ref (Types.AGENTT(Types.TOP))
let encodedParamType  = ref (Types.RECORD([]))
let encodedProgramParamType  = ref (Types.RECORD([]))


(* 
 * Utilities for names. 
 *)
let findVId fi c s =  (* this function is a hack; an elegant solution 
			 would require changes in data structures 
			 generated by the typechecking *)
  let retId x =
    let n = Var.str_of x in
    let base = Var.getModuleName() in
    try
      if (String.sub n 0 (String.length base) <> base) then 
	createId s (int_of (fresh ""))  (* retain the original name *)
      else 
	createId n (Var.int_of x)  (* take the internal name *)    
    with 
      Invalid_argument _ -> createId s (int_of (fresh ""))
  in
  let (_,a) = Types.typeOfX fi s c in 
  match a with
    Inter.DYNAMIC(x) -> retId x
  | Inter.STATIC(x) -> retId x
  | Inter.INT(_) -> createId s (int_of (fresh ""))  (* temporary hack *)
    (* For debugging:
       | Inter.INT(_) -> Error.bugAt fi "Precheck.findVId: found INT"
       | Inter.CHAR(_) -> Error.bugAt fi "Precheck.findVId: found CHAR"
       | Inter.BOOL(_) -> Error.bugAt fi "Precheck.findVId: found BOOL"
       | Inter.ADDR(_) -> Error.bugAt fi "Precheck.findVId: found ADDR"
       | Inter.PROJECT(a,i) -> Error.bugAt fi "Precheck.findVId: found PROJECT"
       | Inter.CCONST(_) -> Error.bugAt fi "Precheck.findVId: found CCONST"
    *)
  | _ -> Error.bugAt fi "Precheck.findVId: Expected DYNAMIC or STATIC"
	
(* 
 * Flags and error messages.
 *)

let check = Flags.createFlag "_check" "*Debug typechecker" false

let verbose = Flags.createFlag "_CHECK" "*Debug typechecker verbosely" false

let assertions =
  Flags.createFlag "_assertions"
  "*Check all assertions made in the library code" false

let checks =
  Flags.createFlag "_checks"
  "*Do all runtime checks specified in the library code" true

let debug = Flags.createFlag "_debug"
  "*Include debugging code in libraries" false

let ochErr () =
  Format.print_string "expected an output-channel type"
let ichErr () =
  Format.print_string "expected an input-channel type"

let atErr () =
  Format.print_string "expected a Loc"
let agentErr () =
  Format.print_string "expected an Agent"
let siteErr () =
  Format.print_string "expected a Site"
let timeoutErr () =
  Format.print_string "expected a timeout value of type integer"

let patErr () =
  Format.print_string "pattern's type does not match type of channel"

let matchErr () =
  Format.print_string "expected type does not match actual"

let asErr () =
  Format.print_string
  "Pattern type is not a subtype of the given explicit type"
let dynamicErr () =
  Format.print_string
  "argument of dynamic does not have required type"
let recpErr () =
  Format.print_string
  "Recursive pattern type does not match explicitly given type"
let recordpErr () =
  Format.print_string
  "Record pattern type does not match annotation"
let outputErr () =
  Format.print_string "value's type does not match type of channel"
let typecaseErr () =
  Format.print_string "argument to typecase is not a Dyn"
let caseErr () =
  Format.print_string "argument to case is not a variant"
let typecasevErr () =
  Format.print_string "branch of typecase does not have expected type"
let casevErr () =
  Format.print_string "branch of case does not have expected type"
let ifErr () =
  Format.print_string "argument to conditional is not a boolean"
let thenElseErr () =
  Format.print_string
  "minimal types of 'then' and 'else' branches must be the same"
let appErr () =
  Format.print_string "argument has incorrect type"
let recordErr () =
  Format.print_string "expected a record type containing the field "
let valPatErr () =
  Format.print_string "value does not match type of pattern"
let patValErr () =
  Format.print_string "pattern does not match type of value"
let foldErr () =
  Format.print_string "ill-typed fold"
let newErr () =
  Format.print_string "declared type is not a channel type"
let abbrvErr () =
  Format.print_string "type abbreviation fields are not identical"
let boundErr () =
  Format.print_string "bounds do not match on type fields"

exception NeedTypeAnnotation of Error.info

let printReqd = function
  None -> ()
| Some(t,_) -> Format.print_string "required type: "; Types.printT t

let checkReqd fi c t = function
  None -> t
| Some(reqdt,err) -> Leq.isLeq fi c t reqdt err; reqdt

let getReqd fi = function
  None -> Error.errAt fi "Type annotation required"
| Some(t,_) -> t

let getProvidedType fi c ec topt reqd =
  match (topt,reqd) with
    (None,None) -> raise (NeedTypeAnnotation fi)
  | (Some t,None) -> let (t,tt,ec) = Prekinding.hasKindType c ec t in (t,Some tt,ec)
  | (None,Some(t,_)) -> (t,None,ec)
  | (Some t,Some(treqd,err)) ->
      let (t,tt,ec) = Prekinding.hasKindType c ec t in 
      Leq.isLeq fi c treqd t err; (t,Some tt,ec)

let silentGetProvidedType fi c ec topt reqd =
  match (topt,reqd) with
    (None,None) -> (Types.TOP,Some (TOP(fi)),ec)
  | (Some t,None) -> let (t,tt,ec) = Prekinding.hasKindType c ec t in (t,Some tt,ec)
  | (None,Some(t,_)) -> (t,None,ec)
  | (Some t,Some(treqd,err)) ->
      let (t,tt,ec) = Prekinding.hasKindType c ec t in 
      Leq.isLeq fi c treqd t err; (t,Some tt,ec)

let rec checkPat' topLevel c ec reqd getProvidedType = function
    Syntax.RECORDp(fi,l) ->
      (match reqd with
      	None ->
          let (l,tl,inc,ec) = synthFieldPat topLevel c ec 0 l in
          (RECORDp(fi,l), Prekinding.mkRecord fi tl, inc, ec)
      | Some(reqdt,err) ->
          let tl = Types.getFields fi c reqdt in
          let (l,inc,ec) = checkFieldPat topLevel fi c ec 0 l tl in
          (RECORDp(fi,l), Prekinding.mkRecord fi tl, inc, ec))
  | Syntax.ASp(fi,n,topt,pat) ->
      let ((pat,t,inc,ec),ttopt) =
      	match (reqd,topt) with
          (None,None) -> (checkPat topLevel c ec None getProvidedType pat, None)
      	| (_,_) ->
	    let (t,ttopt,ec) = getProvidedType fi c ec topt reqd in
	    (checkPat topLevel c ec (Some(t,asErr)) getProvidedType pat, ttopt)
      in
      if topLevel && Types.isBoundVar c n then
      	Error.errAt fi "Attempt to rebind top-level identifier"
      else
  	let nid = Enco.fresh n in
  	let x = Var.createVar (str_of nid) (int_of nid) in
        (ASp(fi,nid,ttopt,pat), t, 
	 Types.union (Types.xbind n t (Inter.DYNAMIC x) Types.empty) inc,
	 Enco.xbind nid t (Inter.DYNAMIC x) ec)
  | Syntax.VARp(fi,n,topt) ->
      let (t,ttopt,ec) = getProvidedType fi c ec topt reqd in
      if topLevel && Types.isBoundVar c n then
      	Error.errAt fi "Attempt to rebind top-level identifier"
      else
   	let nid = Enco.fresh n in
        let x = Var.createVar (str_of nid) (int_of nid) in
      	(VARp(fi,nid,ttopt), t, 
	 Types.xbind n t (Inter.DYNAMIC x) Types.empty,
	 Enco.xbind nid t (Inter.DYNAMIC x) ec)
  | Syntax.WILDp(fi,topt) ->
      let (t,ttopt,ec) = getProvidedType fi c ec topt reqd in
      (WILDp(fi,ttopt),t,Types.empty,ec)
  | Syntax.RECp(fi,topt,pat) ->
      let (t,ttopt,ec) = getProvidedType fi c ec topt reqd in
      let reqd = Some(Types.unroll fi c t,recpErr) in
      let (pat,_,inc,ec) = checkPat topLevel c ec reqd getProvidedType pat in
      (RECp(fi,ttopt,pat),t,inc,ec)
  | Syntax.ENCODE_PAT(fi,l) ->
      (match l with
      	(n::[]) ->
          let (t,_,ec) = getProvidedType fi c ec None reqd in
      	  (ENCODE_PAT(fi,findVId fi c n), t, Types.empty, ec)  (* Bogus? *)
      | _ -> Error.errAt fi "Wrong arguments in ENCODE PAT")

and synthFieldPat topLevel c ec i = function
    [] ->
      ([],[],Types.empty,ec)
  | Syntax.ANONp(fi,pat) :: l ->
      let (pat,t,inc,ec) = checkPat topLevel c ec None getProvidedType pat in
      let (l,tl,incs,ec) = synthFieldPat topLevel c ec (i+1) l in
      (ANONp(fi,pat) :: l, Types.ANON t :: tl,Types.union inc incs,ec)
  | Syntax.LABELEDp(fi,n,pat) :: l ->
      let (pat,t,inc,ec) = checkPat topLevel c ec None getProvidedType pat in
      let (l,tl,incs,ec) = synthFieldPat topLevel c ec (i+1) l in
      (LABELEDp(fi,n,pat) :: l, Types.LABELED(n,t) :: tl,Types.union inc incs,ec)
  | Syntax.TYp(fi,n,Syntax.LT ty) :: l ->
      let tag = Types.freshTag() in let (t,tt,ec) = Prekinding.wellKinded c ec ty in
      let nid = Enco.fresh n in
      let ec = Enco.ibind ec tag nid in
      let (l,tl,incs,ec) =
      	synthFieldPat topLevel
	  (Types.abind (Types.constr c tag (Types.LT t)) n tag)
	  (Enco.abind (Enco.constr ec tag (Types.LT t)) nid tag)
	  i l
      in
      if topLevel && Types.isBoundTyVar c n then
	Error.errAt fi "Attempt to rebind top-level type identifier"
      else
    	(TYp(fi,nid,LT tt) :: l, 
	 Types.TY(n,tag,Types.LT t) :: tl,
	 Types.abind (Types.constr incs tag (Types.LT t)) n tag,
         Enco.ibind (Enco.abind (Enco.constr ec tag (Types.LT t)) nid tag) tag nid)  (* or just ec ? *)
  | Syntax.TYp(fi,n,Syntax.EQ ty) :: l ->
      let tag = Types.freshTag() and (t,tt,ec) = Prekinding.wellKinded c ec ty in
      let nid = Enco.fresh n in
      let ec = Enco.ibind ec tag nid in
      let (l,tl,incs,ec) =
      	synthFieldPat topLevel
	  (Types.abind (Types.constr c tag (Types.EQ t)) n tag)
	  (Enco.abind (Enco.constr ec tag (Types.EQ t)) nid tag)
	  (i+1) l
      in
      if topLevel && Types.isBoundTyVar c n then
	Error.errAt fi "Attempt to rebind top-level type identifier"
      else
	(TYp(fi,nid,EQ tt) :: l, 
	 Types.TY(n,tag,Types.EQ t) :: tl,
	 Types.abind (Types.constr incs tag (Types.EQ t)) n tag,
         Enco.ibind (Enco.abind (Enco.constr ec tag (Types.EQ t)) nid tag) tag nid)  (* or just ec ? *)

and checkFieldPat topLevel fi c ec i l tl = match (l,tl) with
  ([],[]) -> ([],Types.empty,ec)
| (Syntax.ANONp(fi,pat)::l, Types.ANON(t)::tl) ->
   (match t with 
      Types.IGNORE -> 
        let (pat,_,inc,ec) = checkPat topLevel c ec (Some(t,matchErr)) getProvidedType pat in
	let (l,incs,ec) = 
          (match l with
             [] -> checkFieldPat topLevel fi c ec (i+1) l []
           | _ -> checkFieldPat topLevel fi c ec (i+1) l (Types.ANON(Types.IGNORE)::[]))
        in
        (ANONp(fi,pat) :: l, Types.union inc incs, ec)
    | _ ->
      let (pat,_,inc,ec) = checkPat topLevel c ec (Some(t,matchErr)) getProvidedType pat in
      let (l,incs,ec) = checkFieldPat topLevel fi c ec (i+1) l tl in
      (ANONp(fi,pat) :: l, Types.union inc incs, ec)
    )
| (Syntax.LABELEDp(fi,n,pat)::l, Types.LABELED(reqdn,reqdt)::tl) ->
    if n<>reqdn then Error.errfAt fi (fun () ->
      Format.print_string "Expected label "; Format.print_string reqdn;
      Format.print_string " but found "; Format.print_string n);
    let (pat,_,inc,ec) = checkPat topLevel c ec (Some(reqdt,matchErr)) getProvidedType pat in
    let (l,incs,ec) = checkFieldPat topLevel fi c ec (i+1) l tl in
    (LABELEDp(fi,n,pat) :: l, Types.union inc incs, ec)
| (Syntax.TYp(fi,n,constrain)::l, Types.TY(_,tag,reqdConstr)::tl) ->
    let (constr,con,ec) = Prekinding.wellKindedConstr c ec constrain in
    Leq.leqConstr fi c reqdConstr constr boundErr;
    let c = Types.abind (Types.constr c tag constr) n tag in
    let nid = Enco.fresh n in
    let ec = Enco.abind (Enco.constr ec tag constr) nid tag in
    let ec = Enco.ibind ec tag nid in
    let (l,incs,ec) = checkFieldPat topLevel fi c ec i l tl in
    (TYp(fi,nid,con) :: l, 
     Types.abind (Types.constr incs tag constr) n tag, 
     Enco.ibind (Enco.abind (Enco.constr ec tag constr) nid tag) tag nid)  (* or just ec ? *)
| (_,_) ->
    Error.errAt fi "Field mismatch in record pattern"
  
and checkPat topLevel c ec reqd f pat =
  if !check then
    Debug.wrap "checkPat"
      (fun () -> checkPat' topLevel c ec reqd f pat)
      (fun () -> Error.printInfo (Syntax.patInfo pat); printReqd reqd)
      (fun (p,t,inc,ec) -> Types.printT t; if !verbose then Types.printC inc)
  else
    checkPat' topLevel c ec reqd f pat

let safeCheckPat topLevel c ec reqd pat =
  try checkPat topLevel c ec reqd getProvidedType pat with NeedTypeAnnotation(fi) ->
    Error.errAt fi "Type annotation required"

let leqReqd fi c ec topt reqd =
  match (topt,reqd) with
    (None,None) -> (None,None,ec)
  | (Some t,None) -> 
      let (t,tt,ec) = Prekinding.hasKindType c ec t in (Some(t,matchErr),Some tt,ec)
  | (None,Some(_,_)) -> (reqd,None,ec)
  | (Some t,Some(treqd,err)) ->
      let (t,tt,ec) = Prekinding.hasKindType c ec t in
      Leq.isLeq fi c t treqd err; (Some(t,err),Some tt,ec)
	
let rec checkProc' c ec = function
  Syntax.PRL(fi,p,q) ->
    PRL(fi,checkProc c ec p,checkProc c ec q)
| Syntax.INPUT(fi,v,pat,p) ->
    let (v,t) = checkVal c ec None v in
    let argTy = Types.getIChan fi c t in
    let (pat,_,inc,ec) = safeCheckPat false c ec (Some(argTy,matchErr)) pat in
    INPUT(fi,v,pat,checkProc (Types.union c inc) ec p)
| Syntax.OUTPUT(fi,v,av) ->
    let (v,t) = checkVal c ec None v in
    let argTy = Types.getOChan fi c t in
    (*
    if Types.leqRChan c t then
      OUTPUT(RESPONSIVE,v,fst(checkVal c (Some(argTy,matchErr)) av))
    else
      OUTPUT(UNKNOWN,v,fst(checkVal c (Some(argTy,matchErr)) av))
    *)
    OUTPUT(fi,v,fst(checkVal c ec (Some(argTy,matchErr)) av))
| Syntax.INPA(fi,pat,p) ->
    let (v,t) = checkVal c ec None (Syntax.VARv(fi,!thisAgentName)) in
    let argTy = Types.getIAgent fi c t in
    let (pat,_,inc,ec) = safeCheckPat false c ec (Some(argTy,matchErr)) pat in
    INPA(fi,v,pat,checkProc (Types.union c inc) ec p)
| Syntax.OUTA(fi,v,av) ->
    let (v,t) = checkVal c ec None v in
    let argTy = Types.getOAgent fi c t in
    OUTA(fi,v,fst(checkVal c ec (Some(argTy,matchErr)) av))
| Syntax.LDOUTA(fi,v,sv,av) ->
    let (v,t) = checkVal c ec None v in
    let argTy = Types.getOAgent fi c t in
    let (sv,_)= checkVal c ec (Some(Types.SITE,siteErr)) sv in
    (*
    if Types.leqRChan c t then
      LDOUTA(UNKNOWN,v,sv,fst(checkVal c (Some(argTy,matchErr)) av))
    else
      LDOUTA(UNKNOWN,v,sv,fst(checkVal c (Some(argTy,matchErr)) av))
    *)
    LDOUTA(fi,v,sv,fst(checkVal c ec (Some(argTy,matchErr)) av))
| Syntax.TRYHERE(fi,v,av,p,q) ->
    let (v,t) = checkVal c ec None v in
    let argTy = Types.getOAgent fi c t in
    (*
    if Types.leqRChan c t then
      TRYHERE(v,fst(checkVal c (Some(argTy,matchErr)) av),checkProc c p,checkProc c q)
    else
      TRYHERE(v,fst(checkVal c (Some(argTy,matchErr)) av),checkProc c p,checkProc c q)
    *)
    TRYHERE(fi,v,fst(checkVal c ec (Some(argTy,matchErr)) av),checkProc c ec p,checkProc c ec q)
| Syntax.REP_INPUT(fi,v,pat,p) ->
    let (v,t) = checkVal c ec None v in
    let argTy = Types.getIChan fi c t in
    let (pat,t,inc,ec) = safeCheckPat false c ec (Some(argTy,matchErr)) pat in
    REP_INPUT(fi,v,pat,checkProc (Types.union c inc) ec p)
| Syntax.TIMED_INPUT(fi,w,v,pat,p,q) ->
    let (w,_) = checkVal c ec (Some(Types.INT,timeoutErr)) w in
    let (v,t) = checkVal c ec None v in
    let argTy = Types.getIChan fi c t in
    let (pat,_,inc,ec) = safeCheckPat false c ec (Some(argTy,matchErr)) pat in
    TIMED_INPUT(fi,w,v,pat,checkProc (Types.union c inc) ec p,checkProc c ec q)
| Syntax.REP_TIMED_INPUT(fi,w,v,pat,p,q) ->
    let (w,_) = checkVal c ec (Some(Types.INT,timeoutErr)) w in
    let (v,t) = checkVal c ec None v in
    let argTy = Types.getIChan fi c t in
    let (pat,_,inc,ec) = safeCheckPat false c ec (Some(argTy,matchErr)) pat in
    REP_TIMED_INPUT(fi,w,v,pat,checkProc (Types.union c inc) ec p,checkProc c ec q)
| Syntax.LI_SEND(fi,v,chv,av) ->
    let (v,_) = checkVal c ec (Some(Types.AGENT,agentErr)) v in
    let (chv,cht) = checkVal c ec None chv in
    let argTy = Types.getOChan fi c cht in
    LI_SEND(fi,v,chv,fst(checkVal c ec (Some(argTy,matchErr)) av))
| Syntax.REMOTE_SEND(fi,v,sv,chv,av) ->
    let (v,_) = checkVal c ec (Some(Types.AGENT,agentErr)) v in
    let (sv,_)= checkVal c ec (Some(Types.SITE,siteErr)) sv in
    let (chv,cht) = checkVal c ec None chv in
    let argTy = Types.getOChan fi c cht in
    REMOTE_SEND(fi,v,sv,chv,fst(checkVal c ec (Some(argTy,matchErr)) av))
| Syntax.LOCAL_SEND(fi,v,chv,av) ->
    let (v,_) = checkVal c ec (Some(Types.AGENT,agentErr)) v in
    let (chv,cht) = checkVal c ec None chv in
    let argTy = Types.getOChan fi c cht in
    LOCAL_SEND(fi,v,chv,fst(checkVal c ec (Some(argTy,matchErr)) av))
| Syntax.IFLOCAL(fi,v,chv,av,p,q) ->
    let (v,_) = checkVal c ec (Some(Types.AGENT,agentErr)) v in
    let (chv,cht) = checkVal c ec None chv in
    let argTy = Types.getOChan fi c cht in
    IFLOCAL(fi,v,chv,fst(checkVal c ec (Some(argTy,matchErr)) av),checkProc c ec p,checkProc c ec q)
| Syntax.SKIP(fi) ->
    SKIP(fi)
| Syntax.LET(fi,d,p) ->
    let (d,inc,ec) = checkDec false c ec d in
    LET(fi,d,checkProc (Types.union c inc) ec p)
| Syntax.IF(fi,v,p,q) ->
    let (v,_) = checkVal c ec (Some(Types.BOOL,ifErr)) v in
    IF(fi,v,checkProc c ec p,checkProc c ec q)
| Syntax.TERMINATE(fi) ->
    TERMINATE(fi)
| Syntax.TYPECASE(fi,v,ts,p) ->
    let (v,_) = checkVal c ec (Some(Types.DYN,typecaseErr)) v in
    let ts = List.map
               (fun (pati,pi) ->
                  let (pati,ti,inci,ec) = safeCheckPat false c ec (None) pati in
                  let pi = checkProc (Types.union c inci) ec pi in
                  (pati,pi))
               ts in
    let p = checkProc c ec p in
    TYPECASE(fi,v,ts,p)
| Syntax.CASE(fi,topt,v,ts) ->
    let ((v,t),ttopt,ec) =
      match topt with
  	None -> (checkVal c ec None v, None, ec)
      | Some t -> 
  	  let (t,tt,ec) = Prekinding.hasKindType c ec t in 
  	  (checkVal c ec (Some(t,matchErr)) v, Some tt, ec)
    in
    let ts = List.map
        (fun (idi,pati,pi) ->
          let (pati,_,inci,ec) = safeCheckPat false c ec None pati in
                     (* Bogus: Some(Types.VART(..),caseErr) instead of None *)
	  let pi = checkProc (Types.union c inci) ec pi in
                  (*
                     let x = Var.fresh() in
                     (Var.createVar idi (Var.snd x),pati,pi)
		  *)
          (idi,pati,pi))
   	ts in
    CASE(fi,ttopt,v,ts)
| Syntax.ENCODE_PROC(fi,l) -> (match l with
    (n::pa::[]) ->
      ENCODE_PROC(fi,findVId fi c n,findVId fi c pa)
  | _ -> Error.errAt fi "Wrong arguments in ENCODE PROC")

and checkVal' c ec reqd = function
  Syntax.CCODEv(fi,ci,l) ->
    let (l,tl) = synthFieldVal c ec l in
    (match ci.Ccode.kind with
       Ccode.EXP ->
         (CCODEv(fi,ci,l), getReqd fi reqd)
         (*
	 (match ci.Ccode.code with
	   ["%async"] ->
	      let async =
		try Sys.getenv "PICTASYNC" = "yes" with Not_found -> false
	      in
		(BOOLv async, checkReqd fi c Types.BOOL reqd)
	 | ["%assertions"] ->
	     (BOOLv (!assertions), checkReqd fi c Types.BOOL reqd)
	 | ["%checks"] ->
	     (BOOLv (!checks), checkReqd fi c Types.BOOL reqd)
	 | ["%debug"] ->
	     (BOOLv (!debug), checkReqd fi c Types.BOOL reqd)
         | _ ->
	   (CCODEv(ci,l), getReqd fi reqd))
            *)
     | Ccode.STATEMENT ->
         (CCODEv(fi,ci,l), checkReqd fi c Types.mkUnit reqd)
     | Ccode.NEXP ->
	 (CCODEv(fi,ci,l), getReqd fi reqd)
     | Ccode.NSTATEMENT ->
         (CCODEv(fi,ci,l), checkReqd fi c Types.mkUnit reqd))
| Syntax.RECORDv(fi,l) ->
    (match reqd with
       Some(t,_) ->
         let (_,l) =
	   checkFieldVal fi c ec l (Types.getFields fi c (Types.clone t)) false
           (fun _ _ l tl ->
	      match (l,tl) with
	        ([],[]) -> Types.mkUnit
	      | (l, []) ->
		  Error.warningAt fi "Record has extra fields";
		  ignore (synthFieldVal c ec l); Types.mkUnit
              | ([], _) -> Error.errAt fi "Record does not have enough fields"
	      | (fv :: _,_) -> Error.errAt (Syntax.fvInfo fv)
		"Record field mismatch")
	 in (RECORDv(fi,l),t)
     | None ->
	 let (l,tl) = synthFieldVal c ec l in
	 (RECORDv(fi,l), Prekinding.mkRecord fi tl))
| Syntax.WITHv(fi,topt,v,l) ->
    let ((v,t),ttopt,ec) =
      match topt with
        None -> (checkVal c ec None v, None,ec)
      | Some t -> 
	  let (t,tt,ec) = Prekinding.hasKindType c ec t in 
 	  (checkVal c ec (Some(t,matchErr)) v, Some tt,ec)
    in let fields = Types.getFields fi c t in
    let (l,tl) = synthFieldVal c ec l in
    (*
    let var = Var.fresh() in
    let rec loop i accum =
      if i < 0 then
	LETv(VAL(VARp(var,BOXED),v),RECORDv accum)
      else
	loop (i-1) (ANONv(SELECTv(VARv(Inter.DYNAMIC var),i)) :: accum)
    in
      (loop (List.length fields - 1) l, Prekinding.mkRecord fi (fields @ tl))
    *)
    (WITHv(fi,ttopt,v,l), Prekinding.mkRecord fi (fields @ tl))
| Syntax.WHEREv(fi,topt,v,l) ->
    let ((v,t),ttopt,ec) =
      match topt with
        None -> (checkVal c ec None v, None, ec)
      | Some t -> 
	  let (t,tt,ec) = Prekinding.hasKindType c ec t in 
	  (checkVal c ec (Some(t,matchErr)) v, Some tt, ec)
    in let fields = Types.getFields fi c t in
    let (l,tl) = synthFieldVal c ec l in
(*
    let rcd = Var.fresh() in
    let (d,map) =
      List.fold_left2 (fun (d,map) fv ft ->
	(match ft with
	  Types.LABELED(n,ty) ->
	    let var = Var.fresh() in
	    let (_,i) = Types.fieldType fi c t n in
	    (match fv with
	       ANONv(v) ->
	         (SEQ(d,VAL(VARp(var,BOXED),v)),
		  Misc.IntMap.add i (ANONv(VARv(Inter.DYNAMIC var)),ft) map)
	     | TYv(_) ->
	         Error.bugAt fi "Precheck.checkVal: WHERE")
	| Types.ANON(_) -> Error.errAt fi "Expected a labeled value field"
	| Types.TY(_) -> Error.errAt fi "Expected a labeled value field")
      ) (VAL(VARp(rcd,BOXED),v),Misc.IntMap.empty) l tl
*)
    let rcd = Enco.fresh "DUMMY" in
    let (d,map) =
      List.fold_left2 (fun (d,map) fv ft ->
	(match ft with
	  Types.LABELED(n,ty) ->
	    let var = Enco.fresh "DUMMY" in
	    let (_,i) = Types.fieldType fi c t n in
	    (match fv with
	      ANONv(fi,v) ->
	        (SEQ(fi,d,VAL(fi,VARp(fi,var,None),v)),
		 Misc.IntMap.add i (ANONv(fi,VARv(fi,var)),ft) map)
	    | TYv(fi,_) ->
	        Error.bugAt fi "Precheck.checkVal: WHERE"
	    | LABELEDv(fi,_,_) ->
	        Error.bugAt fi "Precheck.checkVal: WHERE")
	| Types.ANON(_) -> Error.errAt fi "Expected a labeled value field"
	| Types.TY(_) -> Error.errAt fi "Expected a labeled value field")
	  ) (VAL(fi,VARp(fi,rcd,None),v),Misc.IntMap.empty) l tl
    (*
    in let (_,l,tl) =
      List.fold_right (fun ft (i,l,tl) ->
        let (fv,ft) =
	  try Misc.IntMap.find i map with Not_found ->
	  (ANONv(SELECTv(VARv(Inter.DYNAMIC rcd),i)), ft)
	in
	  (i-1, fv :: l, ft :: tl)
      ) fields (List.length fields-1,[],[])
    in
      (LETv(d,RECORDv l), checkReqd fi c (Types.RECORD tl) reqd)
    *)
    in let (_,l',tl) =
      List.fold_right (fun ft (i,l,tl) ->
        let (fv,ft) =
	  try Misc.IntMap.find i map with Not_found ->
	    (ANONv(fi,SELECTv(fi,VARv(fi,rcd),(string_of_int i))), ft)
	in
	(i-1, fv :: l, ft :: tl)
	  ) fields (List.length fields-1,[],[])
    in
    (WHEREv(fi,ttopt,v,l'), checkReqd fi c (Types.RECORD tl) reqd)
| Syntax.APPv(fi,topt,v,al) ->
    let (v,t) = checkVal c ec None v in
    let (res,l) =
      checkFieldVal fi c ec al
      	(Types.getFields fi c (Types.clone (Types.getOChan fi c t))) false
      	(fun c ec l tl ->
	  match (l,tl) with
  	    ([],[Types.ANON t]) ->
	      (let (tyopt,ttopt,ec) = leqReqd fi c ec topt reqd in
	      match tyopt with
	       	None -> Types.getOChan fi c t
	      | Some(reqdt,err) ->
		  Leq.isLeq fi c (Types.getOChan fi c t) reqdt err; reqdt)
          | ([],_ :: _ :: _) -> Error.errAt fi "Too few arguments"
          | ([],_ :: _) -> Error.errAt fi "Illegal function type"
          | ([],[]) -> Error.errAt fi "Too many arguments"
          | (fv :: _,[]) -> Error.errAt (Syntax.fvInfo fv) "Too many arguments"
          | (fv :: _,_ :: _) -> Error.errAt (Syntax.fvInfo fv)
	   	"Record field mismatch")
    in 
    (*
    let r = if Types.leqRChan c t then RESPONSIVE else UNKNOWN 
    in
      (APPv(r,v,l), res)
    *)
    let ttopt = Prekinding.scanOptT c ec topt in
    (APPv(fi,ttopt,v,l), res)
| Syntax.SELECTv(fi,v,n) ->
    let (v,t) = checkVal c ec None v in
    let (ft,i) = Types.fieldType fi c t n in
    (SELECTv(fi,v,n), checkReqd fi c ft reqd)
| Syntax.STRINGv(fi,s) ->
    (STRINGv(fi,s), checkReqd fi c Types.STRING reqd)
| Syntax.SITEv(fi,s,i) ->
    let s = try 
      if (s = "localhost") 
        then string_of_inet_addr((gethostbyname(gethostname())).h_addr_list.(0))
        else string_of_inet_addr((gethostbyname(s)).h_addr_list.(0))
    with Not_found -> Error.errAt fi ("Unknown IP address " ^ s)
    in
    (SITEv(fi,s,i), checkReqd fi c Types.SITE reqd)
| Syntax.CHARv(fi,x) ->
    (CHARv(fi,x), checkReqd fi c Types.CHAR reqd)
| Syntax.VARv(fi,x) ->
    let (t,a) = Types.typeOfX fi x c in
    (VARv(fi,findVId fi c x), checkReqd fi c t reqd)
| Syntax.BOOLv(fi,b) ->
    (BOOLv(fi,b), checkReqd fi c Types.BOOL reqd)
| Syntax.INTv(fi,i) ->
    (INTv(fi,i), checkReqd fi c Types.INT reqd)
| Syntax.LETv(fi,d,v) ->
    let (d,inc,ec) = checkDec false c ec d in
    let (v,t) = checkVal (Types.union c inc) ec reqd v in
    (match reqd with
       None ->
         Error.warningAt fi "Result type for LETv might be wrong";
	 (LETv(fi,d,v),t)
    | Some(reqdt,_) ->
	 (LETv(fi,d,v),reqdt))
| Syntax.RECv(fi,v,topt) ->
    (match leqReqd fi c ec topt reqd with
      (None,_,_) -> Error.errAt fi "Type annotation required"
    | (Some(t,_),ttopt,ec) ->
        (RECv(fi,fst(checkVal c ec (Some(Types.unroll fi c t,foldErr)) v),ttopt),t))
| Syntax.IFv(fi,topt,b,v1,v2) ->
    let (b,_) = checkVal c ec (Some(Types.BOOL,ifErr)) b in
    let (tyopt,ttopt,ec) = leqReqd fi c ec topt reqd in
    let (v1,t) = checkVal c ec tyopt v1 in
    let (v2,_) = checkVal c ec (Some(t,thenElseErr)) v2 in
    (IFv(fi,ttopt,b,v1,v2),t)
| Syntax.ABSv(fi,pat,p) ->
    (match reqd with
      None ->
        let (pat,t,inc,ec) = safeCheckPat false c ec None pat in
        (ABSv(fi,pat,checkProc (Types.union c inc) ec p),Types.RCHAN t)
    | Some(reqdt,_) ->
        let t = Types.getOChan fi c reqdt in
        let (pat,_,inc,ec) = safeCheckPat false c ec (Some(t,matchErr)) pat in
        (ABSv(fi,pat,checkProc (Types.union c inc) ec p),reqdt))
| Syntax.DYNAMICv(fi,v,topt) ->
    let ((v,t),ttopt,ec) =
      match topt with
        None -> (checkVal c ec None v,None, ec)
      | Some(t) ->
          let (t,tt,ec) = Prekinding.hasKindType c ec t in
          (checkVal c ec (Some(t,dynamicErr)) v, Some tt, ec) 
    in
    (DYNAMICv(fi,v,ttopt), Types.DYN)
| Syntax.TYPECASEv(fi,topt,v,ts,velse) ->
    let (v,_) = checkVal c ec (Some(Types.DYN,typecaseErr)) v in
    let (tyopt,ttopt,ec) = leqReqd fi c ec topt reqd in
    let (velse,t) = checkVal c ec tyopt velse in
    let ts = List.map
        (fun (pati,vi) ->
          let (pati,ti,inci,ec) = safeCheckPat false c ec None pati in
          let (vi,_) = checkVal (Types.union c inci) ec
              (Some(t,typecasevErr)) vi in
          (pati,vi))
        ts in
    (TYPECASEv(fi,ttopt,v,ts,velse), t)
| Syntax.VARIANTv(fi,n,v) ->
    let (v,t) = checkVal c ec None v in
    (VARIANTv(fi,n,v), Types.IGNORE)   (* Bogus: VART(..) instead of IGNORE *)
| Syntax.CASEv(fi,topt,v,ts) ->
    let (v,_) = checkVal c ec None v in  (* Bogus: VART(..) instead of None *)
    let ts = List.map
        (fun (idi,pati,vi) ->
          let (pati,_,inci,ec) = safeCheckPat false c ec None pati in
          let (vi,_) = checkVal (Types.union c inci) ec None vi in
                                (* Bogus: Some(Types.VART(..),casevErr) instead of None *)
          (idi,pati,vi))
        ts in
    let ttopt = Prekinding.scanOptT c ec topt in
    (CASEv(fi,ttopt,v,ts), Types.IGNORE)   (* Bogus: The actual type of values returned in cases instead of DYN *)
| Syntax.ENCODE_VALUE(fi,l) -> (match l with
    (n::[]) ->
      let (t,a) = Types.typeOfX fi n c in 
 	(match t with
	  Types.IGNORE -> (ENCODE_VALUE(fi,findVId fi c n), Types.IGNORE) (* to avoid typechecking the macro argument *)
        | _ -> (ENCODE_VALUE(fi,findVId fi c n), checkReqd fi c t reqd)  (* Bogus? *)
	)
  | _ -> Error.errAt fi "Wrong arguments in ENCODE VALUE")

and synthFieldVal c ec = function
  [] -> ([],[])
| Syntax.ANONv(fi,v) :: l ->
    let (v,t) = checkVal c ec None v in
    let (l,tl) = synthFieldVal c ec l in
    (ANONv(fi,v) :: l, Types.ANON t :: tl)
| Syntax.LABELEDv(fi,n,v) :: l ->
    let (v,t) = checkVal c ec None v in
    let (l,tl) = synthFieldVal c ec l in
    (LABELEDv(fi,n,v) :: l, Types.LABELED(n,t) :: tl)
| Syntax.TYv(fi,ty) :: l ->
    let (t,tt,ec) = Prekinding.wellKinded c ec ty in
    let (l,tl) = synthFieldVal c ec l in
    (TYv(fi,tt) :: l, Types.TY(":x",Types.freshTag(),Types.EQ t) :: tl)

and checkFieldVal fi c ec l tl synth finalise = match (l,tl) with
  (Syntax.ANONv(fi,v) :: l, Types.ANON t :: tl) ->
    if synth then
      let (v,vt) = checkVal c ec None v in Leq.isLeq fi c vt t matchErr;
      let (res,l) = checkFieldVal fi c ec l tl true finalise in
      (res, ANONv(fi,v) :: l)
    else
      let (v,_) = checkVal c ec (Some(t,matchErr)) v in
      let (res,l) = checkFieldVal fi c ec l tl false finalise in
      (res, ANONv(fi,v) :: l)
| (Syntax.LABELEDv(fi,n,v) :: l, Types.LABELED(reqdn,t) :: tl) ->
    if n<>reqdn then Error.errfAt fi (fun () ->
      Format.print_string "Expected label "; Format.print_string reqdn;
      Format.print_string " but found "; Format.print_string n);
    if synth then
      let (v,vt) = checkVal c ec None v in Leq.isLeq fi c vt t matchErr;
      let (res,l) = checkFieldVal fi c ec l tl true finalise in
      (res, LABELEDv(fi,n,v) :: l)
    else
      let (v,_) = checkVal c ec (Some(t,matchErr)) v in
      let (res,l) = checkFieldVal fi c ec l tl false finalise in
      (res, LABELEDv(fi,n,v) :: l)
| (Syntax.TYv(fi,ty) :: l, Types.TY(_,tag,constr) :: tl) ->
    let (t,tt,ec) = Prekinding.wellKinded c ec ty in
    Leq.leqConstr fi c (Types.EQ t) constr boundErr;
    let (res,l) =
      checkFieldVal fi (Types.constr c tag (Types.EQ t)) ec l tl synth finalise
    in
    (Types.subst res tag t, TYv(fi,tt) :: l)
| (_, Types.TY(_,tag,Types.LT t) :: tl) ->
    (match Types.kindOf t with
      Kind.TYPE ->
        let alpha = Types.mkUvar() in
	let (res,l) =
	  checkFieldVal fi (Types.constr c tag (Types.EQ alpha)) ec
	    l tl true finalise
	in
	Leq.isLeq fi c alpha t boundErr;
	(Types.subst res tag alpha, l)
    | Kind.ARROW(_,_) ->
        Error.errAt fi "Types arguments of higher kind cannot be omitted")
| (l, Types.TY(_,tag,Types.EQ t) :: tl) ->
    let (res,l) =
      checkFieldVal fi (Types.constr c tag (Types.EQ t)) ec l tl synth finalise
    in
    (*
       (Types.subst res tag t, TYv BOXED :: l)
    *)
    (Types.subst res tag t, l)
| (l, tl) ->
    (finalise c ec l tl, [])  (* Warning: extra fields are cut! *)

and checkDec' topLevel c ec = function
    Syntax.SEQ(fi,d1,d2) ->
      let (d1,inc1,ec) = checkDec topLevel c ec d1 in
      let (d2,inc2,ec) = checkDec topLevel (Types.union c inc1) ec d2 in
      (SEQ(fi,d1,d2), Types.union inc1 inc2, ec)
  | Syntax.DEF(fi,l) ->
      let (inc,ec,l) = List.fold_left (scanDef topLevel c) (Types.empty,ec,[]) l in
      (DEF(fi,List.map (checkDef (Types.union c inc) ec) l), inc, ec)
  | Syntax.INLINE(fi,n,pat,p) ->
      if topLevel && Types.isBoundVar c n then
      	Error.errAt fi "Attempt to rebind top-level identifier"
      else
      	let nid = Enco.fresh n in
      	let x = Var.createVar (str_of nid) (int_of nid) in
      	let (pat,t,inc,ec) = safeCheckPat false c ec None pat in
      	(INLINE(fi,nid,pat,checkProc (Types.union c inc) ec p),
	 Types.xbind n (Types.RCHAN t) (Inter.DYNAMIC x) Types.empty,
	 Enco.xbind nid (Types.RCHAN t) (Inter.DYNAMIC x) ec)
  | Syntax.VAL(fi,pat,v) ->
      (try
       	let (pat,t,inc,ec) = checkPat topLevel c ec None getProvidedType pat in
       	let (v,_) = checkVal c ec (Some(t,valPatErr)) v in
       	(VAL(fi,pat,v), inc, ec)
      with NeedTypeAnnotation(_) ->
       	let (v,t) = checkVal c ec None v in
       	let (pat,_,inc,ec) = safeCheckPat topLevel c ec (Some(t,patValErr)) pat in
       	(VAL(fi,pat,v), inc, ec))
  | Syntax.NEW(fi,n,ty) ->
      let nid = Enco.fresh n in
      let x = Var.createVar (str_of nid) (int_of nid) in
      let (t,tty,ec) = Prekinding.hasKindType c ec ty in
      if not(Types.isChan c t) then
      	Error.errfAt fi (fun () ->
          Format.print_string "Expected an channel type but found ";
          Types.displayT t);
      (NEW(fi,nid,tty), Types.xbind n t (Inter.DYNAMIC x) Types.empty,
       Enco.xbind nid t (Inter.DYNAMIC x) ec)
  | Syntax.RUN(fi,p) ->
      (RUN(fi,checkProc c ec p),Types.empty,ec)
  | Syntax.NEWLOC(fi,n) ->
      let nid = Enco.fresh n in
      let x = Var.createVar (str_of nid) (int_of nid) in
      (NEWLOC(fi,nid), Types.xbind n Types.LOC (Inter.DYNAMIC x) Types.empty,
       Enco.xbind nid Types.LOC (Inter.DYNAMIC x) ec)
  | Syntax.CREATE(fi,l) ->
      let (inc,ec,l) = List.fold_left (scanCreate topLevel c) (Types.empty,ec,[]) l in
      (CREATE(fi,List.map (checkCreate (Types.union c inc) ec) l), inc, ec)
  | Syntax.CREATE_AGENT(fi,l) ->
      let (inc,ec,l) = List.fold_left (scanCreateAgent topLevel c) (Types.empty,ec,[]) l in
      (CREATE_AGENT(fi,List.map (checkCreateAgent (Types.union c inc) ec) l), inc, ec)
  | Syntax.AT(fi,x) ->
      let (xt,a) = Types.typeOfX fi x c in
      let x = match a with
      	Inter.DYNAMIC(x) -> createId (Var.str_of x) (Var.int_of x) 
      | _ -> Error.bugAt fi "precheck.checkDec': AT"
      in
      Leq.isLeq fi c xt Types.LOC atErr;
      (AT(fi,x), Types.empty, ec)
  | Syntax.MIGRATE(fi,v) ->
      let (v,_) = checkVal c ec (Some(Types.SITE,siteErr)) v in
      (MIGRATE(fi,v), Types.empty, ec)
  | Syntax.MACRO(fi,strkey,v) ->
      let (v,_) = checkVal c ec None v in
      (MACRO(fi,strkey,v), Types.empty, ec)
  | Syntax.MACROLET(fi,strkey,n,v) ->
      let (v,_) = checkVal c ec None v in
      let nid = Enco.fresh n in
      let x = Var.createVar (str_of nid) (int_of nid) in
      (MACROLET(fi,strkey,nid,v), Types.xbind n (Types.IGNORE) (Inter.DYNAMIC x) Types.empty,
       Enco.xbind nid (Types.IGNORE) (Inter.DYNAMIC x) ec)
  | Syntax.DEFER_DEC(fi,dec) -> 
      Enco.anotherPass := true; 
      (DEFER_DEC(fi,dec), c, ec)
  | Syntax.INLINE_DEC(fi,dec) -> 
      let (dec,c,ec) = checkDec topLevel c ec dec in 
      (INLINE_DEC(fi,dec), c, ec)
  | Syntax.INLINE_VAL(fi,v) -> 
      let (v,t) = checkVal c ec None v in
      encodedParamType := t;
      (INLINE_VAL(fi,v), c, ec)
  | Syntax.DEF_TYPE(fi,s,l,typ) -> defType c ec fi s l typ
  | Syntax.DEF_DEF(fi,n,pa,pat,p)  -> defDef topLevel c ec fi n pa pat p 
  | Syntax.DEF_DEC(fi,s,l,dec)  -> defDec c ec fi s l dec 
  | Syntax.DEF_PROC(fi,s,l,pro) -> defProc c ec fi s l pro 
  | Syntax.ENCODE_DEC(fi,l) -> 
      (match l with
      	(n::pa::[]) ->
      	  (ENCODE_DEC(fi,findVId fi c n,findVId fi c pa), c, ec)
      | (n::[]) ->
      	  (ENCODE_DEC(fi,findVId fi c n,fresh "dummy"), c, ec)
      | _ -> Error.errAt fi "Wrong arguments in ENCODE DEC")
	
and checkDef c ec (fi,nid,pat,inc,ec,p) =
  (fi,nid,pat,checkProc (Types.union c inc) ec p)

and scanDef topLevel c (inc,ec,l) (fi,n,pat,p) =
  if topLevel && Types.isBoundVar c n then
    Error.errAt (Syntax.patInfo pat) "Attempt to rebind top-level identifier"
  else
    let nid = Enco.fresh n in
    let x = Var.createVar (str_of nid) (int_of nid) in
    try
      let (pat,t,patInc,ec) =
 	try (*safe*)checkPat false c ec None getProvidedType pat 
	with NeedTypeAnnotation(fi) ->
          Error.errAt fi "Type annotation required"
      in
      (Types.xbind n (Types.RCHAN t) (Inter.DYNAMIC x) inc,
       Enco.xbind nid (Types.RCHAN t) (Inter.DYNAMIC x) ec,
       (fi,nid,pat,patInc,ec,p) :: l)
    with NeedTypeAnnotation(_) ->
      let (pat,_,patInc,ec) = 
        checkPat false c ec None silentGetProvidedType pat 
      in
      (Types.xbind n (Types.RCHAN Types.TOP) (Inter.DYNAMIC x) inc,
       Enco.xbind nid (Types.RCHAN Types.TOP) (Inter.DYNAMIC x) ec,
       (fi,nid,pat,patInc,ec,p) :: l)

and checkCreate c ec (fi,n,nid,tt,p) =
  thisAgentInfo := fi;
  thisAgentName := n;
  (fi,nid,tt,checkProc c ec p)

and scanCreate topLevel c (inc, ec (*ec might be useful for GL type-checking!*), l) (fi,n,t',p) =
  if topLevel && Types.isBoundVar c n then
    Error.errAt fi "Attempt to rebind top-level identifier"
  else
    let nid = Enco.fresh n in
    let x = Var.createVar (str_of nid) (int_of nid) in
    let (t,tt,ec) = Prekinding.hasKindType c ec t' in
    if not(Types.isAgent c t) then
      Error.errfAt fi (fun()->Format.print_string "Expected an agent type but found "; Types.displayT t);
    (Types.xbind n t (Inter.DYNAMIC x) inc,
     Enco.xbind nid t (Inter.DYNAMIC x) ec,
     (fi,n,nid,tt,p) :: l)

and checkCreateAgent c ec (fi,n,nid,p) =
  thisAgentInfo := fi;
  thisAgentName := n;
  (fi,nid,checkProc c ec p)

and scanCreateAgent topLevel c (inc, ec , l) (fi,n,p) =  (* btw, ec might be useful for GL typechecking *)
  if topLevel && Types.isBoundVar c n then
    Error.errAt fi "Attempt to rebind top-level identifier"
  else
    let nid = Enco.fresh n in
    let x = Var.createVar (str_of nid) (int_of nid) in
    (Types.xbind n (Types.AGENT) (Inter.DYNAMIC x) inc,
     Enco.xbind nid (Types.AGENT) (Inter.DYNAMIC x) ec,
     (fi,n,nid,p) :: l)

(*
 * Debugging versions of some of the above
 *)

and checkProc c ec p =
  if !check then
    Debug.wrap "checkProc"
    (fun () -> checkProc' c ec p)
    (fun () -> Error.printInfo (Syntax.procInfo p))
    (fun _ -> ())
  else
    checkProc' c ec p

and checkVal c ec reqd v =
  if !check then
    Debug.wrap "checkVal"
    (fun () -> checkVal' c ec reqd v)
    (fun () -> Error.printInfo (Syntax.valInfo v); printReqd reqd)
    (fun (_,t) -> Types.printT t)
  else
    checkVal' c ec reqd v

and checkDec topLevel c ec d =
  if !check then
    Debug.wrap "checkDec"
    (fun cont -> checkDec' topLevel c ec d)
    (fun () -> Error.printInfo (Syntax.decInfo d))
    (fun (_,inc,inec) -> if !verbose then Types.printC inc)
  else
    checkDec' topLevel c ec d


(*
 * Typechecking and extracting the translation definitions 
 *)

and defType c ec fi s l def = match (s,l) with

  ("SITE",[]) -> encodedSiteType := (let (t,_,_) = Prekinding.hasKindType c ec def in t);
    (DEF(fi,[]), c, Enco.ebind "TYPE_SITE" (DEF_TYPE([],Prekinding.scanT c ec def)) ec)
| ("SITE",_) -> Error.bugAt fi "Precheck.defType: Parametric type SITE is not supported"

| ("DAEMON",[]) -> encodedDaemonType := (let (t,_,_) = Prekinding.hasKindType c ec def in t);
    (DEF(fi,[]), c, Enco.ebind "TYPE_DAEMON" (DEF_TYPE([],Prekinding.scanT c ec def)) ec)
| ("DAEMON",_) -> Error.bugAt fi "Precheck.defType: Parametric type DAEMON is not supported"

| ("AGENT",[]) -> encodedAgentType := (let (t,_,_) = Prekinding.hasKindType c ec def in t);
    (DEF(fi,[]), c, Enco.ebind "TYPE_AGENT" (DEF_TYPE([],Prekinding.scanT c ec def)) ec)
| ("AGENT",l) -> (match l with
    [x] -> 
      (* if many parameters in list l:

      let c = List.fold_left (fun c n -> 
      	(let tag = Types.freshTag() in
      	Types.abind (Types.constr c tag (Types.LT(Types.TOP))) n tag)) c l in    
      
      let ec = List.fold_left (fun ec n -> 
	(let nid = Enco.fresh n in let tag = Types.lookupTag "precheck.AGENT" fi n c in
	Enco.ibind (Enco.abind (Enco.constr ec tag (Types.LT(Types.TOP))) nid tag) tag nid)) ec l in
      *)

      let xtag = Types.freshTag() in 
      let c = Types.abind (Types.constr c xtag (Types.LT(Types.TOP))) x xtag in
      let xid = createId x (int_of (Enco.fresh x)) in 
      let ec = Enco.abind (Enco.constr ec xtag (Types.LT(Types.TOP))) xid xtag in
      let ec = Enco.ibind ec xtag xid in

      encodedAgentTypeX := x;
      encodedAgentTypeXtag := xtag;
      encodedAgentTypeXid := xid;
      encodedAgentType := (let (t,_,_) = Prekinding.hasKindType c ec def in t);

      (DEF(fi,[]), c, Enco.ebind "TYPE_AGENT"
	 (DEF_TYPE(List.map (fun n -> Enco.lookupTVId fi (Types.lookupTag "precheck.AGENT" fi n c) ec) l,
		   Prekinding.scanT c ec def)) ec)

  | _ -> Error.errAt fi "Precheck.defType: Type AGENT expects only one type variable")
      
| ("TRANSLATION_PARAMETERS",[]) -> encodedParamType := (let (t,_,_) =Prekinding.hasKindType c ec def in t);
    (DEF(fi,[]), c, Enco.ebind "TYPE_PARAMETERS" (DEF_TYPE([],Prekinding.scanT c ec def)) ec)
| ("TRANSLATION_PARAMETERS",_) -> Error.bugAt fi "Precheck.defType: Parametric type TRANSLATION_PARAMETERS is not supported"
      
| ("PROGRAM_PARAMETERS",[]) -> encodedProgramParamType := (let (t,_,_) =Prekinding.hasKindType c ec def in t);
    (DEF(fi,[]), c, Enco.ebind "TYPE_PARAMETERS" (DEF_TYPE([],Prekinding.scanT c ec def)) ec)
| ("PROGRAM_PARAMETERS",_) -> Error.bugAt fi "Precheck.defType: Parametric type TRANSLATION_PARAMETERS is not supported"

| _ -> Error.errAt fi "Precheck.defType: Encodings definition not recognised"


and defDef topLevel c ec fi n pa pat p =
(*
  if topLevel && Types.isBoundVar c n then
    Error.errAt fi "Attempt to rebind top-level identifier"
  else
*)
  let nid = Enco.fresh n in
  let nx = Var.createVar (str_of nid) (int_of nid) in

  let paid = Enco.fresh pa in 
  let pax = Var.createVar (str_of paid) (int_of paid) in
  let c = Types.xbind pa (!encodedParamType) (Inter.DYNAMIC pax) c in
  let ec = Enco.xbind paid (!encodedParamType) (Inter.DYNAMIC pax) ec in  

  let (pat,t,inc,inec) = safeCheckPat false c ec None pat in
  let inc = Types.xbind n (Types.RCHAN t) (Inter.DYNAMIC nx) inc in
  let inec = Enco.xbind nid (Types.RCHAN t) (Inter.DYNAMIC nx) inec in
  let p = checkProc (Types.union c inc) inec p in

  (DEF(fi,[]), c, Enco.fbind (Enco.str_of nid) (DEF_DEF(fi,nid,paid,pat,p)) ec)
    
and defDec c ec fi s l def = match (s,l) with
(*
  ("DEF",(n::pat::p::pa::cont::[])) ->
| ("INLINE",(n::pat::p::pa::cont::[])) ->
| ("NEW",(n::x::pa::cont::[])) ->
| ("VAL",(pat::v::pa::cont::[])) ->
| ("SEQ",(dec1::dec2::pa::cont::[])) ->
| ("RUN",(p::pa::cont::[])) ->
| ("NEWLOC",(n::pa::cont::[])) ->
| ("AT",(n::pa::cont::[])) ->
| *) 

  ("TOPLEVEL",(p::pa::[])) ->

    (* A user-defined program *)
    let pid = Enco.fresh p in
    let px = Var.createVar (str_of pid) (int_of pid) in
    let c = Types.xbind p (Types.TOP) (Inter.DYNAMIC px) c in
    let ec = Enco.xbind pid (Types.TOP) (Inter.DYNAMIC px) ec in

    (* The program parameters *)
    let paid = Enco.fresh pa in 
    let pax = Var.createVar (str_of paid) (int_of paid) in
    let c = Types.xbind pa (!encodedProgramParamType) (Inter.DYNAMIC pax) c in
    let ec = Enco.xbind paid (!encodedProgramParamType) (Inter.DYNAMIC pax) ec in
  
    let (def,_,ec) = checkDec false c ec def in
    (DEF(fi,[]), c, Enco.ebind "TOPLEVEL" (DEF_DEC(pid::paid::[],def)) ec)

(* 
 * Old agent calculus 
 *)

| ("CREATE",(n::x::p::pa::cont::[])) ->

    let xtag = Types.freshTag() in 
    let c = Types.abind (Types.constr c xtag (Types.LT(Types.TOP))) x xtag in
    let xid = Enco.fresh x in 
    let ec = Enco.abind (Enco.constr ec xtag (Types.LT(Types.TOP))) xid xtag in
    let ec = Enco.ibind ec xtag xid in

    let nid = Enco.fresh n in
    let nx = Var.createVar (str_of nid) (int_of nid) in
    let c = Types.xbind n (!encodedAgentType) (Inter.DYNAMIC nx) c in
    let ec = Enco.xbind nid (!encodedAgentType) (Inter.DYNAMIC nx) ec in

    let pid = Enco.fresh p in
    let px = Var.createVar (str_of pid) (int_of pid) in
    let c = Types.xbind p (Types.TOP) (Inter.DYNAMIC px) c in
    let ec = Enco.xbind pid (Types.TOP) (Inter.DYNAMIC px) ec in

    let paid = Enco.fresh pa in 
    let pax = Var.createVar (str_of paid) (int_of paid) in
    let c = Types.xbind pa (!encodedParamType) (Inter.DYNAMIC pax) c in
    let ec = Enco.xbind paid (!encodedParamType) (Inter.DYNAMIC pax) ec in    

    let contid = Enco.fresh cont in
    let contx = Var.createVar (str_of contid) (int_of contid) in
    let c = Types.xbind cont (Types.TOP) (Inter.DYNAMIC contx) c in
    let ec = Enco.xbind contid (Types.TOP) (Inter.DYNAMIC contx) ec in

    let (def,_,ec) = checkDec false c ec def in
    (DEF(fi,[]), c, Enco.ebind "CREATE" (DEF_DEC(nid::xid::pid::paid::contid::[],def)) ec)

| ("MIGRATE",(n::x::v::pa::cont::[])) ->

    let xtag = Types.freshTag() in 
    let c = Types.abind (Types.constr c xtag (Types.LT(Types.TOP))) x xtag in
    let xid = Enco.fresh x in 
    let ec = Enco.abind (Enco.constr ec xtag (Types.LT(Types.TOP))) xid xtag in
    let ec = Enco.ibind ec xtag xid in

    let nid = Enco.fresh n in
    let nx = Var.createVar (str_of nid) (int_of nid) in
    let c = Types.xbind n (!encodedAgentType) (Inter.DYNAMIC nx) c in
    let ec = Enco.xbind nid (!encodedAgentType) (Inter.DYNAMIC nx) ec in

    let vid = Enco.fresh v in
    let vx = Var.createVar (str_of vid) (int_of vid) in
    let c = Types.xbind v (!encodedSiteType) (Inter.DYNAMIC vx) c in
    let ec = Enco.xbind vid (!encodedSiteType) (Inter.DYNAMIC vx) ec in

    let paid = Enco.fresh pa in 
    let pax = Var.createVar (str_of paid) (int_of paid) in
    let c = Types.xbind pa (!encodedParamType) (Inter.DYNAMIC pax) c in
    let ec = Enco.xbind paid (!encodedParamType) (Inter.DYNAMIC pax) ec in    

    let contid = Enco.fresh cont in
    let contx = Var.createVar (str_of contid) (int_of contid) in
    let c = Types.xbind cont (Types.TOP) (Inter.DYNAMIC contx) c in
    let ec = Enco.xbind contid (Types.TOP) (Inter.DYNAMIC contx) ec in

    let x = !encodedAgentTypeX in
    let xtag = !encodedAgentTypeXtag in
    let a = "$DEF_MIGRATE__TOP_AGENT" in
    let aid = Enco.fresh a in
    let ax = Var.createVar (str_of aid) (int_of aid) in
    let c = Types.xbind a (Types.AGENTT(Types.TVAR(Kind.TYPE,x,xtag))) (Inter.DYNAMIC ax) c in
    let ec = Enco.xbind aid (Types.AGENTT(Types.TVAR(Kind.TYPE,x,xtag))) (Inter.DYNAMIC ax) ec in
    thisAgentName := a;

    let (def,_,ec) = checkDec false c ec def in
    (DEF(fi,[]), c, Enco.ebind "MIGRATE" (DEF_DEC(nid::xid::vid::paid::contid::[],def)) ec)

(* 
 * The new calculus 
 *)

| ("CREATE",(n::p::pa::cont::[])) ->

    let nid = Enco.fresh n in
    let nx = Var.createVar (str_of nid) (int_of nid) in
    let c = Types.xbind n (!encodedAgentType) (Inter.DYNAMIC nx) c in
    let ec = Enco.xbind nid (!encodedAgentType) (Inter.DYNAMIC nx) ec in

    let pid = Enco.fresh p in
    let px = Var.createVar (str_of pid) (int_of pid) in
    let c = Types.xbind p (Types.TOP) (Inter.DYNAMIC px) c in
    let ec = Enco.xbind pid (Types.TOP) (Inter.DYNAMIC px) ec in

    let paid = Enco.fresh pa in 
    let pax = Var.createVar (str_of paid) (int_of paid) in
    let c = Types.xbind pa (!encodedParamType) (Inter.DYNAMIC pax) c in
    let ec = Enco.xbind paid (!encodedParamType) (Inter.DYNAMIC pax) ec in    

    let contid = Enco.fresh cont in
    let contx = Var.createVar (str_of contid) (int_of contid) in
    let c = Types.xbind cont (Types.TOP) (Inter.DYNAMIC contx) c in
    let ec = Enco.xbind contid (Types.TOP) (Inter.DYNAMIC contx) ec in

    let (def,_,ec) = checkDec false c ec def in
    (DEF(fi,[]), c, Enco.ebind "CREATE" (DEF_DEC(nid::pid::paid::contid::[],def)) ec)

| ("MIGRATE",(v::pa::cont::[])) ->

    let vid = Enco.fresh v in
    let vx = Var.createVar (str_of vid) (int_of vid) in
    let c = Types.xbind v (!encodedSiteType) (Inter.DYNAMIC vx) c in
    let ec = Enco.xbind vid (!encodedSiteType) (Inter.DYNAMIC vx) ec in

    let paid = Enco.fresh pa in 
    let pax = Var.createVar (str_of paid) (int_of paid) in
    let c = Types.xbind pa (!encodedParamType) (Inter.DYNAMIC pax) c in
    let ec = Enco.xbind paid (!encodedParamType) (Inter.DYNAMIC pax) ec in    

    let contid = Enco.fresh cont in
    let contx = Var.createVar (str_of contid) (int_of contid) in
    let c = Types.xbind cont (Types.TOP) (Inter.DYNAMIC contx) c in
    let ec = Enco.xbind contid (Types.TOP) (Inter.DYNAMIC contx) ec in

    let a = "$DEF_MIGRATE__TOP_AGENT" in
    let aid = Enco.fresh a in
    let ax = Var.createVar (str_of aid) (int_of aid) in
    let c = Types.xbind a (Types.AGENT) (Inter.DYNAMIC ax) c in
    let ec = Enco.xbind aid (Types.AGENT) (Inter.DYNAMIC ax) ec in
    thisAgentName := a;

    let (def,_,ec) = checkDec false c ec def in
    (DEF(fi,[]), c, Enco.ebind "MIGRATE" (DEF_DEC((* nid::xid:: *) vid::paid::contid::[],def)) ec)

| (macro,(v::pa::cont::[])) ->

    let vid = Enco.fresh v in
    let vx = Var.createVar (str_of vid) (int_of vid) in
    let c = Types.xbind v (Types.IGNORE) (Inter.DYNAMIC vx) c in
    let ec = Enco.xbind vid (Types.IGNORE) (Inter.DYNAMIC vx) ec in

    let paid = Enco.fresh pa in 
    let pax = Var.createVar (str_of paid) (int_of paid) in
    let c = Types.xbind pa (!encodedParamType) (Inter.DYNAMIC pax) c in
    let ec = Enco.xbind paid (!encodedParamType) (Inter.DYNAMIC pax) ec in   

    let contid = Enco.fresh cont in
    let contx = Var.createVar (str_of contid) (int_of contid) in
    let c = Types.xbind cont (Types.TOP) (Inter.DYNAMIC contx) c in
    let ec = Enco.xbind contid (Types.TOP) (Inter.DYNAMIC contx) ec in

    let (def,_,ec) = checkDec false c ec def in
    (DEF(fi,[]), c, Enco.ebind macro (DEF_DEC(vid::paid::contid::[],def)) ec)

| (macro,(n::v::pa::cont::[])) ->

    let nid = Enco.fresh n in
    let nx = Var.createVar (str_of nid) (int_of nid) in
    let c = Types.xbind n (Types.IGNORE) (Inter.DYNAMIC nx) c in
    let ec = Enco.xbind nid (Types.IGNORE) (Inter.DYNAMIC nx) ec in

    let vid = Enco.fresh v in
    let vx = Var.createVar (str_of vid) (int_of vid) in
    let c = Types.xbind v (Types.IGNORE) (Inter.DYNAMIC vx) c in
    let ec = Enco.xbind vid (Types.IGNORE) (Inter.DYNAMIC vx) ec in

    let paid = Enco.fresh pa in 
    let pax = Var.createVar (str_of paid) (int_of paid) in
    let c = Types.xbind pa (!encodedParamType) (Inter.DYNAMIC pax) c in
    let ec = Enco.xbind paid (!encodedParamType) (Inter.DYNAMIC pax) ec in   

    let contid = Enco.fresh cont in
    let contx = Var.createVar (str_of contid) (int_of contid) in
    let c = Types.xbind cont (Types.TOP) (Inter.DYNAMIC contx) c in
    let ec = Enco.xbind contid (Types.TOP) (Inter.DYNAMIC contx) ec in

    let (def,_,ec) = checkDec false c ec def in
    (DEF(fi,[]), c, Enco.ebind macro (DEF_DEC(nid::vid::paid::contid::[],def)) ec)

| _ -> Error.errAt fi "Precheck.defDec: Encodings definition not recognised"


and defProc c ec fi s l def = match (s,l) with
(*
  ("PRL",(p::q::pa::[])) ->
| ("IF",(v::p::q::pa::[])) ->
| ("INPUT",(v::pat::p::pa::[])) ->
| ("OUTPUT",(v::w::pa::[])) ->
*)

  ("PROGRAM",(pa::[])) ->

    let paid = Enco.fresh pa in 
    let pax = Var.createVar (str_of paid) (int_of paid) in
    let c = Types.xbind pa (!encodedProgramParamType) (Inter.DYNAMIC pax) c in
    let ec = Enco.xbind paid (!encodedProgramParamType) (Inter.DYNAMIC pax) ec in
    
    let def = checkProc c ec def in
    (DEF(fi,[]), c, Enco.ebind "PROGRAM" (DEF_PROC(paid::[],def)) ec)

(* 
 * Old calculus 
 *)

| ("AGENT_INPUT",(pat::p::pa::cont::[])) ->

    let patid = Enco.fresh pat in
    let patx = Var.createVar (str_of patid) (int_of patid) in
    let c = Types.xbind pat (Types.TOP) (Inter.DYNAMIC patx) c in
    let ec = Enco.xbind patid (Types.TOP) (Inter.DYNAMIC patx) ec in

    let pid = Enco.fresh p in
    let px = Var.createVar (str_of pid) (int_of pid) in
    let c = Types.xbind p (Types.TOP) (Inter.DYNAMIC px) c in
    let ec = Enco.xbind pid (Types.TOP) (Inter.DYNAMIC px) ec in

    let paid = Enco.fresh pa in 
    let pax = Var.createVar (str_of paid) (int_of paid) in
    let c = Types.xbind pa (!encodedParamType) (Inter.DYNAMIC pax) c in
    let ec = Enco.xbind paid (!encodedParamType) (Inter.DYNAMIC pax) ec in    

    let contid = Enco.fresh cont in
    let contx = Var.createVar (str_of contid) (int_of contid) in
    let c = Types.xbind cont (Types.TOP) (Inter.DYNAMIC contx) c in
    let ec = Enco.xbind contid (Types.TOP) (Inter.DYNAMIC contx) ec in

    let def = checkProc c ec def in
    (DEF(fi,[]), c, Enco.ebind "INPA" (DEF_PROC(patid::pid::paid::contid::[],def)) ec)

| ("AGENT_OUTPUT",(v::w::pa::cont::[])) ->

    let vid = Enco.fresh v in
    let vx = Var.createVar (str_of vid) (int_of vid) in
    let c = Types.xbind v (!encodedAgentType) (Inter.DYNAMIC vx) c in
    let ec = Enco.xbind vid (!encodedAgentType) (Inter.DYNAMIC vx) ec in

    let wid = Enco.fresh w in
    let wx = Var.createVar (str_of wid) (int_of wid) in
    let c = Types.xbind w (Types.ENCODE_TYPE(!encodedAgentTypeX::[])) (Inter.DYNAMIC wx) c in
    let ec = Enco.xbind wid (Types.ENCODE_TYPE(!encodedAgentTypeX::[])) (Inter.DYNAMIC wx) ec in
    let paid = Enco.fresh pa in 
    let pax = Var.createVar (str_of paid) (int_of paid) in
    let c = Types.xbind pa (!encodedParamType) (Inter.DYNAMIC pax) c in
    let ec = Enco.xbind paid (!encodedParamType) (Inter.DYNAMIC pax) ec in    

    let contid = Enco.fresh cont in
    let contx = Var.createVar (str_of contid) (int_of contid) in
    let c = Types.xbind cont (Types.TOP) (Inter.DYNAMIC contx) c in
    let ec = Enco.xbind contid (Types.TOP) (Inter.DYNAMIC contx) ec in

    let def = checkProc c ec def in
    (DEF(fi,[]), c, Enco.ebind "OUTA" (DEF_PROC(vid::wid::paid::contid::[],def)) ec)

| ("AGENT_OUTPUT_TO_SITE",(v::x::u::w::pa::cont::[])) -> 

    let xtag = Types.freshTag() in 
    let c = Types.abind (Types.constr c xtag (Types.LT(Types.TOP))) x xtag in
    let xid = Enco.fresh x in 
    let ec = Enco.abind (Enco.constr ec xtag (Types.LT(Types.TOP))) xid xtag in
    let ec = Enco.ibind ec xtag xid in

    let vid = Enco.fresh v in
    let vx = Var.createVar (str_of vid) (int_of vid) in
    let c = Types.xbind v (!encodedAgentType) (Inter.DYNAMIC vx) c in
    let ec = Enco.xbind vid (!encodedAgentType) (Inter.DYNAMIC vx) ec in

    let uid = Enco.fresh u in
    let ux = Var.createVar (str_of uid) (int_of uid) in
    let c = Types.xbind u (!encodedSiteType) (Inter.DYNAMIC ux) c in
    let ec = Enco.xbind uid (!encodedSiteType) (Inter.DYNAMIC ux) ec in

    let wid = Enco.fresh w in
    let wx = Var.createVar (str_of wid) (int_of wid) in
    let c = Types.xbind w (Types.ENCODE_TYPE(!encodedAgentTypeX::[])) (Inter.DYNAMIC wx) c in
    let ec = Enco.xbind wid (Types.ENCODE_TYPE(!encodedAgentTypeX::[])) (Inter.DYNAMIC wx) ec in
    let paid = Enco.fresh pa in 
    let pax = Var.createVar (str_of paid) (int_of paid) in
    let c = Types.xbind pa (!encodedParamType) (Inter.DYNAMIC pax) c in
    let ec = Enco.xbind paid (!encodedParamType) (Inter.DYNAMIC pax) ec in    

    let contid = Enco.fresh cont in
    let contx = Var.createVar (str_of contid) (int_of contid) in
    let c = Types.xbind cont (Types.TOP) (Inter.DYNAMIC contx) c in
    let ec = Enco.xbind contid (Types.TOP) (Inter.DYNAMIC contx) ec in

    let def = checkProc c ec def in
    (DEF(fi,[]), c, Enco.ebind "LDOUTA" (DEF_PROC(vid::xid::uid::wid::paid::contid::[],def)) ec)

| ("TEST_AND_SEND",(v::x::w::p::q::pa::cont::[])) ->

    let xtag = Types.freshTag() in 
    let c = Types.abind (Types.constr c xtag (Types.LT(Types.TOP))) x xtag in
    let xid = Enco.fresh x in 
    let ec = Enco.abind (Enco.constr ec xtag (Types.LT(Types.TOP))) xid xtag in
    let ec = Enco.ibind ec xtag xid in

    let vid = Enco.fresh v in
    let vx = Var.createVar (str_of vid) (int_of vid) in
    let c = Types.xbind v (!encodedAgentType) (Inter.DYNAMIC vx) c in
    let ec = Enco.xbind vid (!encodedAgentType) (Inter.DYNAMIC vx) ec in

    let wid = Enco.fresh w in
    let wx = Var.createVar (str_of wid) (int_of wid) in
    let c = Types.xbind w (Types.ENCODE_TYPE(!encodedAgentTypeX::[])) (Inter.DYNAMIC wx) c in
    let ec = Enco.xbind wid (Types.ENCODE_TYPE(!encodedAgentTypeX::[])) (Inter.DYNAMIC wx) ec in
    let pid = Enco.fresh p in
    let px = Var.createVar (str_of pid) (int_of pid) in
    let c = Types.xbind p (Types.TOP) (Inter.DYNAMIC px) c in
    let ec = Enco.xbind pid (Types.TOP) (Inter.DYNAMIC px) ec in

    let qid = Enco.fresh q in
    let qx = Var.createVar (str_of qid) (int_of qid) in
    let c = Types.xbind q (Types.TOP) (Inter.DYNAMIC qx) c in
    let ec = Enco.xbind qid (Types.TOP) (Inter.DYNAMIC qx) ec in

    let paid = Enco.fresh pa in 
    let pax = Var.createVar (str_of paid) (int_of paid) in
    let c = Types.xbind pa (!encodedParamType) (Inter.DYNAMIC pax) c in
    let ec = Enco.xbind paid (!encodedParamType) (Inter.DYNAMIC pax) ec in    

    let contid = Enco.fresh cont in
    let contx = Var.createVar (str_of contid) (int_of contid) in
    let c = Types.xbind cont (Types.TOP) (Inter.DYNAMIC contx) c in
    let ec = Enco.xbind contid (Types.TOP) (Inter.DYNAMIC contx) ec in

    let def = checkProc c ec def in
    (DEF(fi,[]), c, Enco.ebind "TRYHERE" (DEF_PROC(vid::xid::wid::pid::qid::paid::contid::[],def)) ec)

(* 
 * A new calculus 
 *)

| ("REPLICATED_INPUT",(v::pat::p::pa::[])) ->

    let vid = Enco.fresh v in
    let vx = Var.createVar (str_of vid) (int_of vid) in
    let c = Types.xbind v (Types.CHAN(Types.IGNORE)) (Inter.DYNAMIC vx) c in
    let ec = Enco.xbind vid (Types.CHAN(Types.IGNORE)) (Inter.DYNAMIC vx) ec in

    let patid = Enco.fresh pat in
    let patx = Var.createVar (str_of patid) (int_of patid) in
    let c = Types.xbind pat Types.IGNORE (Inter.DYNAMIC patx) c in
    let ec = Enco.xbind patid Types.IGNORE (Inter.DYNAMIC patx) ec in

    let pid = Enco.fresh p in
    let px = Var.createVar (str_of pid) (int_of pid) in
    let c = Types.xbind p (Types.TOP) (Inter.DYNAMIC px) c in
    let ec = Enco.xbind pid (Types.TOP) (Inter.DYNAMIC px) ec in

    let paid = Enco.fresh pa in 
    let pax = Var.createVar (str_of paid) (int_of paid) in
    let c = Types.xbind pa (!encodedParamType) (Inter.DYNAMIC pax) c in
    let ec = Enco.xbind paid (!encodedParamType) (Inter.DYNAMIC pax) ec in

    let def = checkProc c ec def in 
    (DEF(fi,[]), c, Enco.ebind "REP_INPUT" (DEF_PROC(vid::patid::pid::paid::[],def)) ec)

| ("TRANSPARENT_SEND",(v::u::w::pa::[])) ->

    let vid = Enco.fresh v in
    let vx = Var.createVar (str_of vid) (int_of vid) in
    let c = Types.xbind v (!encodedAgentType) (Inter.DYNAMIC vx) c in
    let ec = Enco.xbind vid (!encodedAgentType) (Inter.DYNAMIC vx) ec in

    let uid = Enco.fresh u in
    let ux = Var.createVar (str_of uid) (int_of uid) in
    let c = Types.xbind u (Types.CHAN(Types.TOP)) (Inter.DYNAMIC ux) c in
    let ec = Enco.xbind uid (Types.CHAN(Types.TOP)) (Inter.DYNAMIC ux) ec in

    let wid = Enco.fresh w in
    let wx = Var.createVar (str_of wid) (int_of wid) in
    let c = Types.xbind w (Types.TOP) (Inter.DYNAMIC wx) c in
    let ec = Enco.xbind wid (Types.TOP) (Inter.DYNAMIC wx) ec in

    let paid = Enco.fresh pa in 
    let pax = Var.createVar (str_of paid) (int_of paid) in
    let c = Types.xbind pa (!encodedParamType) (Inter.DYNAMIC pax) c in
    let ec = Enco.xbind paid (!encodedParamType) (Inter.DYNAMIC pax) ec in    

    let def = checkProc c ec def in
    (DEF(fi,[]), c, Enco.ebind "LI_SEND" (DEF_PROC(vid::uid::wid::paid::[],def)) ec)

| ("REMOTE_SEND",(v::s::u::w::pa::[])) -> 

    let vid = Enco.fresh v in
    let vx = Var.createVar (str_of vid) (int_of vid) in
    let c = Types.xbind v (!encodedAgentType) (Inter.DYNAMIC vx) c in
    let ec = Enco.xbind vid (!encodedAgentType) (Inter.DYNAMIC vx) ec in

    let sid = Enco.fresh s in
    let sx = Var.createVar (str_of sid) (int_of sid) in
    let c = Types.xbind s (!encodedSiteType) (Inter.DYNAMIC sx) c in
    let ec = Enco.xbind sid (!encodedSiteType) (Inter.DYNAMIC sx) ec in

    let uid = Enco.fresh u in
    let ux = Var.createVar (str_of uid) (int_of uid) in
    let c = Types.xbind u (Types.CHAN(Types.TOP)) (Inter.DYNAMIC ux) c in
    let ec = Enco.xbind uid (Types.CHAN(Types.TOP)) (Inter.DYNAMIC ux) ec in

    let wid = Enco.fresh w in
    let wx = Var.createVar (str_of wid) (int_of wid) in
    let c = Types.xbind w (Types.TOP) (Inter.DYNAMIC wx) c in
    let ec = Enco.xbind wid (Types.TOP) (Inter.DYNAMIC wx) ec in

    let paid = Enco.fresh pa in 
    let pax = Var.createVar (str_of paid) (int_of paid) in
    let c = Types.xbind pa (!encodedParamType) (Inter.DYNAMIC pax) c in
    let ec = Enco.xbind paid (!encodedParamType) (Inter.DYNAMIC pax) ec in    

    let def = checkProc c ec def in
    (DEF(fi,[]), c, Enco.ebind "REMOTE_SEND" (DEF_PROC(vid::sid::uid::wid::paid::[],def)) ec)

| ("LOCAL_SEND",(v::u::w::pa::[])) ->

    let vid = Enco.fresh v in
    let vx = Var.createVar (str_of vid) (int_of vid) in
    let c = Types.xbind v (!encodedAgentType) (Inter.DYNAMIC vx) c in
    let ec = Enco.xbind vid (!encodedAgentType) (Inter.DYNAMIC vx) ec in

    let uid = Enco.fresh u in
    let ux = Var.createVar (str_of uid) (int_of uid) in
    let c = Types.xbind u (Types.CHAN(Types.TOP)) (Inter.DYNAMIC ux) c in
    let ec = Enco.xbind uid (Types.CHAN(Types.TOP)) (Inter.DYNAMIC ux) ec in

    let wid = Enco.fresh w in
    let wx = Var.createVar (str_of wid) (int_of wid) in
    let c = Types.xbind w (Types.TOP) (Inter.DYNAMIC wx) c in
    let ec = Enco.xbind wid (Types.TOP) (Inter.DYNAMIC wx) ec in

    let paid = Enco.fresh pa in 
    let pax = Var.createVar (str_of paid) (int_of paid) in
    let c = Types.xbind pa (!encodedParamType) (Inter.DYNAMIC pax) c in
    let ec = Enco.xbind paid (!encodedParamType) (Inter.DYNAMIC pax) ec in    

    let def = checkProc c ec def in
    (DEF(fi,[]), c, Enco.ebind "LOCAL_SEND" (DEF_PROC(vid::uid::wid::paid::[],def)) ec)

| ("ATOMIC_SEND",(v::u::w::p::q::pa::[])) ->

    let vid = Enco.fresh v in
    let vx = Var.createVar (str_of vid) (int_of vid) in
    let c = Types.xbind v (!encodedAgentType) (Inter.DYNAMIC vx) c in
    let ec = Enco.xbind vid (!encodedAgentType) (Inter.DYNAMIC vx) ec in

    let uid = Enco.fresh u in
    let ux = Var.createVar (str_of uid) (int_of uid) in
    let c = Types.xbind u (Types.CHAN(Types.TOP)) (Inter.DYNAMIC ux) c in
    let ec = Enco.xbind uid (Types.CHAN(Types.TOP)) (Inter.DYNAMIC ux) ec in

    let wid = Enco.fresh w in
    let wx = Var.createVar (str_of wid) (int_of wid) in
    let c = Types.xbind w (Types.TOP) (Inter.DYNAMIC wx) c in
    let ec = Enco.xbind wid (Types.TOP) (Inter.DYNAMIC wx) ec in

    let pid = Enco.fresh p in
    let px = Var.createVar (str_of pid) (int_of pid) in
    let c = Types.xbind p (Types.TOP) (Inter.DYNAMIC px) c in
    let ec = Enco.xbind pid (Types.TOP) (Inter.DYNAMIC px) ec in

    let qid = Enco.fresh q in
    let qx = Var.createVar (str_of qid) (int_of qid) in
    let c = Types.xbind q (Types.TOP) (Inter.DYNAMIC qx) c in
    let ec = Enco.xbind qid (Types.TOP) (Inter.DYNAMIC qx) ec in

    let paid = Enco.fresh pa in 
    let pax = Var.createVar (str_of paid) (int_of paid) in
    let c = Types.xbind pa (!encodedParamType) (Inter.DYNAMIC pax) c in
    let ec = Enco.xbind paid (!encodedParamType) (Inter.DYNAMIC pax) ec in    

    let def = checkProc c ec def in
    (DEF(fi,[]), c, Enco.ebind "IFLOCAL" (DEF_PROC(vid::uid::wid::pid::qid::paid::[],def)) ec)

(*
| ("LET",(dec::p::pa::[])) ->
| ("SKIP",(pa::[])) ->
| ("TERMINATE",(pa::[])) ->
| ("TYPECASE",?::pa::) ->
| ("CASE",?::pa::) ->
*)
| _ -> Error.errAt fi "Translation definition not recognised."


let rec addImportProc c = function
  Inter.STRUCT(inc) -> Types.union c inc
| Inter.PRL(p,q) -> addImportProc (addImportProc c p) q
| Inter.VAL(l,p) -> addImportProc (List.fold_left addImportVal c l) p
| Inter.CCODE(_,_,_,p) -> addImportProc c p
| Inter.ATOM(_,_,p) -> addImportProc c p
| Inter.NEW(_,_,p) -> addImportProc c p
| Inter.IF(_) | Inter.INPUT(_) | Inter.OUTPUT(_)
| Inter.SEND(_) | Inter.SKIP -> c
| Inter.ICODE(_,_,p) -> addImportProc c p
| Inter.REP_INPUT(_) | Inter.TIMED_INPUT(_) | Inter.REP_TIMED_INPUT(_) | Inter.IFLOCAL(_) | Inter.TERMINATE -> c
| Inter.MIGRATE(_,_,p) -> addImportProc c p
| _ -> Error.bug "Error in precheck.ml: pattern-matching is not exhaustive"
  
and addImportVal c = function
  Inter.DEF(_,_,_,p) -> addImportProc c p
| Inter.TUPLE(_) | Inter.STRING(_) -> c
| Inter.CREATE_AGENT(_,p) -> addImportProc c p

let addImport c prog = addImportProc c prog.Inter.proc

let precheck imports prog =
  let c = List.fold_left addImport Types.empty imports in
  let (d,_,ec) = checkDec true c Enco.empty prog.Syntax.dec in
  {imports = prog.Syntax.imports; static = prog.Syntax.static;
   includes = prog.Syntax.includes; dec = d; ec = ec}

