(*
 * CPS coversion from high-level Nomadic Pict to core Nomadic Pict.
 *)

(*
   NOTE: we made some attempt to converse the Nomadic Pict primitives 
   for the execution by Pict back-end, however, a full semantics cannot
   be preserved and so the conversion is marked as "Bogus".
*)

type status = RESPONSIVE | UNKNOWN

type proc =
  PRL of proc * proc
| IF of value * proc * proc
| INPUT of value * pat * proc
| OUTPUT of status * value * value
| INPA of value * pat * proc
| OUTA of status * value * value
| LDOUTA of status * value * value * value
| TRYHERE of value * value * proc * proc
| REP_INPUT of value * pat * proc
| TIMED_INPUT of value * value * pat * proc * proc
| REP_TIMED_INPUT of value * value * pat * proc * proc
| LI_SEND of value * value * value
| REMOTE_SEND of value * value * value * value
| LOCAL_SEND of value * value * value
| IFLOCAL of value * value * value * proc * proc
| LET of dec * proc
| SKIP
| TERMINATE
| TYPECASE of value * (pat * Types.t * proc) list * proc
| CASE of value * (Var.var * pat * proc) list

and pat =
  RECORDp of fieldPat list
| ASp of Var.var * ty * pat
| VARp of Var.var * ty
| WILDp

and fieldPat =
  ANONp of pat
| TYp of Var.var

and dec =
  DEF of (Var.var * pat * proc) list
| INLINE of Var.var * pat * proc
| NEW of Var.var * ty
| VAL of pat * value
| SEQ of dec * dec
| RUN of proc
| NEWLOC of  Var.var
| CREATE of (Var.var * proc) list 
| CREATE_AGENT of (Var.var * proc) list 
| AT of Inter.atom
| MIGRATE of value

and ty =
  BOXED
| UNBOXED
| VAR of Var.var

and value =
  CCODEv of Ccode.info * fieldVal list
| APPv of status * value * fieldVal list
| RECORDv of fieldVal list
| STRINGv of string
| SITEv of string * int
| SELECTv of value * int
| VARv of Inter.atom
| BOOLv of bool
| CHARv of char
| INTv of int
| LETv of dec * value
| IFv of value * value * value
| ABSv of pat * proc
| DYNAMICv of  value * Types.t
| TYPECASEv of value * (pat * Types.t * value) list * value
| VARIANTv of Var.var * value
| CASEv of value * (Var.var * pat * value) list

and fieldVal =
  ANONv of value
| TYv of ty

type toplevel =
  {includes : string list;
   static : string list;
   types : Inter.atom Types.c;
   dec : dec;
   context : Inter.atom Types.c}

type cont =
  CONT of (Inter.atom -> Inter.proc)
| CHAN of Inter.atom

let throw a = function 
  (CONT f) -> f a
| (CHAN ch) -> Inter.SEND(ch,[a])

let lookupChan c s = let (_,a) = Types.typeOfX Error.unknown s c in a

let rec simplifyPat c root p = function
  RECORDp(l) -> simplifyFieldPat c 0 root p l
| ASp(n,ty,pat) -> Inter.ATOM(n,root,simplifyPat c root p pat)
| VARp(n,ty) -> Inter.ATOM(n,root,p)
| WILDp -> p

and simplifyFieldPat c i root p = function
  [] -> p
| TYp(x) :: l -> simplifyFieldPat c i root p l
| ANONp(pat) :: l ->
    let var = Var.fresh() in
    Inter.ATOM(
      var,Inter.PROJECT(root,i),
      simplifyPat c (Inter.DYNAMIC var) (simplifyFieldPat c (i+1) root p l) pat
    )

let rec simplifyProc c = function
  PRL(p,q) ->
    Inter.PRL(simplifyProc c p,simplifyProc c q)
| INPUT(v,pat,p) ->
    simplifyVal c v (CONT (fun v ->
      let cl = Var.fresh() and arg = Var.fresh() in
      Inter.VAL([Inter.DEF(cl,Inter.INLINABLE,[Some arg],
			   simplifyPat c (Inter.DYNAMIC arg)
			   (simplifyProc c p) pat)],
		Inter.INPUT(v,Inter.DYNAMIC cl,Inter.UNKNOWN))
    ))
| OUTPUT(UNKNOWN,v,av) ->
    simplifyVal c v (CONT (fun v ->
      simplifyVal c av (CONT (fun av -> Inter.OUTPUT(v,av,Inter.UNKNOWN)))
    ))
| OUTPUT(RESPONSIVE,v,av) ->
    simplifyVal c v (CONT (fun v -> simplifyVal c av (CHAN v)))
| INPA(v,pat,p) ->
    simplifyVal c v (CONT (fun v ->
      let cl = Var.fresh() and arg = Var.fresh() in
      Inter.VAL([Inter.DEF(cl,Inter.INLINABLE,[Some arg],
			   simplifyPat c (Inter.DYNAMIC arg)
			     (simplifyProc c p) pat)],
		Inter.INPUT(v,Inter.DYNAMIC cl,Inter.UNKNOWN))
    			  ))
| OUTA(_,v,av) -> simplifyProc c (LI_SEND(v,v,av))
| LDOUTA(_,v,sv,av) -> simplifyProc c (REMOTE_SEND(v,sv,v,av))
| TRYHERE(v,av,p,q) -> simplifyProc c (IFLOCAL(v,v,av,p,q))
| REP_INPUT(v,pat,p) -> 
    if !Ccode.cc then  (* Bogus *)
      simplifyVal c v (CONT (fun v ->
      	let cl = Var.fresh() and arg = Var.fresh() in
      	Inter.VAL([Inter.DEF(cl,Inter.INLINABLE,[Some arg],
			     simplifyPat c (Inter.DYNAMIC arg)
			       (simplifyProc c p) pat)],
		  Inter.INPUT(v,Inter.DYNAMIC cl,Inter.UNKNOWN))
			    ))
    else
      simplifyVal c v (CONT (fun v ->
      	let cl = Var.fresh() and arg = Var.fresh() in
      	Inter.VAL([Inter.DEF(cl,Inter.INLINABLE,[Some arg],
			     simplifyPat c (Inter.DYNAMIC arg)
			       (simplifyProc c p) pat)],
		  Inter.REP_INPUT(v,Inter.DYNAMIC cl))
			    ))
| TIMED_INPUT(w,v,pat,p,q) -> 
    if !Ccode.cc then  (* Bogus: q is discarded, p executed *)
      simplifyVal c v (CONT (fun v ->
      	let cl = Var.fresh() and arg = Var.fresh() in
      	Inter.VAL([Inter.DEF(cl,Inter.INLINABLE,[Some arg],
			     simplifyPat c (Inter.DYNAMIC arg)
			       (simplifyProc c p) pat)],
		  Inter.INPUT(v,Inter.DYNAMIC cl,Inter.UNKNOWN))
			    ))
    else
      simplifyVal c w (CONT (fun w -> 
      	simplifyVal c v (CONT (fun v ->
      	  let cl = Var.fresh() and arg = Var.fresh() in
      	  Inter.VAL([Inter.DEF(cl,Inter.INLINABLE,[Some arg],
			       simplifyPat c (Inter.DYNAMIC arg)
				 (simplifyProc c p) pat)],
		    Inter.TIMED_INPUT(w,v,Inter.DYNAMIC cl,simplifyProc c q,
				      0 (* this must be equal zero! *)))
			      ))))
| REP_TIMED_INPUT(w,v,pat,p,q) -> 
    if !Ccode.cc then  (* Bogus: q is discarded, p executed *)
      simplifyVal c v (CONT (fun v ->
      	let cl = Var.fresh() and arg = Var.fresh() in
      	Inter.VAL([Inter.DEF(cl,Inter.INLINABLE,[Some arg],
			     simplifyPat c (Inter.DYNAMIC arg)
			       (simplifyProc c p) pat)],
		  Inter.INPUT(v,Inter.DYNAMIC cl,Inter.UNKNOWN))
			    ))
    else
      simplifyVal c w (CONT (fun w -> 
      	simplifyVal c v (CONT (fun v ->
      	  let cl = Var.fresh() and arg = Var.fresh() in
      	  Inter.VAL([Inter.DEF(cl,Inter.INLINABLE,[Some arg],
			       simplifyPat c (Inter.DYNAMIC arg)
				 (simplifyProc c p) pat)],
		    Inter.REP_TIMED_INPUT(w,v,Inter.DYNAMIC cl,simplifyProc c q,
					  0 (* this must be equal zero! *)))
			      ))))
| LI_SEND(v,u,av) ->
    Error.warning "Location-independent output not defined -- replace by a local output";
    if !Ccode.cc then  (* Bogus *)
      simplifyVal c u (CONT (fun u ->
      	simplifyVal c av (CONT (fun av -> Inter.OUTPUT(u,av,Inter.UNKNOWN)))
    			    ))
    else
      simplifyVal c v (CONT (fun v ->
      	simplifyVal c u (CONT (fun u ->
      	  simplifyVal c av (CONT (fun av -> 
	    Inter.IFLOCAL(v,u,av,Inter.SKIP,Inter.SKIP)))
    			      ))))
| REMOTE_SEND(v,sv,u,av) -> 
    if !Ccode.cc then  (* Bogus *)
      simplifyVal c u (CONT (fun u ->
      	simplifyVal c av (CONT (fun av -> Inter.OUTPUT(u,av,Inter.UNKNOWN)))
    			    ))
    else
      simplifyVal c v (CONT (fun v ->
      	simplifyVal c sv (CONT (fun sv ->
	  simplifyVal c u (CONT (fun u ->
	    simplifyVal c av (CONT (fun av ->
			    Inter.VAL([
				      Inter.CREATE_AGENT(Var.fresh(),
				      Inter.MIGRATE(sv,1,Inter.IFLOCAL(v,u,av,Inter.TERMINATE,Inter.TERMINATE)))
                                      (* Note that agent will terminate regardless 
					 if the message can be delivered or not *)
                  		    ],Inter.SKIP)
			    ))))))))
(* a rather embarassing bug..
      simplifyVal c v (CONT (fun v ->
      	simplifyVal c sv (CONT (fun sv ->
	  simplifyVal c u (CONT (fun u ->
	    simplifyVal c av (CONT (fun av -> 
	      Inter.IFLOCAL(v,u,av,
			    Inter.SKIP,
			    Inter.VAL([
				      Inter.CREATE_AGENT(Var.fresh(),
				      Inter.MIGRATE(sv,Inter.IFLOCAL(v,u,av,Inter.TERMINATE,Inter.TERMINATE)))
                                      (* Note that agent will terminate regardless 
					 if the message can be delivered or not *)
                  		    ],Inter.SKIP)) 
				   ))))))))
*)
| LOCAL_SEND(v,u,av) ->
    if !Ccode.cc then  (* Bogus *)
      simplifyVal c u (CONT (fun u ->
      	simplifyVal c av (CONT (fun av -> Inter.OUTPUT(u,av,Inter.UNKNOWN)))
    			    ))
    else
      simplifyVal c v (CONT (fun v ->
      	simplifyVal c u (CONT (fun u ->
      	  simplifyVal c av (CONT (fun av ->
	    Inter.IFLOCAL(v,u,av,Inter.SKIP,Inter.SKIP)))
    			      ))))
| IFLOCAL(v,u,av,p,q) ->
    if !Ccode.cc then  (* Bogus: q is discarded, p executed *)
      simplifyVal c u (CONT (fun u ->
      	simplifyVal c av (CONT (fun av -> Inter.PRL(Inter.OUTPUT(u,av,Inter.UNKNOWN),
						    simplifyProc c p)))
    			    ))
    else
      simplifyVal c v (CONT (fun v -> 
      	simplifyVal c u (CONT (fun u -> 
	  simplifyVal c av (CONT (fun av -> 
	    Inter.IFLOCAL(v,u,av,simplifyProc c p,simplifyProc c q)))
			      ))))
| SKIP -> Inter.SKIP
| LET(d,p) -> simplifyDec c d (simplifyProc c p)
| IF(v,p,q) -> simplifyVal c v (CONT (fun v -> Inter.IF(v,simplifyProc c p,simplifyProc c q)))
| TERMINATE -> if !Ccode.cc then Inter.SKIP else Inter.TERMINATE
| TYPECASE(v, ts, p) ->  (* Bogus *)
    (match ts with
      [] -> Error.bug "simplify.simplifyProc: TYPECASE"
    | tsl -> 
    	let tyStr = Var.fresh() in 
 	let dummy = Var.fresh() in 
      	simplifyProc c (LET(
			VAL(RECORDp(ANONp(VARp(dummy,BOXED))::
				    ANONp(VARp(tyStr,BOXED))::[]),v),
			simplifyTCase c tyStr v tsl p)))
| CASE(v, ts) ->
    (match ts with
      [] -> Error.bug "simplify.simplifyProc:CASE"
    | tsl -> 
    	let n0 = Var.fresh() in 
 	let dummy = Var.fresh() in 
      	simplifyProc c (LET(
			VAL(RECORDp(ANONp(VARp(n0,BOXED))
				    ::ANONp(VARp(dummy,BOXED))::[]),v),
			simplifyCase c n0 v tsl)))

and simplifyCase c n0 v0 = function
    [] -> SKIP
  | (n1,pat1,p1)::rest -> 
      (IF(
       APPv(RESPONSIVE,VARv(lookupChan c "==$" 
			      (* Inter.STATIC(Var.createVar "pict_Std_Prim_" 87) *)
			   ),[ANONv(VARv(Inter.DYNAMIC(n0))); 
			      ANONv(STRINGv(Var.str_of n1))]),
                 (* then *) (LET( VAL(RECORDp(ANONp(WILDp)::ANONp(pat1)::[]),v0), p1)),
		 (* else *) simplifyCase c n0 v0 rest))
	
and simplifyTCase c tyStr v tcase p =
  match tcase with
    [] -> p  (* else p *)
  | (pat1,t1,p1)::rest -> 
      (IF(
       APPv(RESPONSIVE,VARv(lookupChan c "==$"),[
	    ANONv(VARv(Inter.DYNAMIC(tyStr))); 
	    ANONv(STRINGv(Types.toString t1))]),
                 (* then *) (LET( VAL(RECORDp(ANONp(pat1)::ANONp(WILDp)::[]),v), p1)),
		 (* else *) simplifyTCase c tyStr v rest p))
	
and simplifyVal c v cont = match v with
  CCODEv(ci,l) ->
    simplifyFieldVal c l [] (fun l ->
      (match ci.Ccode.kind with
       	Ccode.EXP ->
	  (match ci.Ccode.code with
	    [s] when
	    not(ci.Ccode.reader || ci.Ccode.writer ||
		ci.Ccode.alloc > 0)
	    ->
	      throw (Inter.CCONST ci) cont
          | _ ->
	      let res = Var.fresh() in
	      Inter.CCODE(Some res,ci,List.rev l,
			  throw (Inter.DYNAMIC res) cont))
      | Ccode.STATEMENT ->
	  Inter.CCODE(None,ci,List.rev l,throw (Inter.INT 0) cont)
	    
      | Ccode.NEXP ->
	  (match ci.Ccode.code with
	    [s] when
	    not(ci.Ccode.reader || ci.Ccode.writer ||
	      	ci.Ccode.alloc > 0)
	    ->
	      throw (Inter.CCONST ci) cont
          | _ ->
	      if !Ccode.cc then encodeFunction ci.Ccode.code l cont
	      else 
		(let res = Var.fresh() in
	      	Inter.ICODE(Some res,Calls.yieldCall ci.Ccode.code l,
			    throw (Inter.DYNAMIC res) cont)))
      | Ccode.NSTATEMENT ->
	  if !Ccode.cc then encodeFunction ci.Ccode.code l cont
	  else Inter.ICODE(None,Calls.yieldCall ci.Ccode.code l,
			   throw (Inter.INT 0) cont)
	      ))
| RECORDv([]) ->
    throw (Inter.INT 0) cont
| RECORDv(l) ->
    simplifyFieldVal c l [] (fun l ->
      let var = Var.fresh() in
      Inter.VAL([Inter.TUPLE(var,List.rev l)],throw (Inter.DYNAMIC var) cont)
    )
| APPv(s,v,al) ->
    simplifyVal c v (CONT (fun v ->
      simplifyFieldVal c al [] (fun l ->
	let arg = Var.fresh() in
	let body =
	  match s with
	    RESPONSIVE -> Inter.SEND(v,[Inter.DYNAMIC arg])
	  | UNKNOWN -> Inter.OUTPUT(v,Inter.DYNAMIC arg,Inter.UNKNOWN)
	in
	  match cont with
	    (CONT f) ->
	      let ch = Var.fresh() and res = Var.fresh() in
	      let a = Inter.DYNAMIC ch in
	      Inter.VAL(
		[Inter.DEF(ch,Inter.NOTINLINABLE,[Some res],
			   f (Inter.DYNAMIC res));
		 Inter.TUPLE(arg,List.rev (a :: l))],
		 body
	      )
	  | (CHAN ch) ->
	      Inter.VAL([Inter.TUPLE(arg,List.rev (ch :: l))],body)
      )
    ))
| SELECTv(v,i) ->
    simplifyVal c v (CONT (fun v -> throw (Inter.PROJECT(v,i)) cont))
| STRINGv(s) ->
    let x = Var.fresh() in
    Inter.VAL([Inter.STRING(x,s)],throw (Inter.DYNAMIC x) cont)
| SITEv(s,i) -> simplifyVal c (RECORDv(ANONv(STRINGv s)::ANONv(INTv i)::[])) cont
| CHARv(x) -> throw (Inter.CHAR x) cont
| VARv(a) -> throw a cont
| BOOLv(b) -> throw (Inter.BOOL b) cont
| INTv(i) -> throw (Inter.INT i) cont
| LETv(d,v) -> simplifyDec c d (simplifyVal c v cont)
| ABSv(pat,p) ->
    let x = Var.fresh() and arg = Var.fresh() in
    let p = simplifyPat c (Inter.DYNAMIC arg) (simplifyProc c p) pat in
    Inter.VAL([Inter.DEF(x,Inter.NOTINLINABLE,[Some arg],p)],
	      throw (Inter.DYNAMIC x) cont) 
| IFv(b,v1,v2) ->
    begin
      match cont with
        (CONT f) ->
	  simplifyVal c b (CONT (fun b ->
	    let ch = Var.fresh() and res = Var.fresh() in
	    let cont = CHAN(Inter.DYNAMIC ch) in
	    Inter.VAL(
	      [Inter.DEF(ch,Inter.NOTINLINABLE,[Some res],
			 f (Inter.DYNAMIC res))],
	      Inter.IF(b,simplifyVal c v1 cont,simplifyVal c v2 cont)
	    )
	  ))
      | (CHAN ch) ->
	  simplifyVal c b (CONT (fun b ->
	    Inter.IF(b,simplifyVal c v1 cont,simplifyVal c v2 cont)
	  ))
    end
| DYNAMICv(v,t) -> 
    (* Note that a dynamic value is encoded as a record of the actual value 
                               followed by the value type converted to string *)
    simplifyVal c (RECORDv(ANONv(v)::ANONv(STRINGv(Types.toString t))::[])) cont
| TYPECASEv(v, ts, elsev) ->
    (match ts with
      [] -> Error.bug "simplify.simplifyVal:TYPECASEv"
    | tsl -> 
    	let tyStr = Var.fresh() in 
 	let dummy = Var.fresh() in 
      	simplifyVal c (LETv(
		       VAL(RECORDp(ANONp(VARp(dummy,BOXED))::
				   ANONp(VARp(tyStr,BOXED))::[]),v),
		       simplifyTCaseV c tyStr v tsl elsev)) cont)
| VARIANTv(n,v) -> simplifyVal c (RECORDv(ANONv(STRINGv(Var.str_of n))::ANONv(v)::[])) cont
| CASEv(v, ts) ->
    (match ts with
      [] -> Error.bug "simplify.simplifyVal:CASEv"
    | tsl -> 
    	let n0 = Var.fresh() in 
 	let dummy = Var.fresh() in 
      	simplifyVal c (LETv(
		       VAL(RECORDp(ANONp(VARp(n0,BOXED))
			           ::ANONp(VARp(dummy,BOXED))::[]),v),
		       simplifyCaseV c n0 v tsl)) cont)

and simplifyCaseV c n0 v0 = function
    [] -> INTv(0)  (* i.e. a null value *)
  | (n1,pat1,p1)::rest -> 
      (IFv(
       APPv(RESPONSIVE,VARv(lookupChan c "==$"),[
	    ANONv(VARv(Inter.DYNAMIC(n0))); 
	    ANONv(STRINGv(Var.str_of n1))]),
                 (* then *) (LETv( VAL(RECORDp(ANONp(WILDp)::ANONp(pat1)::[]),v0), p1)),
		 (* else *) simplifyCaseV c n0 v0 rest))

and simplifyTCaseV c tyStr v tcase elsev =
  match tcase with
    [] -> elsev
  | (pat1,t1,p1)::rest -> 
      (IFv(
       APPv(RESPONSIVE,VARv(lookupChan c "==$"),[
	    ANONv(VARv(Inter.DYNAMIC(tyStr))); 
	    ANONv(STRINGv(Types.toString t1))]),
                 (* then *) (LETv( VAL(RECORDp(ANONp(pat1)::ANONp(WILDp)::[]),v), p1)),
		 (* else *) simplifyTCaseV c tyStr v rest elsev))

and simplifyFieldVal c l accum cont = match l with
  [] -> cont accum
| ANONv(v) :: l ->
    simplifyVal c v (CONT (fun v -> simplifyFieldVal c l (v :: accum) cont))
| TYv(_) :: l ->
    simplifyFieldVal c l accum cont

and simplifyDec c d p = match d with
  SEQ(d1,d2) -> simplifyDec c d1 (simplifyDec c d2 p)
| DEF(l) -> Inter.VAL(List.map (simplifyDef c) l,p)
| INLINE(n,pat,q) ->
    let arg = Var.fresh() in
    let q = simplifyPat c (Inter.DYNAMIC arg) (simplifyProc c q) pat in
    Inter.VAL([Inter.DEF(n,Inter.INLINABLE,[Some arg],q)],p)
| VAL(pat,v) -> simplifyVal c v (CONT (fun v -> simplifyPat c v p pat))
| NEW(n,t) -> Inter.NEW(n,Inter.REQUIRED,p)
| RUN(q) -> Inter.PRL(simplifyProc c q,p)
| NEWLOC(n) -> Inter.NEW(n,Inter.REQUIRED,p)  (* not supported *)
| CREATE(l) ->
    if !Ccode.cc then 
      List.fold_right (simplifyCreateToC c) l p
    else
      Inter.VAL(List.map (simplifyCreateAgent c) l,p)
| CREATE_AGENT(l) -> 
    if !Ccode.cc then 
      List.fold_right (simplifyCreateToC c) l p
    else
      Inter.VAL(List.map (simplifyCreateAgent c) l,p)
| AT(a) -> p
| MIGRATE(v) -> 
    if !Ccode.cc then 
      simplifyVal c v (CONT (fun v -> p))
    else
      simplifyVal c v (CONT (fun v -> Inter.MIGRATE(v,0,p)))

and simplifyDef c (x,pat,p) =
  let arg = Var.fresh() in
  Inter.DEF(x,Inter.NOTINLINABLE,[Some arg],
	    simplifyPat c (Inter.DYNAMIC arg) (simplifyProc c p) pat)

and simplifyCreateToC c (x,p) i =
  Inter.PRL(Inter.NEW(x,Inter.REQUIRED,Inter.PRL(simplifyProc c p,i)),Inter.SKIP)

and simplifyCreateAgent c (x,p) =
  Inter.CREATE_AGENT(x,simplifyProc c p)

(*
 *  Encode system functions for a C back-end; not supported since `ccode' is used instead 
 *)

and encodeFunction sl l cont = 
  match Calls.yieldCall sl l with
    Inter.TEST_EQUALITY(x,y) ->
      let res = Var.fresh() in
      Inter.CCODE(Some res, 
		  {Ccode.alloc=0;
		    Ccode.const=false;
		    Ccode.reader=true;
		    Ccode.writer=true;
		    Ccode.kind=Ccode.EXP;
		    Ccode.code=["("; "=="; ")"]},
		  List.rev l,
		  throw (Inter.DYNAMIC res) cont)
 (*
      simplifyVal c
      	(APPv(RESPONSIVE,VARv(Inter.STATIC(("pict_Std_Bool_",2))),[
	      ANONv(APPv(RESPONSIVE,VARv(Inter.STATIC(("pict_Std_Prim_",87))),
			 [ANONv(STRINGv(fst x)); ANONv(STRINGv(fst y))])); 
	      ANONv(APPv(RESPONSIVE,VARv(Inter.STATIC(("pict_Std_Prim_",19))),
			 [ANONv(INTv(snd x)); ANONv(INTv(snd y))]))
	    ]))
      	cont
   *)
  | Inter.DUMMY -> Inter.SKIP
  | Inter.GETSITE -> Inter.SKIP

let simplify _ prog =
  {Inter.includes = prog.includes;
   Inter.static = prog.static; Inter.constants = [];
   Inter.proc = simplifyDec prog.context prog.dec (Inter.STRUCT prog.types)}
