(*
 * Types and contexts for the pi-calculus.
 *)

type tag

val freshTag : unit -> tag
val printTag : tag -> unit

val str_of : tag -> string
val int_of : tag -> int
val name : tag -> string

module Map : Map.S with type key = tag

type t =
  TOP
| UVAR of t option ref * tag
| TVAR of Kind.kind * string * tag
| TABS of Kind.kind * (string * tag * Kind.kind) list * t
| REC of Kind.kind * string * tag * t
| TAPP of Kind.kind * t * t list
| RECORD of field list  
| RCHAN of t
| OCHAN of t
| ICHAN of t
| CHAN of t
| STRING
| CHAR
| BOOL
| INT
| SITE
| LOC
| AGENTT of t
| AGENT
| DYN
| VART of (string * t) list
| MAP of t * t
| ENCODE_TYPE of string list
| IGNORE

and field =
  ANON of t
| LABELED of string * t
| TY of string * tag * constr

and constr =
  LT of t
| EQ of t

type 'a c =
  {types: tag Misc.StringMap.t;     (* Maps type variables to unique tags *)
   constr: constr Map.t;            (* Maps tags to constraints *)  
   vars: (t * 'a) Misc.StringMap.t} (* Maps identifiers to types, plus an 'a *)

val empty : 'a c
val abind : 'a c -> string -> tag -> 'a c
val constr : 'a c -> tag -> constr -> 'a c
val xbind : string -> t -> 'a -> 'a c -> 'a c
val union : 'a c -> 'a c -> 'a c

val mkUnit : t
val mkUvar : unit -> t
val mkTop : Kind.kind -> t
val getFields : Error.info -> 'a c -> t -> field list
val fieldType : Error.info -> 'a c -> t -> string -> t * int

val kindOf : t -> Kind.kind
val kindOfConstr : constr -> Kind.kind

val typeOfX : Error.info -> string -> 'a c -> t * 'a
val isBoundVar : 'a c -> string -> bool
val isBoundTyVar : 'a c -> string -> bool

val printT : t -> unit
val printConstr : constr -> unit
val displayT : t -> unit
val printC : 'a c -> unit

val caml_formatT : t -> unit
val caml_formatC : 'a c -> ('a -> unit) -> unit

val toString : t -> string

val mapBindings : (t * 'a -> t * 'b) -> 'a c -> 'b c
val allBindings : (t * 'a -> unit) -> 'a c -> unit
val lookupBound : tag -> 'a c -> t

val lookupTag : string-> Error.info -> string -> 'a c -> tag
val preLookupTag : string-> Error.info -> string -> 'a c -> tag

val clone : t -> t
val subst : t -> tag -> t -> t
val expose : 'a c -> t -> t
val promote : 'a c -> t -> t option
val basis : 'a c -> t -> t
val unroll : Error.info -> 'a c -> t -> t
val getIChan : Error.info -> 'a c -> t -> t
val getOChan : Error.info -> 'a c -> t -> t
val getRChan : Error.info -> 'a c -> t -> t
val getIAgent : Error.info -> 'a c -> t -> t
val getOAgent : Error.info -> 'a c -> t -> t
val isChan : 'a c -> t -> bool
val leqRChan : 'a c -> t -> bool
val isAgent : 'a c -> t -> bool

